/*
    XCI: eXperimental C Interpreter for ANSI C
    -- Copyright (c) 2000-2001, gondow@jaist.ac.jp
    $Id: attr.c,v 1.1.1.1 2002/06/06 01:41:56 gondow Exp $
*/ 

#include <stdio.h>
#include <assert.h>
#include <malloc.h>
#include <string.h>

#include "AST.h"
#include "xci.h"
#include "type.h"
#include "symbol.h"
#include "vm.h"
#include "attr.h"

/* ---------------------------------------------------------------------- */
static struct Attribute *get_attr (struct AST const *ast,
				   enum AttrKind attr_kind);
static void              put_attr (struct AST *ast, struct Attribute *attr);
static struct Attribute *create_attr_goto (struct AST *ast_dest,
					   struct AST **ast_path,
					   int path_num);
static struct Attribute *create_attr_break (struct AST *ast_dest);
static struct Attribute *create_attr_continue (struct AST *ast_dest);
static struct Attribute *create_attr_block_sp_offset (int offset);
/* ---------------------------------------------------------------------- */
struct Attribute *
create_attr_sw_table (struct AST *ast_default, struct AST **ast_case,
		      int case_num)
{
    struct Attribute *attr = emalloc (sizeof (struct Attribute));
    attr->kind = ATTR_SWITCH_TABLE;
    attr->u.a_sw_table.ast_default = ast_default;
    attr->u.a_sw_table.ast_case    = ast_case;
    attr->u.a_sw_table.case_num    = case_num;
    return attr;
}

struct AST *
get_break_exit (struct AST const *ast_break)
{
    struct Attribute *attr;
    attr = get_attr (ast_break, ATTR_BREAK_DEST);
    assert (attr != NULL);
    return attr->u.a_break_continue.ast_dest;
}

void
set_break_exit (struct AST *ast_break)
{
    struct AST *ast;
    struct Attribute *attr;
    assert (ast_break->ast_type == AST_jump_statement_break);

    /* find the lowest upper 'while', 'do-while', 'for', or 'switch' */
    for (ast = ast_break; !is_AST_function_def (ast); ast = ast->parent) {
	if (is_AST_break_exit (ast)) {
	    break;
    	}
    }
    if (is_AST_function_def (ast))
	ast = NULL;    /* this means there exists no break-exit */
    attr = create_attr_break (ast);
    put_attr ((struct AST *) ast_break, attr);
}

struct AST *
get_continue_exit (struct AST const *ast_continue)
{
    struct Attribute *attr;
    attr = get_attr (ast_continue, ATTR_CONTINUE_DEST);
    assert (attr != NULL);
    return attr->u.a_break_continue.ast_dest;
}

void
set_continue_exit (struct AST *ast_continue)
{
    struct AST *ast;
    struct Attribute *attr;
    assert (ast_continue->ast_type == AST_jump_statement_continue);

    /* find the lowest upper 'while', 'do-while', 'for', or 'switch' */
    for (ast = ast_continue; !is_AST_function_def (ast); ast = ast->parent) {
	if (is_AST_continue_exit (ast)) {
	    break;
    	}
    }
    if (is_AST_function_def (ast))
	ast = NULL;    /* this means there exists no continue-exit */
    attr = create_attr_continue (ast);
    put_attr ((struct AST *) ast_continue, attr);
}

int
get_block_sp_offset   (struct AST const *ast)
{
    struct Attribute *attr;
    assert (is_AST_compound_statement (ast));
    attr = get_attr (ast, ATTR_BLOCK_SP_OFFSET);
    return attr->u.a_block_sp_offset.offset;
}

void
set_block_sp_offset   (struct AST *ast, int offset)
{
    struct Attribute *attr;
    assert (is_AST_compound_statement (ast));
    attr = create_attr_block_sp_offset (offset);
    put_attr (ast, attr);
}

struct Attribute *
get_goto_dest   (struct AST const *ast_goto)
{
    assert (ast_goto->ast_type == AST_jump_statement_goto);
    return get_attr (ast_goto, ATTR_GOTO_DEST);
}

void
set_goto_dest   (struct AST *ast_goto)
{
    char const       *label;
    struct AST       *ast_label, *ast;
    struct AST 	     **ast_path = NULL;
    int              i, path_num;
    struct Attribute *attr;

    assert (ast_goto->ast_type == AST_jump_statement_goto);
    label = ast_goto->u.child [0]->u.string;
    ast_label = sym_lookup_label (label);
    if (ast_label == NULL) {
	xci_error_args (ast_goto, "no such label: %s", label);
    }

    /* enumerate all blocks into ast_path [] between ast_label to 
       AST_function_definition_spec or AST_function_definition_nospec */
    path_num = 0;
    for (ast = ast_label; !is_AST_function_def (ast); ast = ast->parent) {
	if (is_AST_compound_statement (ast)) {
	    ast_path = erealloc (ast_path, sizeof (struct AST *) * ++path_num);
	    ast_path [path_num - 1] = ast;
    	}
    }

    /* ast_common should be the lowest common AST between ast_label
       and ast_goto; now search ast_common! */
    for (ast = ast_goto; !is_AST_function_def (ast); ast = ast->parent) {
	if (is_AST_compound_statement (ast)) {
	    for (i = 0; i < path_num; i++) {
		if (ast_path [i] == ast) {
		    /* Yeah! ast_path [i] is the lowest common. */
		    goto found;
    	    	}
    	    }
    	}
    }
  /* not found */
    assert (0);
  found:
    path_num = i + 1;
    /* cut the unnecessary tail */
    ast_path = erealloc (ast_path, sizeof (struct AST *) * path_num);
    attr = create_attr_goto (ast_label, ast_path, path_num);
    put_attr (ast_goto, attr);
}

struct Attribute *
get_block_sw_table (struct AST const *ast_switch)
{
    assert (ast_switch->ast_type == AST_selection_statement_switch);
    return get_attr (ast_switch, ATTR_SWITCH_TABLE);
}

void
set_block_sw_table (struct AST *ast_switch, struct Attribute *attr_switch)
{
    assert (ast_switch->ast_type == AST_selection_statement_switch);
    put_attr (ast_switch, attr_switch);
}

#if 0 /* for debugging */
void *
emalloc (int size)
{ 
    void *p = malloc (size);
    if (p == NULL) perror ("malloc");
    return p;
}

int
main (void)
{
    struct AST ast;
    struct Attribute *attr1, *attr2, *attr3, *attr4;

    ast.attr = NULL;
    attr1 = create_attr_break (NULL);
    put_attr (&ast, attr1);
    printf ("%p\n", get_attr (&ast, ATTR_BREAK_DEST)) ;
    printf ("%p\n", get_attr (&ast, ATTR_CONTINUE_DEST)) ;

    attr2 = create_attr_break (NULL);
    put_attr (&ast, attr2);
    printf ("%p\n", get_attr (&ast, ATTR_CONTINUE_DEST)) ;
    attr3 = create_attr_continue (NULL);
    put_attr (&ast, attr3);
}
#endif

/* ---------------------------------------------------------------------- */
static struct Attribute *
get_attr (struct AST const *ast, enum AttrKind attr_kind)
{
    struct Attribute *attr;

    for (attr = ast->attr; attr != NULL; attr = attr->next) {
	if (attr->kind == attr_kind)
	    return attr;
    }
    return NULL;
}

static void
put_attr (struct AST *ast, struct Attribute *attr)
{
    struct Attribute *attr2;

    attr2 = get_attr (ast, attr->kind);

    if (attr2 == NULL) {  /* new entry */
	attr->next = ast->attr;
	ast->attr = attr;
    } else {
	struct Attribute *next;
	next = attr2->next;
	*attr2 = *attr;   /* overwrite */
	attr2->next = next;
    }
}

static struct Attribute *
create_attr_goto (struct AST *ast_dest, struct AST **ast_path, int path_num)
{
    struct Attribute *attr = emalloc (sizeof (struct Attribute));
    attr->kind = ATTR_GOTO_DEST;
    attr->u.a_goto.ast_dest = ast_dest;
    attr->u.a_goto.ast_path = ast_path;
    attr->u.a_goto.path_num = path_num;
    return attr;
}

static struct Attribute *
create_attr_break (struct AST *ast_dest)
{
    struct Attribute *attr = emalloc (sizeof (struct Attribute));
    attr->kind = ATTR_BREAK_DEST;
    attr->u.a_break_continue.ast_dest = ast_dest;
    return attr;
}

static struct Attribute *
create_attr_continue (struct AST *ast_dest)
{
    struct Attribute *attr = emalloc (sizeof (struct Attribute));
    attr->kind = ATTR_CONTINUE_DEST;
    attr->u.a_break_continue.ast_dest = ast_dest;
    return attr;
}

static struct Attribute *
create_attr_block_sp_offset (int offset)
{
    struct Attribute *attr = emalloc (sizeof (struct Attribute));
    attr->kind = ATTR_BLOCK_SP_OFFSET;
    attr->u.a_block_sp_offset.offset = offset;
    return attr;
}
/* ---------------------------------------------------------------------- */
