/*
    XCI: eXperimental C Interpreter for ANSI C
    -- Copyright (c) 2000-2001, gondow@jaist.ac.jp
    $Id: eval_block.c,v 1.5 2002/07/08 13:16:33 gondow Exp $
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "xci.h"
#include "AST.h"
#include "type.h"
#include "symbol.h"
#include "vm.h"
#include "eval_block.h"
#include "eval_exp.h"

/* macros for tree traversal */
#define GO_PARENT()    (vm.pc.nth = vm.pc.ast->nth + 1, \
    	    	        vm.pc.ast = vm.pc.ast->parent)
#define GO_CHILD(i)    (vm.pc.nth = 0, vm.pc.ast = vm.pc.ast->u.child[(i)])

/* ---------------------------------------------------------------------- */
static void eval_block_break    (struct AST const *ast_break);
static void eval_block_continue (struct AST const *ast_continue);
static void eval_block_goto     (struct AST const *ast);
static void pop_local_symbols   (struct AST const *ast_from,
				 struct AST const *ast_to);
static void push_local_symbols  (struct AST **ast_path, int path_num);
static void push_local_symbols_for_switch (struct AST const *ast_switch,
					   struct AST *ast_case_default);
static struct AST *find_case_or_default (struct AST const *ast_switch);
/* ---------------------------------------------------------------------- */
void
eval_block (struct AST const * const ast)
{
    vm.pc.ast = ast;
    vm.pc.nth = 0;

    while (1) {
	assert (vm.pc.ast != NULL);

	switch (vm.pc.ast->ast_type) {
	/* leaves; just returning */
	case AST_statement_list_opt_null:
	    GO_PARENT ();
	    break;

	/* 1 child visiting, then returning */
	case AST_statement_labeled:
	case AST_statement_compound:
	case AST_statement_selection:
	case AST_statement_iteration:
	case AST_statement_jump:
	case AST_statement_expression:
	case AST_statement_list_opt_direct:
	case AST_statement_list_single:
	case AST_labeled_statement_default:
	    switch (vm.pc.nth) {
	    case 0: GO_CHILD (0); break;
	    case 1: GO_PARENT();  break;
	    default: assert (0);  break;
    	    }
	    break;

	case AST_labeled_statement_label:
	case AST_labeled_statement_case:
	    switch (vm.pc.nth) {
	    case 0: GO_CHILD (1); break;
	    case 2: GO_PARENT();  break;
	    default: assert (0);  break;
    	    }
	    break;

	/* two children traversing */
	case AST_statement_list_pair:
	    switch (vm.pc.nth) {
	    case 0: GO_CHILD (0); break;
	    case 1: GO_CHILD (1); break;
	    case 2: GO_PARENT();  break;
	    default: assert (0);  break;
    	    }
	    break;

	case AST_function_definition_nospec:
	case AST_function_definition_spec:
	    switch (vm.pc.nth) {
	    case 0:
		sym_begin_funcall  (vm.pc.ast);
		vm.push_p (vm.get_bp ()); /* store base pointer */
		vm.set_bp (vm.get_sp ()); /* create new stack frame */
		switch (vm.pc.ast->ast_type) {
		case AST_function_definition_nospec: GO_CHILD (2); break;
		case AST_function_definition_spec:   GO_CHILD (3); break;
		default:                             assert (0);   break;
    	    	}
		break;
	    case 3:
	    case 4:
		sym_end_funcall ();
		vm.set_sp (vm.get_bp ());
		vm.set_bp (vm.pop_p ()); /* restore base pointer */
		vm.pc.ast = vm.pop_p ();
		return;
	    default: assert (0); break;
	    }
	    break;

	case AST_expression_statement_direct:
	    {
		void *old_sp;
		assert (vm.pc.ast->ast_type
			== AST_expression_statement_direct);
		switch (vm.pc.ast->u.child [0]->ast_type) {
		default:                         assert (0);   break;
		case AST_expression_opt_null:    GO_PARENT (); break;
		case AST_expression_opt_direct:
		    old_sp = vm.get_sp ();
		    eval_exp (vm.pc.ast->u.child [0]->u.child [0]);
		    vm.set_sp (old_sp);
		    GO_PARENT ();
		    break;
		}
    	    }
	    break;
	    
	case AST_compound_statement_nodecn:
	case AST_compound_statement_decn:
	    switch (vm.pc.nth) {
	    case 0:
		sym_begin_block ();
		sym_copy_static_block (vm.pc.ast);
		switch (vm.pc.ast->parent->ast_type) {
		case AST_function_definition_spec:
		case AST_function_definition_nospec:
		    assign_address_to_local_sym (vm.pc.ast->parent->type->
						 u.t_function.ret_type);
		    break;
		default:
		    assign_address_to_local_sym (NULL);
		    break;
		}
		vm.add_sp (get_block_sp_offset (vm.pc.ast));
		switch (vm.pc.ast->ast_type) {
		case AST_compound_statement_nodecn:
		    GO_CHILD (0);
		    break;
		case AST_compound_statement_decn:
		    initialize_local_vars ();
		    GO_CHILD (1);
		    break;
		default: assert (0); break;
    	    	}
		break;
	    case 1:
	    case 2:
		sym_end_block ();
		assert (is_AST_compound_statement (vm.pc.ast));
		vm.sub_sp (get_block_sp_offset (vm.pc.ast));
		GO_PARENT ();
		break;
	    default: assert (0); break;
	    }
	    break;

	case AST_selection_statement_if:
	    switch (vm.pc.nth) {
	    case 0: 
		eval_exp (vm.pc.ast->u.child [0]);
		if (vm.pop_i ()) {
		    GO_CHILD (1);
		} else {
		    GO_PARENT ();
		}
		break;
	    case 2: GO_PARENT();  break;
	    default: assert (0);  break;
    	    }
	    break;

	case AST_selection_statement_if_else:
	    switch (vm.pc.nth) {
	    case 0: 
		eval_exp (vm.pc.ast->u.child [0]);
		if (vm.pop_i ()) {
		    GO_CHILD (1);
		} else {
		    GO_CHILD (2);
		}
		break;
	    case 2:
	    case 3:
		GO_PARENT();  break;
	    default: assert (0);  break;
    	    }
	    break;

	case AST_iteration_statement_while:
	    switch (vm.pc.nth) {
	    case 0: 
	    case 2:
		eval_exp (vm.pc.ast->u.child [0]);
		if (vm.pop_i ()) {
		    GO_CHILD (1);
		} else {
		    GO_PARENT ();
		}
		break;
	    case 3: /* for break */
		GO_PARENT ();
		break;
	    case 1: /* falling through */
	    default: assert (0);   break;
    	    }
	    break;

	case AST_iteration_statement_do_while:
	    switch (vm.pc.nth) {
	    case 0: 
		    GO_CHILD (0);
		    break;
	    case 1:
		eval_exp (vm.pc.ast->u.child [1]);
    	    	/* falling through */
	    case 2:
		if (vm.pop_i ()) {
		    GO_CHILD (0);
		} else {
		    GO_PARENT ();
		}
		break;
	    case 3: /* for break */
		GO_PARENT ();
		break;
	    default: assert (0);  break;
    	    }
	    break;

	case AST_iteration_statement_for:
	    switch (vm.pc.nth) {
		void *sp;
	    case 0: 
	    case 4:
		sp = vm.get_sp ();
		switch (vm.pc.nth) {
		case 0: eval_exp_opt (vm.pc.ast->u.child [0]); break;
		case 4: eval_exp_opt (vm.pc.ast->u.child [2]); break;
		default: assert (0);                           break;
    	    	}
		vm.set_sp (sp); /* discarding the value */
		eval_exp_opt (vm.pc.ast->u.child [1]);
		/* falling through */
	    case 2:
		if (vm.pop_i ()) {
		    GO_CHILD (3);
		} else {
		    GO_PARENT ();
		}
		break;
	    case 5: /* for break */
		GO_PARENT ();
		break;
	    default: assert (0);  break;
    	    }
	    break;

	case AST_jump_statement_return:
	    /* compute return value and then store it. */
	    if (vm.pc.ast->u.child [0]->ast_type
		== AST_expression_opt_direct) {
		struct AST *ast_exp = vm.pc.ast->u.child [0]->u.child [0];
		struct AST const *ast_func;
		int   ret_size;
		void *ret_addr;
		for (ast_func = vm.pc.ast; ; ast_func = ast_func->parent) {
		    if (ast_func->ast_type == AST_function_definition_spec)
			break;
		    if (ast_func->ast_type == AST_function_definition_nospec)
			break;
		}
		ret_size = get_type_size (ast_func->type->u.t_function.ret_type);
		ret_addr = vm.get_bp () - vm.VM_SIZEOF_VOID_P * 2 - ret_size;
		eval_exp (ast_exp);
		vm.pop_b (ret_addr, ret_size);
    	    }
	    sym_end_funcall ();
	    vm.set_sp (vm.get_bp ());
	    vm.set_bp (vm.pop_p ()); /* restore base pointer */
	    vm.pc.ast = vm.pop_p ();
	    return;

	case AST_jump_statement_break:
	    eval_block_break (vm.pc.ast);
	    break;

	case AST_jump_statement_continue:
	    eval_block_continue (vm.pc.ast);
	    break;

	case AST_jump_statement_goto:
	    eval_block_goto (vm.pc.ast);
	    break;

	case AST_selection_statement_switch:
	    switch (vm.pc.nth) {
		struct AST *ast;
	    case 0: 
		ast = find_case_or_default (vm.pc.ast);
		if (ast == NULL) {
		    /* no matched label, so just skip the switch statement  */
		    GO_PARENT ();
		} else {
		    push_local_symbols_for_switch (vm.pc.ast, ast);

		    /* jump the case or default label */
		    vm.pc.ast = ast;
		    vm.pc.nth = 0;
		}
		break;
	    case 2:
	    case 3: /* for break */
		GO_PARENT();  break;
	    default: assert (0);  break;
    	    }
	    break;

	default:
	    assert (0);
	    break;
    	}
    }
}
/* ---------------------------------------------------------------------- */
static void
eval_block_break (struct AST const *ast_break)
{
    struct AST *ast_dest;
    assert (ast_break->ast_type == AST_jump_statement_break);

    ast_dest = get_break_exit (ast_break);
    if (ast_dest == NULL) {
	xci_error (ast_break, "no loop or switch for break");
    }

    pop_local_symbols (ast_break, ast_dest);
    vm.pc.ast = ast_dest;

    switch (vm.pc.ast->ast_type) {
    case AST_iteration_statement_while:    vm.pc.nth = 3; break;
    case AST_iteration_statement_do_while: vm.pc.nth = 3; break;
    case AST_iteration_statement_for:      vm.pc.nth = 5; break;
    case AST_selection_statement_switch:   vm.pc.nth = 3; break;
    default:                               assert (0);    break;
    }
}

static void
eval_block_continue (struct AST const *ast_continue)
{
    struct AST *ast_dest;
    assert (ast_continue->ast_type == AST_jump_statement_continue);

    ast_dest = get_continue_exit (ast_continue);
    if (ast_dest == NULL) {
	xci_error (ast_continue, "no loop for continue");
    }

    pop_local_symbols (ast_continue, ast_dest);
    vm.pc.ast = ast_dest;

    switch (vm.pc.ast->ast_type) {
    case AST_iteration_statement_while:    vm.pc.nth = 2; break;
    case AST_iteration_statement_do_while: vm.pc.nth = 1; break;
    case AST_iteration_statement_for:      vm.pc.nth = 4; break;
    default:                               assert (0);    break;
    }
}

static void
eval_block_goto (struct AST const *ast_goto)
{
    struct AST       *ast_label;
    struct AST       **ast_path;
    struct Attribute *attr;
    int              path_num;

    assert (ast_goto->ast_type == AST_jump_statement_goto);
    attr = get_goto_dest (ast_goto);
    assert (attr->kind == ATTR_GOTO_DEST);

    ast_label = attr->u.a_goto.ast_dest;
    ast_path  = attr->u.a_goto.ast_path;
    path_num  = attr->u.a_goto.path_num;

    /* pop all local variables in the exiting blocks. */
    pop_local_symbols (ast_goto, ast_path [path_num - 1]);

    /* push all local variables in the entering blocks. */
    push_local_symbols (ast_path, path_num - 1);

    vm.pc.ast = ast_label;
    vm.pc.nth = 0;
}

static void
pop_local_symbols (struct AST const *ast_from, struct AST const *ast_to)
{
    struct AST const *ast;
    for (ast = ast_from; ast != ast_to; ast = ast->parent) {
	if (is_AST_compound_statement (ast)) {
	    sym_end_block ();
	    vm.sub_sp (get_block_sp_offset (ast));
    	}
    }
}

static void
push_local_symbols (struct AST **ast_path, int path_num)
{
    int i;
    for (i = path_num - 1; i >= 0; i--) {
	assert (is_AST_compound_statement (ast_path [i]));
	sym_begin_block ();
	sym_copy_static_block (ast_path [i]);
	assign_address_to_local_sym (NULL);
	vm.add_sp (get_block_sp_offset (ast_path [i]));
    }
}

static void
push_local_symbols_for_switch (struct AST const *ast_switch,
			       struct AST *ast_case_default)
{
    struct AST *ast;
    struct AST **ast_path = NULL;
    int    path_num = 0;
    
    for (ast = ast_case_default; ast != ast_switch; ast = ast->parent) {
	if (is_AST_compound_statement (ast)) {
	    ast_path = erealloc (ast_path, sizeof (struct AST *) * ++path_num);
	    ast_path [path_num - 1] = ast;
    	}
    }
    push_local_symbols (ast_path, path_num);
    efree (ast_path);
}

static struct AST *
find_case_or_default (struct AST const *ast_switch)
{
    int    i;
    int    exp_val;
    
    struct Attribute *attr = get_block_sw_table (ast_switch);
    assert (attr != NULL);
    assert (attr->kind == ATTR_SWITCH_TABLE);

    eval_exp (ast_switch->u.child [0]);
    exp_val = vm.pop_i ();

    /* find case label with the same value as 'exp_val' */
    for (i = 0; i < attr->u.a_sw_table.case_num; i++) {
	int case_val;
	struct AST *ast_case = attr->u.a_sw_table.ast_case [i];
	eval_exp (ast_case->u.child [0]);
	case_val = vm.pop_i ();
	if (exp_val == case_val)
	    return ast_case;
    }

    /* no case label found, so return 'default'.
       note: this returns NULL value, but it's OK.
       NULL means no action in the switch statement */
    return attr->u.a_sw_table.ast_default;
}
/* ---------------------------------------------------------------------- */
