/*
    XCI: eXperimental C Interpreter for ANSI C
    -- Copyright (c) 2000-2001, gondow@jaist.ac.jp
    $Id: eval_exp.c,v 1.5 2002/07/08 13:16:33 gondow Exp $
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "xci.h"
#include "AST.h"
#include "type.h"
#include "symbol.h"
#include "vm.h"
#include "builtin.h"
#include "eval_block.h"
#include "eval_exp.h"

/* ---------------------------------------------------------------------- */
static void eval_promote_ast  (struct AST *ast);
static void eval_promote_type (struct Type *type_from, struct Type *type_to);

static void eval_exp_cond   (struct AST *ast);    /* ? : */
static void eval_exp_comma  (struct AST *ast);    /* , */
static void eval_exp_sizeof (struct AST *ast);    /* sizeof */
static void eval_exp_assign (struct AST *ast);    /* =, +=, -=, ... */

static void eval_exp_add    (struct AST *ast);    /* + */
static void eval_exp_sub    (struct AST *ast);    /* - */
static void eval_exp_mult   (struct AST *ast);    /* * */
static void eval_exp_div    (struct AST *ast);    /* / */
static void eval_exp_mod    (struct AST *ast);    /* % */
static void eval_exp_rel    (struct AST *ast);    /* <, <=, >, >= */
static void eval_exp_eq     (struct AST *ast);    /* ==, != */
static void eval_exp_shift  (struct AST *ast);    /* >>, << */
static void eval_exp_usub   (struct AST *ast);	  /* unary - */
static void eval_exp_logic  (struct AST *ast);    /* &&, || */
static void eval_exp_bit    (struct AST *ast);    /* &, |, ^ */

static void eval_exp_addr   (struct AST *ast);    /* &exp */
static void eval_exp_deref  (struct AST *ast);	  /* *addr */
static void eval_exp_comp   (struct AST *ast);	  /* ~ */
static void eval_exp_neg    (struct AST *ast);    /* ! */

static void eval_exp_inc    (struct AST *ast);    /* ++, -- */
static void eval_exp_id     (struct AST *ast);    /* identifier */

static void eval_exp_array  (struct AST *ast);	  /* exp [exp] */
static void eval_exp_dot    (struct AST *ast);    /* exp . exp */
static void eval_exp_arrow  (struct AST *ast);    /* exp -> exp */

static void eval_exp_assign_eq   (struct AST *ast);    /* =  */
static void eval_exp_assign_add  (struct AST *ast);    /* +=  */
static void eval_exp_assign_sub  (struct AST *ast);    /* -=  */
static void eval_exp_assign_mult (struct AST *ast);    /* *=  */
static void eval_exp_assign_div  (struct AST *ast);    /* /=  */
static void eval_exp_assign_mod  (struct AST *ast);    /* %=  */
static void eval_exp_assign_sh   (struct AST *ast);    /* <<=, >>= */
static void eval_exp_assign_bit  (struct AST *ast);    /* &=, ^=, |= */

static void eval_exp_funcall  (struct AST *ast);
static void exp_push_addr     (void *addr, struct Type *type);

static void eval_lvalue_id    (struct AST *ast);
static void eval_lvalue_deref (struct AST *ast);
static void eval_lvalue_dot   (struct AST *ast);
static void eval_lvalue_array (struct AST *ast);
static void eval_lvalue_arrow (struct AST *ast);

static struct StructUnionField *lookup_field (struct Type *type,
					      char *field_name);

static struct AST *eval_initializer_list (struct AST *ast, struct Type *type,
					  void *addr);

static void push_arguments_for_builtin (struct AST *ast, struct Type *type);
static void push_arguments_for_normal (struct AST *ast, struct Type *type);
/* ---------------------------------------------------------------------- */
void
eval_exp (struct AST *ast)
{
    struct Symbol *sym;

    switch (ast->ast_type) {
    /* do nothing */
    case AST_unary_operator_addr:
    case AST_unary_operator_deref:
    case AST_unary_operator_add:
    case AST_unary_operator_sub:
    case AST_unary_operator_comp:
    case AST_unary_operator_neg:
    case AST_assignment_operator_eq:
    case AST_assignment_operator_mult:
    case AST_assignment_operator_div:
    case AST_assignment_operator_mod:
    case AST_assignment_operator_add:
    case AST_assignment_operator_sub:
    case AST_assignment_operator_lshift:
    case AST_assignment_operator_rshift:
    case AST_assignment_operator_band:
    case AST_assignment_operator_beor:
    case AST_assignment_operator_bior:
    case AST_TYPEDEF_NAME:
	break;


    /* forwarding (0) */
    case AST_expression_direct:
    case AST_assignment_expression_direct:
    case AST_conditional_expression_direct:
    case AST_constant_expression_direct:
    case AST_logical_or_expression_direct:
    case AST_logical_and_expression_direct:
    case AST_inclusive_or_expression_direct:
    case AST_exclusive_or_expression_direct:
    case AST_and_expression_direct:
    case AST_equality_expression_direct:
    case AST_relational_expression_direct:
    case AST_shift_expression_direct:
    case AST_additive_expression_direct:
    case AST_multiplicative_expression_direct:
    case AST_cast_expression_direct:
    case AST_unary_expression_direct:
    case AST_postfix_expression_direct:
    case AST_primary_expression_paren:
    case AST_primary_expression_ID:
    case AST_primary_expression_STRING:
    case AST_primary_expression_constant:
    case AST_constant_ENUM:
    case AST_constant_FLOAT:
    case AST_constant_CHAR:
    case AST_constant_INT:
	assert (ast->num_child == 1);
	eval_exp (ast->u.child [0]);
	break;

    /* forwarding (1) */
    case AST_cast_expression_cast:
	eval_exp (ast->u.child [1]);
	break;

    case AST_assignment_expression_assign:
	eval_exp_assign (ast);
	break;

    case AST_conditional_expression_cond:
	eval_exp_cond (ast);
	break;

    case AST_expression_comma:
	eval_exp_comma (ast);
	break;

    case AST_inclusive_or_expression_bior:
    case AST_exclusive_or_expression_beor:
    case AST_and_expression_band:
	eval_exp_bit (ast);
	break;

    case AST_logical_or_expression_lor:
    case AST_logical_and_expression_land:
	eval_exp_logic (ast);
	break;

    case AST_equality_expression_eq:
    case AST_equality_expression_neq:
	eval_exp_eq (ast);
	break;

    case AST_relational_expression_less:
    case AST_relational_expression_more:
    case AST_relational_expression_lesseq:
    case AST_relational_expression_moreeq:
	eval_exp_rel (ast);
	break;

    case AST_shift_expression_lshift:
    case AST_shift_expression_rshift:
	eval_exp_shift (ast);
	break;

    case AST_additive_expression_add:
	eval_exp_add (ast);
	break;

    case AST_additive_expression_sub:
	eval_exp_sub (ast);
	break;

    case AST_multiplicative_expression_mult:
	eval_exp_mult (ast);
	break;

    case AST_multiplicative_expression_div:
	eval_exp_div (ast);
	break;

    case AST_multiplicative_expression_mod:
	eval_exp_mod (ast);
	break;

    case AST_unary_expression_unary_op:
	switch (ast->u.child [0]->ast_type) {
	case AST_unary_operator_addr:  eval_exp_addr (ast);  break;
	case AST_unary_operator_deref: eval_exp_deref (ast); break;
	case AST_unary_operator_add:   /* do  nothing */;    break;
	case AST_unary_operator_sub:   eval_exp_usub (ast);  break;
	case AST_unary_operator_comp:  eval_exp_comp (ast);  break;
	case AST_unary_operator_neg:   eval_exp_neg (ast);   break;
	default:                       assert (0);           break;
	}
	break;

    case AST_unary_expression_sizeof_exp:
    case AST_unary_expression_sizeof_type:
	eval_exp_sizeof (ast);
	break;

    case AST_unary_expression_inc:
    case AST_unary_expression_dec:
    case AST_postfix_expression_inc:
    case AST_postfix_expression_dec:
	eval_exp_inc (ast);
	break;

    case AST_postfix_expression_array:
	eval_exp_array (ast);
	break;
    case AST_postfix_expression_dot:
	eval_exp_dot (ast);
	break;
    case AST_postfix_expression_arrow:
	eval_exp_arrow (ast);
	break;

    case AST_postfix_expression_funcall:
    case AST_postfix_expression_funcall_noarg:
	eval_exp_funcall (ast);
	break;

    case AST_argument_expression_list_single:
    case AST_argument_expression_list_pair:
	/* these AST types are processed in eval_exp_funcall (). */
	assert (0);
	break;

    case AST_STRING:
	vm.push_p (ast->u.string);
	break;

    case AST_IDENTIFIER:
	eval_exp_id (ast);
	break;

    case AST_ENUMERATION_CONSTANT:
	assert (ast->ast_type == AST_ENUMERATION_CONSTANT);
	sym = sym_lookup (ast->u.string, NS_NORMAL);
	assert ((sym != NULL) && (sym->status & SYM_STAT_ENUM_CONST));
	vm.push_i (*(int *) sym->addr);
	break;

    case AST_INTEGER_CONSTANT:
	assert (ast->ast_type == AST_INTEGER_CONSTANT);
	vm.push_i (ast->u.int_ul);
	break;

    case AST_CHARACTER_CONSTANT:
	assert (ast->ast_type == AST_CHARACTER_CONSTANT);
	vm.push_i (ast->u.char_i);
	break;

    case AST_FLOATING_CONSTANT:
	assert (ast->ast_type == AST_FLOATING_CONSTANT);
	vm.push_d (*ast->u.float_ld);
	break;

    default:
	assert (0);
	break;
    }

    eval_promote_ast (ast);
}

void
eval_exp_opt  (struct AST *ast)
{
    assert ((ast->ast_type == AST_expression_opt_null)
	    || (ast->ast_type == AST_expression_opt_direct));
    switch (ast->ast_type) {
    case AST_expression_opt_null:   /* do nothing */             break;
    case AST_expression_opt_direct: eval_exp (ast->u.child [0]); break;
    default:                        assert (0);                  break;
    }
}

void
eval_lvalue (struct AST *ast)
{
    switch (ast->ast_type) {
    /* forwarding (0) */
    case AST_expression_direct:
    case AST_assignment_expression_direct:
    case AST_conditional_expression_direct:
    case AST_constant_expression_direct:
    case AST_logical_or_expression_direct:
    case AST_logical_and_expression_direct:
    case AST_inclusive_or_expression_direct:
    case AST_exclusive_or_expression_direct:
    case AST_and_expression_direct:
    case AST_equality_expression_direct:
    case AST_relational_expression_direct:
    case AST_shift_expression_direct:
    case AST_additive_expression_direct:
    case AST_multiplicative_expression_direct:
    case AST_cast_expression_direct:
    case AST_unary_expression_direct:
    case AST_postfix_expression_direct:
    case AST_primary_expression_ID:
    case AST_primary_expression_paren:
	eval_lvalue (ast->u.child [0]);
	break;

    case AST_IDENTIFIER:
	eval_lvalue_id (ast);
	break;
    case AST_postfix_expression_dot:
	eval_lvalue_dot (ast);
	break;
    case AST_postfix_expression_array:
	eval_lvalue_array (ast);
	break;
    case AST_postfix_expression_arrow:
	eval_lvalue_arrow (ast);
	break;

    case AST_unary_expression_unary_op:
	switch (ast->u.child [0]->ast_type) {
	case AST_unary_operator_deref:
	    eval_lvalue_deref (ast);
	    break;

	/* not lvalue; these must not happen */
	case AST_unary_operator_addr:
	case AST_unary_operator_add:
	case AST_unary_operator_sub:
	case AST_unary_operator_comp:
	case AST_unary_operator_neg:
	default:
	    assert (0);
	    break;
	}
	break;

    /* not lvalue; these must not happen */
    case AST_cast_expression_cast:
    case AST_assignment_expression_assign:
    case AST_conditional_expression_cond:
    case AST_expression_comma:
    case AST_inclusive_or_expression_bior:
    case AST_exclusive_or_expression_beor:
    case AST_and_expression_band:
    case AST_logical_or_expression_lor:
    case AST_logical_and_expression_land:
    case AST_equality_expression_eq:
    case AST_equality_expression_neq:
    case AST_relational_expression_less:
    case AST_relational_expression_more:
    case AST_relational_expression_lesseq:
    case AST_relational_expression_moreeq:
    case AST_shift_expression_lshift:
    case AST_shift_expression_rshift:
    case AST_additive_expression_add:
    case AST_additive_expression_sub:
    case AST_multiplicative_expression_mult:
    case AST_multiplicative_expression_div:
    case AST_multiplicative_expression_mod:
    case AST_unary_expression_sizeof_exp:
    case AST_unary_expression_sizeof_type:
    case AST_unary_expression_inc:
    case AST_unary_expression_dec:
    case AST_postfix_expression_inc:
    case AST_postfix_expression_dec:
    case AST_postfix_expression_funcall:
    case AST_postfix_expression_funcall_noarg:
    case AST_argument_expression_list_single:
    case AST_argument_expression_list_pair:
    case AST_STRING:
    case AST_ENUMERATION_CONSTANT:
    case AST_INTEGER_CONSTANT:
    case AST_CHARACTER_CONSTANT:
    case AST_FLOATING_CONSTANT:
    case AST_primary_expression_STRING:
    case AST_primary_expression_constant:
    case AST_constant_ENUM:
    case AST_constant_FLOAT:
    case AST_constant_CHAR:
    case AST_constant_INT:
    case AST_unary_operator_addr:
    case AST_unary_operator_deref:
    case AST_unary_operator_add:
    case AST_unary_operator_sub:
    case AST_unary_operator_comp:
    case AST_unary_operator_neg:
    case AST_assignment_operator_eq:
    case AST_assignment_operator_mult:
    case AST_assignment_operator_div:
    case AST_assignment_operator_mod:
    case AST_assignment_operator_add:
    case AST_assignment_operator_sub:
    case AST_assignment_operator_lshift:
    case AST_assignment_operator_rshift:
    case AST_assignment_operator_band:
    case AST_assignment_operator_beor:
    case AST_assignment_operator_bior:
    case AST_TYPEDEF_NAME:
	assert (0);
	break;

    default:
	assert (0);
	break;
    }
}

void
eval_initializer (struct AST *ast, struct Type *type, void *addr)
{
    struct AST *ast_init_list;

    assert ((ast->ast_type == AST_initializer_assign)
	    || (ast->ast_type == AST_initializer_list)
	    || (ast->ast_type == AST_initializer_list_comma));
    
    if (ast->ast_type == AST_initializer_assign) {
	eval_exp (ast->u.child [0]);
	eval_promote_type (get_type (ast->u.child [0]), type);
	vm.pop_b (addr, get_type_size (type));
	return;
    }

    assert ((ast->ast_type == AST_initializer_list)
	    || (ast->ast_type == AST_initializer_list_comma));

    /* first search the bottom */
    for (ast_init_list = ast->u.child [0];
	 ast_init_list->ast_type != AST_initializer_list_single;
	 ast_init_list = ast_init_list->u.child [0])
	; /* empty body */
    
    assert (ast_init_list->ast_type == AST_initializer_list_single);
    switch (type->kind) {
    case TYPE_KIND_PRIM:
    case TYPE_KIND_ENUM:
	eval_initializer (ast_init_list->u.child [0], type, addr);
	break;

    case TYPE_KIND_UNION:
	/* union member is initialized with the first member */
	eval_initializer (ast_init_list->u.child [0],
			  type->u.t_struct_union.field [0]->type,
			  addr);
	break;

    case TYPE_KIND_STRUCT:
    case TYPE_KIND_ARRAY:
	eval_initializer_list (ast_init_list, type, addr);
	break;

    case TYPE_KIND_FUNCTION:
    case TYPE_KIND_POINTER:
    case TYPE_KIND_QUAL:
    default:
	assert (0);
	break;
    }
}

void
eval_constant (struct AST *ast, struct Type *type, void *addr)
{
    assert (ast->ast_type == AST_constant_expression_direct);
    eval_exp (ast);
    eval_promote_type (get_type (ast), type);
    vm.pop_b (addr, type->size);
}

/* ---------------------------------------------------------------------- */
static void
eval_promote_ast (struct AST *ast)
{
    struct Type *type_from = ast->type;
    struct Type *type_to   = ast->prom_type;
    eval_promote_type (type_from, type_to);
}

static void
eval_promote_type (struct Type *type_from, struct Type *type_to)
{
    assert (type_from != NULL);

    if (type_to == NULL) {
	return;
    }

    if (type_from->kind == TYPE_KIND_QUAL) {
	type_from = type_from->u.t_qual.type;
    }
    if (type_to->kind == TYPE_KIND_QUAL) {
	type_to = type_to->u.t_qual.type;
    }

    /* if type_from is function, array, structure or union,
       we do not need any converson */
    switch (type_from->kind) {
    case TYPE_KIND_FUNCTION:
    case TYPE_KIND_ARRAY:
    case TYPE_KIND_STRUCT:
    case TYPE_KIND_UNION:  return;
    default:               break;
    }

    switch (type_to->kind) {
    case TYPE_KIND_FUNCTION:
    case TYPE_KIND_ARRAY:
    case TYPE_KIND_STRUCT:
    case TYPE_KIND_UNION:  return;
    default:               break;
    }

    /* if both kind are the same, no (value) conversion is required */
    if (type_from->kind == type_to->kind) {
	if (type_from->kind != TYPE_KIND_PRIM)
	    return;
	if (type_to->u.t_prim.ptype == PRIM_TYPE_VOID)
	    return;
	if (type_from->u.t_prim.ptype == type_to->u.t_prim.ptype)
	    return;
	if ((type_from->u.t_prim.ptype == PRIM_TYPE_FLOAT)
	    && (type_to->u.t_prim.ptype == PRIM_TYPE_DOUBLE))
	    return;
	if ((type_from->u.t_prim.ptype == PRIM_TYPE_FLOAT) 
	    && (type_to->u.t_prim.ptype == PRIM_TYPE_DOUBLE))
	    return;
    }

    /* first convert type_from -> int */
    switch (type_from->kind) {
    case TYPE_KIND_PRIM:
	switch (type_from->u.t_prim.ptype) {
	case PRIM_TYPE_CHAR:   vm.c_to_i ();    break;
	case PRIM_TYPE_INT:    /* do nothing */ break;
	case PRIM_TYPE_FLOAT:  /* falling through */
	case PRIM_TYPE_DOUBLE: vm.d_to_i ();    break;
	case PRIM_TYPE_VOID:   /* falling through */
	default:               assert (0);      break;
	}
	break;
    case TYPE_KIND_ENUM:       /* do nothing */ break;
    case TYPE_KIND_POINTER:    vm.p_to_i ();    break;
    default:                   assert (0);      break;
    }

    /* then convert int -> type_to */
    switch (type_to->kind) {
    case TYPE_KIND_PRIM:
	switch (type_to->u.t_prim.ptype) {
	case PRIM_TYPE_CHAR:   vm.i_to_c ();    break;
	case PRIM_TYPE_INT:    /* do nothing */ break;
	case PRIM_TYPE_FLOAT:  /* falling through */
	case PRIM_TYPE_DOUBLE: vm.i_to_d ();    break;
	case PRIM_TYPE_VOID:   /* falling through */
	default:               assert (0);      break;
	}
	break;
    case TYPE_KIND_ENUM:       /* do nothing */ break;
    case TYPE_KIND_POINTER:    vm.i_to_p ();    break;
    default:                   assert (0);      break;
    }
}

static void
eval_exp_add (struct AST *ast)
{
    /* int     + int     -> int,
       double  + double  -> double,
       pointer + int     -> pointer,
       int     + pointer -> pointer */
    struct Type *type0 = get_type (ast->u.child [0]);
    struct Type *type1 = get_type (ast->u.child [1]);

    if (type0->kind == TYPE_KIND_POINTER) {
	int size = type0->u.t_pointer.type->size;
	assert (is_integral_type (type1));
	eval_exp (ast->u.child [0]); /* pointer is pushed */
	vm.p_to_i ();                /* pointer -> integer */
	eval_exp (ast->u.child [1]); /* offset is pushed */
	vm.push_i (size);            /* size   is pushed */
	vm.mult_i (); vm.add_i ();   /* offset * size + pointer */
	vm.i_to_p ();                /* integer -> pointer */
    } else if (type1->kind == TYPE_KIND_POINTER) {
	int size = type1->u.t_pointer.type->size;
	assert (is_integral_type (type0));
	eval_exp (ast->u.child [0]);	/* offset */
	vm.push_i (size);
	vm.mult_i ();
	eval_exp (ast->u.child [1]);	/* pointer */
	vm.p_to_i ();
	vm.add_i ();
	vm.i_to_p ();
    } else {
	assert (type0->kind == TYPE_KIND_PRIM); 
	assert (type1->kind == TYPE_KIND_PRIM);
	eval_exp (ast->u.child [0]);
	eval_exp (ast->u.child [1]);
	if (type0->u.t_prim.ptype == PRIM_TYPE_INT) {
	    assert (type1->u.t_prim.ptype == PRIM_TYPE_INT);
	    vm.add_i ();
	} else if (type0->u.t_prim.ptype == PRIM_TYPE_FLOAT) {
	    assert (type1->u.t_prim.ptype == PRIM_TYPE_FLOAT);
	    vm.add_d ();
	} else {
	    assert (type0->u.t_prim.ptype == PRIM_TYPE_DOUBLE); 
	    assert (type1->u.t_prim.ptype == PRIM_TYPE_DOUBLE);
	    vm.add_d ();
	} 
    }
}

static void
eval_exp_sub (struct AST *ast)
{
    /* int     - int     -> int,
       double  - double  -> double,
       pointer - int     -> pointer,
       pointer - pionter -> pointer */
    struct Type *type0 = get_type (ast->u.child [0]);
    struct Type *type1 = get_type (ast->u.child [1]);

    if (type0->kind == TYPE_KIND_POINTER) {
	int size = type0->u.t_pointer.type->size;
	if (type1->kind == TYPE_KIND_POINTER) {
	    eval_exp (ast->u.child [0]);
	    vm.p_to_i ();
	    eval_exp (ast->u.child [1]);
	    vm.p_to_i ();
	    vm.sub_i ();
	    vm.push_i (size);
	    vm.div_i ();
	} else {
	    assert (is_integral_type (type1));
	    eval_exp (ast->u.child [0]);
	    vm.p_to_i ();
	    eval_exp (ast->u.child [1]);
	    vm.push_i (size);
	    vm.mult_i ();
	    vm.sub_i ();
	    vm.i_to_p ();
	}
    } else {
	assert (type0->kind == TYPE_KIND_PRIM);
	assert (type1->kind == TYPE_KIND_PRIM);
	eval_exp (ast->u.child [0]);
	eval_exp (ast->u.child [1]);
	if (type0->u.t_prim.ptype == PRIM_TYPE_INT) {
	    assert (type1->u.t_prim.ptype == PRIM_TYPE_INT);
	    vm.sub_i ();
	} else if (type0->u.t_prim.ptype == PRIM_TYPE_FLOAT) {
	    assert (type1->u.t_prim.ptype == PRIM_TYPE_FLOAT);
	    vm.sub_d ();
	} else {
	    assert (type0->u.t_prim.ptype == PRIM_TYPE_DOUBLE); 
	    assert (type1->u.t_prim.ptype == PRIM_TYPE_DOUBLE);
	    vm.sub_d ();
	} 
    }
}

static void
eval_exp_mult (struct AST *ast)
{
    /* int    * int    -> int,
       double * double -> double */
    struct Type *type0 = get_type (ast->u.child [0]);
    struct Type *type1 = get_type (ast->u.child [1]);

    assert (type0->kind == TYPE_KIND_PRIM);

    eval_exp (ast->u.child [0]);
    eval_exp (ast->u.child [1]);
    if (type0->u.t_prim.ptype == PRIM_TYPE_INT) {
	assert (type1->u.t_prim.ptype == PRIM_TYPE_INT);
	vm.mult_i ();
    } else if (type0->u.t_prim.ptype == PRIM_TYPE_FLOAT) {
	assert (type1->u.t_prim.ptype == PRIM_TYPE_FLOAT);
	vm.mult_d ();
    } else {
	assert (type0->u.t_prim.ptype == PRIM_TYPE_DOUBLE);
	assert (type1->u.t_prim.ptype == PRIM_TYPE_DOUBLE);
	vm.mult_d ();
    }
}

static void
eval_exp_div  (struct AST *ast)
{
    /* int    / int    -> int,
       double / double -> double */
    struct Type *type0 = get_type (ast->u.child [0]);
    struct Type *type1 = get_type (ast->u.child [1]);

    assert (type0->kind == TYPE_KIND_PRIM);

    eval_exp (ast->u.child [0]);
    eval_exp (ast->u.child [1]);
    if (type0->u.t_prim.ptype == PRIM_TYPE_INT) {
	assert (type1->u.t_prim.ptype == PRIM_TYPE_INT);
	if (vm.top_i () == 0) {
	    xci_error (ast->u.child [1], "divided by 0");
	}
	vm.div_i ();
    } else if (type0->u.t_prim.ptype == PRIM_TYPE_FLOAT) {
	assert (type1->u.t_prim.ptype == PRIM_TYPE_FLOAT);
	if (vm.top_d () == 0) {
	    xci_error (ast->u.child [1], "divided by 0");
	}
	vm.div_d ();
    } else {
	assert (type0->u.t_prim.ptype == PRIM_TYPE_DOUBLE); 
	assert (type1->u.t_prim.ptype == PRIM_TYPE_DOUBLE);
	if (vm.top_d () == 0) {
	    xci_error (ast->u.child [1], "divided by 0");
	}
	vm.div_d ();
    }
}

static void
eval_exp_mod  (struct AST *ast)
{
    /* int    % int    -> int */
    struct Type *type0 = get_type (ast->u.child [0]);
    struct Type *type1 = get_type (ast->u.child [1]);

    assert (type0->kind == TYPE_KIND_PRIM);

    assert (type0->u.t_prim.ptype == PRIM_TYPE_INT); 
    assert (type1->u.t_prim.ptype == PRIM_TYPE_INT);
    eval_exp (ast->u.child [0]);
    eval_exp (ast->u.child [1]);
    if (vm.top_i () == 0) {
	xci_error (ast->u.child [1], "divided by 0");
    }
    vm.mod_i ();
}

static void
eval_exp_rel    (struct AST *ast)
{
    /* int     < int     -> int, (also <=, >, >=)
       double  < double  -> int,
       pointer < pointer -> int */
    struct Type *type0 = get_type (ast->u.child [0]);
    struct Type *type1 = get_type (ast->u.child [1]);

    if (type0->kind == TYPE_KIND_POINTER) {
	assert (is_type_equivalent (type0, type1));
	eval_exp (ast->u.child [0]);
	vm.p_to_i ();
	eval_exp (ast->u.child [1]);
	vm.p_to_i ();
	switch (ast->ast_type) {
	case AST_relational_expression_less:   vm.less_i (); break;
	case AST_relational_expression_more:   vm.more_i (); break;
	case AST_relational_expression_lesseq: vm.leq_i  (); break;
	case AST_relational_expression_moreeq: vm.meq_i  (); break;
	default:                               assert (0);   break;
	}
    } else {
	assert (type0->kind == TYPE_KIND_PRIM); 
	assert (type1->kind == TYPE_KIND_PRIM);
	eval_exp (ast->u.child [0]);
	eval_exp (ast->u.child [1]);
	if (type0->u.t_prim.ptype == PRIM_TYPE_INT) {
	    assert (type1->u.t_prim.ptype == PRIM_TYPE_INT);
	    switch (ast->ast_type) {
	    case AST_relational_expression_less:   vm.less_i (); break;
	    case AST_relational_expression_more:   vm.more_i (); break;
	    case AST_relational_expression_lesseq: vm.leq_i  (); break;
	    case AST_relational_expression_moreeq: vm.meq_i  (); break;
	    default:                               assert (0);   break;
	    }
	} else {
	    assert (((type0->u.t_prim.ptype == PRIM_TYPE_FLOAT)
		     && (type1->u.t_prim.ptype == PRIM_TYPE_FLOAT))
		    || ((type0->u.t_prim.ptype == PRIM_TYPE_DOUBLE) 
			&& (type1->u.t_prim.ptype == PRIM_TYPE_DOUBLE)));
	    switch (ast->ast_type) {
	    case AST_relational_expression_less:   vm.less_d (); break;
	    case AST_relational_expression_more:   vm.more_d (); break;
	    case AST_relational_expression_lesseq: vm.leq_d  (); break;
	    case AST_relational_expression_moreeq: vm.meq_d  (); break;
	    default:                               assert (0);   break;
	    }
	} 
    }
}

static void
eval_exp_eq     (struct AST *ast)
{
    /* int     == int     -> int, (also !=)
       double  == double  -> int,
       pointer == pointer -> int,
       pointer == 0       -> int,
       0       == pointer -> int  */
    struct Type *type0 = get_type (ast->u.child [0]);
    struct Type *type1 = get_type (ast->u.child [1]);

    if (type0->kind == TYPE_KIND_POINTER) {
	eval_exp (ast->u.child [0]);
	vm.p_to_i ();
	eval_exp (ast->u.child [1]);

	if (type1->kind == TYPE_KIND_POINTER) {
	    assert (is_type_equivalent (type0, type1));
	    vm.p_to_i ();
	    switch (ast->ast_type) {
	    case AST_equality_expression_eq:       vm.eq_i   (); break;
	    case AST_equality_expression_neq:      vm.neq_i  (); break;
	    default:                               assert (0);   break;
	    }
	} else {
	    assert (is_integral_type (type1));
	    if (vm.top_i () != 0) {
		xci_error (ast->u.child [1],
			   "comparison between pointer and non-zero integer");
	    }
	    switch (ast->ast_type) {
	    case AST_equality_expression_eq:       vm.eq_i   (); break;
	    case AST_equality_expression_neq:      vm.neq_i  (); break;
	    default:                               assert (0);   break;
	    }
	}
    } else if (type1->kind == TYPE_KIND_POINTER) {
	assert (is_integral_type (type0));
	eval_exp (ast->u.child [0]);
	if (vm.top_i () != 0) {
	    xci_error (ast->u.child [1],
		       "comparison between pointer and non-zero integer");
	}
	eval_exp (ast->u.child [1]);
	vm.p_to_i ();
	switch (ast->ast_type) {
	case AST_equality_expression_eq:       vm.eq_i   (); break;
	case AST_equality_expression_neq:      vm.neq_i  (); break;
	default:                               assert (0);   break;
	}
    } else {
	assert (type0->kind == TYPE_KIND_PRIM);
	eval_exp (ast->u.child [0]);
	eval_exp (ast->u.child [1]);
	if (type0->u.t_prim.ptype == PRIM_TYPE_INT) {
	    assert (type1->u.t_prim.ptype == PRIM_TYPE_INT);
	    switch (ast->ast_type) {
	    case AST_equality_expression_eq:       vm.eq_i   (); break;
	    case AST_equality_expression_neq:      vm.neq_i  (); break;
	    default:                               assert (0);   break;
	    }
	} else {
	    assert (((type0->u.t_prim.ptype == PRIM_TYPE_FLOAT)
		     && (type1->u.t_prim.ptype == PRIM_TYPE_FLOAT))
		    || ((type0->u.t_prim.ptype == PRIM_TYPE_DOUBLE) 
			&& (type1->u.t_prim.ptype == PRIM_TYPE_DOUBLE)));
	    switch (ast->ast_type) {
	    case AST_equality_expression_eq:       vm.eq_d   (); break;
	    case AST_equality_expression_neq:      vm.neq_d  (); break;
	    default:                               assert (0);   break;
	    }
	}
    }
}

static void
eval_exp_shift (struct AST *ast)
{
    /* int >> int -> int (also <<) */
    struct Type *type0 = get_type (ast->u.child [0]);
    struct Type *type1 = get_type (ast->u.child [1]);

    assert (is_integral_type (type0));
    assert (is_integral_type (type1));

    eval_exp (ast->u.child [0]);
    eval_exp (ast->u.child [1]);

    switch (ast->ast_type) {
    case AST_shift_expression_lshift:
	vm.l_sh ();
	break;
    case AST_shift_expression_rshift:
	if (type0->u.t_prim.modifier & TYPE_MOD_UNSIGNED) {
	    vm.r_ush ();
	} else {
	    vm.r_sh ();
	}
	break;
    default:
	assert (0);
	break;
    }
}

static void
eval_exp_usub   (struct AST *ast)
{
    /* - int    -> int,
       - double -> double */
    struct Type *type1 = get_type (ast->u.child [1]);
    assert (is_arithmetic_type (type1));

    eval_exp (ast->u.child [1]);
    if (type1->u.t_prim.ptype == PRIM_TYPE_INT) {
	vm.push_i (-1);
	vm.mult_i ();
    } else {
	assert ((type1->u.t_prim.ptype == PRIM_TYPE_FLOAT)
		|| (type1->u.t_prim.ptype == PRIM_TYPE_DOUBLE));
	vm.push_d (-1.0);
	vm.mult_d ();
    }
}

static void
eval_exp_id     (struct AST *ast)
{
    /* (none)   -> char,
       (none)   -> int,
       (none)   -> double,
       (none)   -> pointer,
       (none)   -> byte sequence for structure/union */
    
    struct Type   *type0 = get_type (ast);
    struct Symbol *sym;

    sym = sym_lookup (ast->u.string, NS_NORMAL);
    assert (sym != NULL);

    /* no dereference required for function or array symbol.
      (of course, dereference required for an object with the type 
      of pointer to function). */
    if ((sym->type->kind == TYPE_KIND_FUNCTION)
	|| (sym->type->kind == TYPE_KIND_ARRAY)) {
	vm.push_p (sym->addr);
    } else {
	exp_push_addr (sym->addr, type0);
    }
}

static void
eval_exp_logic  (struct AST *ast)
{
    /* int && int -> int (also ||) */
    struct Type   *type0 = get_type (ast->u.child [0]);
    struct Type   *type1 = get_type (ast->u.child [1]);

    eval_exp (ast->u.child [0]);
    switch (type0->kind) {
    case TYPE_KIND_PRIM:
	switch (type0->u.t_prim.ptype) {
	case PRIM_TYPE_CHAR:   vm.c_to_i ();    break;
	case PRIM_TYPE_INT:    /* do nothing */ break;
	case PRIM_TYPE_FLOAT:  /* falling through */
	case PRIM_TYPE_DOUBLE: vm.d_to_i ();    break;
	case PRIM_TYPE_VOID:   /* falling through */
	default:               assert (0);      break;
	}
	break;
    case TYPE_KIND_ENUM:       /* do nothing */ break;
    case TYPE_KIND_POINTER:    vm.p_to_i ();    break;
    default:                   assert (0);      break;
    }

    switch (ast->ast_type) {
    case AST_logical_or_expression_lor:
	if (vm.pop_i () != 0) {
	    vm.push_i (1);
	    return;
	}
	break;
    case AST_logical_and_expression_land:
	if (vm.pop_i () == 0) {
	    vm.push_i (0);
	    return;
	}
	break;
    default:
	assert (0);
	break;
    }

    eval_exp (ast->u.child [1]);
    switch (type1->kind) {
    case TYPE_KIND_PRIM:
	switch (type1->u.t_prim.ptype) {
	case PRIM_TYPE_CHAR:   vm.c_to_i ();    break;
	case PRIM_TYPE_INT:    /* do nothing */ break;
	case PRIM_TYPE_FLOAT:  /* falling through */
	case PRIM_TYPE_DOUBLE: vm.d_to_i ();    break;
	case PRIM_TYPE_VOID:   /* falling through */
	default:               assert (0);      break;
	}
	break;
    case TYPE_KIND_ENUM:       /* do nothing */ break;
    case TYPE_KIND_POINTER:    vm.p_to_i ();    break;
    default:                   assert (0);      break;
    }

    vm.push_i (vm.pop_i () != 0);
}

static void
eval_exp_bit    (struct AST *ast)
{
    /* int & int -> int (also ^, |) */

    struct Type   *type0 = get_type (ast->u.child [0]);
    struct Type   *type1 = get_type (ast->u.child [1]);
    assert (is_integral_type (type0));
    assert (is_integral_type (type1));

    eval_exp (ast->u.child [0]);
    eval_exp (ast->u.child [1]);

    switch (ast->ast_type) {
    case AST_inclusive_or_expression_bior: vm.b_or ();  break;
    case AST_exclusive_or_expression_beor: vm.b_xor (); break;
    case AST_and_expression_band:          vm.b_and (); break;
    default:                               assert (0);  break;
    }
}

static void
eval_exp_comp   (struct AST *ast)
{
    /* ~ int -> int */
    eval_exp (ast->u.child [1]);
    vm.b_comp ();
}

static void
eval_exp_comma  (struct AST *ast)
{
    /* exp , exp -> exp */
    void *old_sp = vm.get_sp ();

    eval_exp (ast->u.child [0]);
    vm.set_sp (old_sp); /* discard the result of first expression */
    eval_exp (ast->u.child [1]);
}

static void
eval_exp_cond   (struct AST *ast)
{
    /* exp ? exp : exp -> exp */
    struct Type   *type0 = get_type (ast->u.child [0]);
    assert (type0->kind == TYPE_KIND_PRIM);

    eval_exp (ast->u.child [0]);
    if (is_floating_type (type0)) {
	vm.d_to_i ();	 /* double -> int */
    }
    if (vm.pop_i () == 0) {
	eval_exp (ast->u.child [2]);
    } else {
	eval_exp (ast->u.child [1]);
    }
}

static void
eval_exp_neg    (struct AST *ast)
{
    /* ! int     -> int,
       ! double  -> int,
       ! pointer -> int */
    struct Type   *type1 = get_type (ast->u.child [1]);

    eval_exp (ast->u.child [1]);
    if (type1->kind == TYPE_KIND_POINTER) {
	vm.p_to_i ();
    } else if (is_floating_type (type1)) {
	vm.d_to_i ();
    } else {
	assert (is_integral_type (type1));
    }
    vm.push_i (!vm.pop_i ());
}

static void
eval_exp_sizeof (struct AST *ast)
{
    struct Type   *type0 = get_type (ast->u.child [0]);

    assert ((ast->ast_type == AST_unary_expression_sizeof_exp)
	    || (ast->ast_type == AST_unary_expression_sizeof_type));

    if (ast->ast_type == AST_unary_expression_sizeof_exp) {
	void *old_sp = vm.get_sp ();
	eval_exp (ast->u.child [0]);
	vm.set_sp (old_sp);
    }
    vm.push_i (type0->size);
}

static void
eval_exp_addr   (struct AST *ast)
{
    eval_lvalue (ast->u.child [1]);
}

static void
eval_exp_deref  (struct AST *ast)
{
    struct Type   *type1 = get_type (ast->u.child [1]);
    eval_exp (ast->u.child [1]);
    if ((type1->kind == TYPE_KIND_POINTER)
	&& (type1->u.t_pointer.type->kind == TYPE_KIND_FUNCTION
	    || type1->u.t_pointer.type->kind == TYPE_KIND_ARRAY)) {
	/* do nothing (no dereference required) */
    } else {
	exp_push_addr (vm.pop_p (), type1);
    }
}

static void
eval_exp_inc    (struct AST *ast)
{
    struct Type   *type0 = get_type (ast->u.child [0]);

    eval_lvalue (ast->u.child [0]);
    if (type0->kind == TYPE_KIND_POINTER) {
	int  size = type0->size;
	void **p = vm.pop_p ();
	switch (ast->ast_type) {
	case AST_unary_expression_inc:   *p += size; vm.push_p (*p); break;
	case AST_unary_expression_dec:   *p -= size; vm.push_p (*p); break;
	case AST_postfix_expression_inc: vm.push_p (*p); *p += size; break;
	case AST_postfix_expression_dec: vm.push_p (*p); *p -= size; break;
	default:                         assert (0);                 break;
	}
    } else {
	assert (type0->kind == TYPE_KIND_PRIM);
	switch (type0->u.t_prim.ptype) {
	case PRIM_TYPE_CHAR:
	    {
		char *p = vm.pop_p ();
		switch (ast->ast_type) {
		case AST_unary_expression_inc:   (*p)++; vm.push_c (*p); break;
		case AST_unary_expression_dec:   (*p)--; vm.push_c (*p); break;
		case AST_postfix_expression_inc: vm.push_c (*p); (*p)++; break;
		case AST_postfix_expression_dec: vm.push_c (*p); (*p)--; break;
		default:                         assert (0);             break;
		}
	    }
	    break;
	case PRIM_TYPE_INT:
	    {
		int *p = vm.pop_p ();
		switch (ast->ast_type) {
		case AST_unary_expression_inc:   (*p)++; vm.push_i (*p); break;
		case AST_unary_expression_dec:   (*p)--; vm.push_i (*p); break;
		case AST_postfix_expression_inc: vm.push_i (*p); (*p)++; break;
		case AST_postfix_expression_dec: vm.push_i (*p); (*p)--; break;
		default:                         assert (0);             break;
		}
	    }
	    break;
	case PRIM_TYPE_FLOAT:
	case PRIM_TYPE_DOUBLE:
	    {
		double *p = vm.pop_p ();
		switch (ast->ast_type) {
		case AST_unary_expression_inc:   (*p)++; vm.push_d (*p); break;
		case AST_unary_expression_dec:   (*p)--; vm.push_d (*p); break;
		case AST_postfix_expression_inc: vm.push_d (*p); (*p)++; break;
		case AST_postfix_expression_dec: vm.push_d (*p); (*p)--; break;
		default:                         assert (0);             break;
		}
	    }
	    break;
	case PRIM_TYPE_VOID:    /* falling through */
	default:
	    assert (0);
            break;
	}
    }
}

static void
eval_exp_array  (struct AST *ast)
{
    struct Type *type_pointer;

    if (get_type (ast->u.child [0])->kind == TYPE_KIND_POINTER) {
	type_pointer = get_type (ast->u.child [0]);
    } else {
	type_pointer = get_type (ast->u.child [1]);
    }

    eval_lvalue_array (ast);

    if ((type_pointer->kind == TYPE_KIND_POINTER)
	&& (type_pointer->u.t_pointer.type->kind == TYPE_KIND_ARRAY)) {
	/* do nothing (no dereference required) */
    } else {
	exp_push_addr (vm.pop_p (), get_type (ast));
    }
}

static void
eval_exp_dot    (struct AST *ast)
{
    eval_lvalue_dot (ast);
    exp_push_addr (vm.pop_p (), get_type (ast));
}

static void
eval_exp_arrow  (struct AST *ast)
{
    eval_lvalue_arrow (ast);
    exp_push_addr (vm.pop_p (), get_type (ast));
}

static void
eval_exp_assign (struct AST *ast)
{

    /* just dispatching */
    switch (ast->u.child [1]->ast_type) {
    case AST_assignment_operator_eq:     eval_exp_assign_eq   (ast); break;
    case AST_assignment_operator_add:    eval_exp_assign_add  (ast); break;
    case AST_assignment_operator_sub:    eval_exp_assign_sub  (ast); break;
    case AST_assignment_operator_mult:   eval_exp_assign_mult (ast); break;
    case AST_assignment_operator_div:    eval_exp_assign_div  (ast); break;
    case AST_assignment_operator_mod:    eval_exp_assign_mod  (ast); break;
    case AST_assignment_operator_lshift:
    case AST_assignment_operator_rshift: eval_exp_assign_sh   (ast); break;
    case AST_assignment_operator_band:
    case AST_assignment_operator_beor:
    case AST_assignment_operator_bior:   eval_exp_assign_bit  (ast); break;
    default:                             assert (0);                 break;
    }
}

static void
eval_exp_assign_eq   (struct AST *ast)
{
    struct Type   *type00 = ast->u.child [0]->type;
    struct Type   *type2  = get_type (ast->u.child [2]);

    eval_exp (ast->u.child [2]);
    eval_promote_type (type2, type00);
    eval_lvalue (ast->u.child [0]);
    vm.top_b (vm.pop_p (), type00->size);
}

static void
eval_exp_assign_add  (struct AST *ast)
{
    struct Type   *type00 = ast->u.child [0]->type;
    struct Type   *type0  = get_type (ast->u.child [0]);
    struct Type   *type2  = get_type (ast->u.child [2]);
    void          *addr;

    eval_lvalue (ast->u.child [0]);      /* push address of left operand */
    addr = vm.pop_p ();                  /* set it aside */
    exp_push_addr (addr, type00);        /* push (*addr) */
    eval_promote_ast (ast->u.child [0]); /* cast it */

    if (type0->kind == TYPE_KIND_POINTER) {
	int size = type0->u.t_pointer.type->size;
	assert (is_integral_type (type2));
	vm.p_to_i ();
	eval_exp (ast->u.child [2]);
	vm.push_i (size);
	vm.mult_i (); vm.add_i ();
	vm.i_to_p ();
    } else if (type2->kind == TYPE_KIND_POINTER) {
	int size = type2->u.t_pointer.type->size;
	assert (is_integral_type (type0));
	vm.push_i (size);
	vm.mult_i ();
	eval_exp (ast->u.child [2]);
	vm.p_to_i ();
	vm.add_i ();
	vm.i_to_p ();
    } else {
	assert (type0->kind == TYPE_KIND_PRIM); 
	assert (type2->kind == TYPE_KIND_PRIM);
	eval_exp (ast->u.child [2]);
	if (type0->u.t_prim.ptype == PRIM_TYPE_INT) {
	    assert (type2->u.t_prim.ptype == PRIM_TYPE_INT);
	    vm.add_i ();
	} else {
	    assert (((type0->u.t_prim.ptype == PRIM_TYPE_FLOAT)
		     && (type2->u.t_prim.ptype == PRIM_TYPE_FLOAT))
		    || ((type0->u.t_prim.ptype == PRIM_TYPE_DOUBLE)
			&& (type2->u.t_prim.ptype == PRIM_TYPE_DOUBLE)));
	    vm.add_d ();
	}
    }

    eval_promote_type (type2, type00);	 /* cast it to the left type */
    vm.top_b (addr, type00->size);       /* store it into the variable */
}

static void
eval_exp_assign_sub  (struct AST *ast)
{
    struct Type   *type00 = ast->u.child [0]->type;
    struct Type   *type0  = get_type (ast->u.child [0]);
    struct Type   *type2  = get_type (ast->u.child [2]);
    void          *addr;

    eval_lvalue (ast->u.child [0]);      /* push address of left operand */
    addr = vm.pop_p ();                  /* set it aside */
    exp_push_addr (addr, type00);        /* push (*addr) */
    eval_promote_ast (ast->u.child [0]); /* cast it */

    if (type0->kind == TYPE_KIND_POINTER) {
	int size = type0->u.t_pointer.type->size;
	vm.p_to_i ();
	eval_exp (ast->u.child [2]);
	if (type2->kind == TYPE_KIND_POINTER) {
	    vm.p_to_i ();
	    vm.sub_i ();
	    vm.push_i (size);
	    vm.div_i ();
	} else {
	    assert (is_integral_type (type2));
	    vm.push_i (size);
	    vm.mult_i ();
	    vm.sub_i ();
	    vm.i_to_p ();
	}
    } else {
	assert (type0->kind == TYPE_KIND_PRIM); 
	assert (type2->kind == TYPE_KIND_PRIM);
	eval_exp (ast->u.child [2]);
	if (type0->u.t_prim.ptype == PRIM_TYPE_INT) {
	    assert (type2->u.t_prim.ptype == PRIM_TYPE_INT);
	    vm.sub_i ();
	} else {
	    assert (((type0->u.t_prim.ptype == PRIM_TYPE_FLOAT)
		     && (type2->u.t_prim.ptype == PRIM_TYPE_FLOAT))
		    || ((type0->u.t_prim.ptype == PRIM_TYPE_DOUBLE)
			&& (type2->u.t_prim.ptype == PRIM_TYPE_DOUBLE)));
	    vm.sub_d ();
	}
    }

    eval_promote_type (type2, type00);	 /* cast it to the left type */
    vm.top_b (addr, type00->size);        /* store it into the variable */
}

static void
eval_exp_assign_mult (struct AST *ast)
{
    struct Type   *type00 = ast->u.child [0]->type;
    struct Type   *type0  = get_type (ast->u.child [0]);
    struct Type   *type2  = get_type (ast->u.child [2]);
    void          *addr;

    eval_lvalue (ast->u.child [0]);      /* push address of left operand */
    addr = vm.pop_p ();                  /* set it aside */
    exp_push_addr (addr, type00);        /* push (*addr) */
    eval_promote_ast (ast->u.child [0]); /* cast it */

    assert (type0->kind == TYPE_KIND_PRIM);
    assert (type2->kind == TYPE_KIND_PRIM);

    eval_exp (ast->u.child [2]);
    if (type0->u.t_prim.ptype == PRIM_TYPE_INT) {
	assert (type2->u.t_prim.ptype == PRIM_TYPE_INT);
	vm.mult_i ();
    } else {
	assert (((type0->u.t_prim.ptype == PRIM_TYPE_FLOAT)
		 && (type2->u.t_prim.ptype == PRIM_TYPE_FLOAT))
		|| ((type0->u.t_prim.ptype == PRIM_TYPE_DOUBLE)
		    && (type2->u.t_prim.ptype == PRIM_TYPE_DOUBLE)));
	vm.mult_d ();
    }

    eval_promote_type (type2, type00);	 /* cast it to the left type */
    vm.top_b (addr, type00->size);       /* store it into the variable */
}

static void
eval_exp_assign_div  (struct AST *ast)
{
    struct Type   *type00 = ast->u.child [0]->type;
    struct Type   *type0  = get_type (ast->u.child [0]);
    struct Type   *type2  = get_type (ast->u.child [2]);
    void          *addr;

    eval_lvalue (ast->u.child [0]);      /* push address of left operand */
    addr = vm.pop_p ();                  /* set it aside */
    exp_push_addr (addr, type00);        /* push (*addr) */
    eval_promote_ast (ast->u.child [0]); /* cast it */

    assert (type0->kind == TYPE_KIND_PRIM);
    assert (type2->kind == TYPE_KIND_PRIM);

    eval_exp (ast->u.child [2]);
    if (type0->u.t_prim.ptype == PRIM_TYPE_INT) {
	assert (type2->u.t_prim.ptype == PRIM_TYPE_INT);
	if (vm.top_i () == 0) {
	    xci_error (ast->u.child [2], "divided by 0");
	}
	vm.div_i ();
    } else {
	assert (((type0->u.t_prim.ptype == PRIM_TYPE_FLOAT) 
		 && (type2->u.t_prim.ptype == PRIM_TYPE_FLOAT))
		|| ((type0->u.t_prim.ptype == PRIM_TYPE_DOUBLE)
		    && (type2->u.t_prim.ptype == PRIM_TYPE_DOUBLE)));
	if (vm.top_d () == 0.0) {
	    xci_error (ast->u.child [2], "divided by 0");
	}
	vm.div_d ();
    }

    eval_promote_type (type2, type00);	 /* cast it to the left type */
    vm.top_b (addr, type00->size);       /* store it into the variable */
}

static void
eval_exp_assign_mod  (struct AST *ast)
{
    struct Type   *type00 = ast->u.child [0]->type;
    struct Type   *type0  = get_type (ast->u.child [0]);
    struct Type   *type2  = get_type (ast->u.child [2]);
    void          *addr;

    eval_lvalue (ast->u.child [0]);      /* push address of left operand */
    addr = vm.pop_p ();                  /* set it aside */
    exp_push_addr (addr, type00);        /* push (*addr) */
    eval_promote_ast (ast->u.child [0]); /* cast it */

    assert (type0->kind == TYPE_KIND_PRIM);
    assert (type2->kind == TYPE_KIND_PRIM);
    assert (type0->u.t_prim.ptype == PRIM_TYPE_INT); 
    assert (type2->u.t_prim.ptype == PRIM_TYPE_INT);

    eval_exp (ast->u.child [2]);
    if (vm.top_i () == 0) {
	xci_error (ast->u.child [2], "divided by 0");
    }
    vm.mod_i ();

    eval_promote_type (type2, type00);	 /* cast it to the left type */
    vm.top_b (addr, type00->size);       /* store it into the variable */
}

static void
eval_exp_assign_sh   (struct AST *ast)
{
    struct Type   *type00 = ast->u.child [0]->type;
    struct Type   *type0  = get_type (ast->u.child [0]);
    struct Type   *type2  = get_type (ast->u.child [2]);
    void          *addr;

    eval_lvalue (ast->u.child [0]);      /* push address of left operand */
    addr = vm.pop_p ();                  /* set it aside */
    exp_push_addr (addr, type00);        /* push (*addr) */
    eval_promote_ast (ast->u.child [0]); /* cast it */

    assert (type0->kind == TYPE_KIND_PRIM);
    assert (type2->kind == TYPE_KIND_PRIM);
    assert (type0->u.t_prim.ptype == PRIM_TYPE_INT); 
    assert (type2->u.t_prim.ptype == PRIM_TYPE_INT);

    eval_exp (ast->u.child [2]);
    switch (ast->u.child [1]->ast_type) {
    case AST_assignment_operator_lshift:
	vm.l_sh ();
	break;
    case AST_assignment_operator_rshift:
	if (type0->u.t_prim.modifier & TYPE_MOD_UNSIGNED) {
	    vm.r_ush ();
	} else {
	    vm.r_sh ();
	}
	break;
    default:
	assert (0);
	break;
    }

    eval_promote_type (type2, type00);	 /* cast it to the left type */
    vm.top_b (addr, type00->size);       /* store it into the variable */
}

static void
eval_exp_assign_bit  (struct AST *ast)
{
    struct Type   *type00 = ast->u.child [0]->type;
    struct Type   *type0  = get_type (ast->u.child [0]);
    struct Type   *type2  = get_type (ast->u.child [2]);
    void          *addr;

    eval_lvalue (ast->u.child [0]);      /* push address of left operand */
    addr = vm.pop_p ();                  /* set it aside */
    exp_push_addr (addr, type00);        /* push (*addr) */
    eval_promote_ast (ast->u.child [0]); /* cast it */

    assert (type0->kind == TYPE_KIND_PRIM);
    assert (type2->kind == TYPE_KIND_PRIM);
    assert (type0->u.t_prim.ptype == PRIM_TYPE_INT); 
    assert (type2->u.t_prim.ptype == PRIM_TYPE_INT);

    eval_exp (ast->u.child [2]);
    switch (ast->u.child [1]->ast_type) {
    case AST_assignment_operator_band:  vm.b_and (); break;
    case AST_assignment_operator_beor:  vm.b_xor (); break;
    case AST_assignment_operator_bior:  vm.b_or ();  break;
    default:                            assert (0);  break;
    }

    eval_promote_type (type2, type00);	 /* cast it to the left type */
    vm.top_b (addr, type00->size);       /* store it into the variable */
}

static void
eval_exp_funcall  (struct AST *ast)
{
    void        *addr;
    struct Type *type;
    void        *old_sp;

    assert ((ast->ast_type == AST_postfix_expression_funcall)
	    || (ast->ast_type == AST_postfix_expression_funcall_noarg));

    eval_exp (ast->u.child [0]);
    addr = vm.pop_p ();	 /* function pointer */

    if (addr == NULL) {
	xci_error (ast, "NULL function called");
    }

    type = get_type (ast->u.child [0]);
    assert (type->kind == TYPE_KIND_POINTER);
    type = type->u.t_pointer.type;
    assert (type->kind == TYPE_KIND_FUNCTION);

    /* push arguments on the stack */
    if (is_builtin (addr)) {
	if (ast->ast_type == AST_postfix_expression_funcall) {
	    push_arguments_for_builtin (ast->u.child [1], type);
	}
	((void (*)(void))addr) ();     /* call builtin function */
    } else {
	struct Type *ret_type = type->u.t_function.ret_type;
	int    ret_size = 0;
	old_sp = vm.get_sp ();

	if (ast->ast_type == AST_postfix_expression_funcall) {
	    push_arguments_for_normal (ast->u.child [1], type);
	}
	if (!is_void_type (ret_type)) {
	    vm.add_sp (ret_type->size); /* skip space to return value */
	}
	vm.push_p ((void *)vm.pc.ast); /* push old context */
	eval_block (addr);             /* call function */

	/* copy the return value, if the function is non-void */
	if (!is_void_type (ret_type)) {
	    ret_size = get_type_size (ret_type);
	    vm.top_b (old_sp, ret_size);
	}

	/* discard arguments */
	vm.set_sp (old_sp + ret_size); /* pop the arguments away */
    }
}

static void
eval_lvalue_id     (struct AST *ast)
{
    struct Symbol *sym;

    sym = sym_lookup (ast->u.string, NS_NORMAL);
    assert (sym != NULL);
    vm.push_p (sym->addr);
}

static void
eval_lvalue_deref (struct AST *ast)
{
    eval_exp (ast->u.child [1]);
}

static void
eval_lvalue_array (struct AST *ast)
{
    int           size;
    struct Type   *type0 = get_type (ast->u.child [0]);
    struct Type   *type1 = get_type (ast->u.child [1]);

    if (type0->kind == TYPE_KIND_POINTER) {
	size = get_type_size (type0->u.t_pointer.type);
	eval_exp (ast->u.child [0]);
	if (vm.top_p () == 0) {
	    xci_error (ast, "pointer to array is NULL");
	}
	vm.p_to_i ();
	eval_exp (ast->u.child [1]);
    } else {
	assert (type1->kind == TYPE_KIND_POINTER);
	size = get_type_size (type1->u.t_pointer.type);
	eval_exp (ast->u.child [1]);
	if (vm.top_p () == 0) {
	    xci_error (ast, "pointer to array is NULL");
	}
	vm.p_to_i ();
	eval_exp (ast->u.child [0]);
    }
    vm.push_i (size);
    vm.mult_i ();
    vm.add_i ();
    vm.i_to_p ();
}

static void
eval_lvalue_dot   (struct AST *ast)
{
    struct StructUnionField *field;
    struct Type   *type0 = get_type (ast->u.child [0]);

    assert ((type0->kind == TYPE_KIND_STRUCT)
	    || (type0->kind == TYPE_KIND_UNION));

    eval_lvalue (ast->u.child [0]);
    vm.p_to_i ();

    field = lookup_field (type0, ast->u.child [1]->u.string);
    if (field == NULL) {
	xci_error_args (ast->u.child [1],
			"no such member in structure or union", 
			ast->u.child [1]->u.string);
    }
    vm.push_i (field->offset);
    vm.add_i ();
    vm.i_to_p ();
}

static void
eval_lvalue_arrow (struct AST *ast)
{
    struct StructUnionField *field;
    struct Type   *type0 = get_type (ast->u.child [0]);
    struct Type   *type1;

    assert (type0->kind == TYPE_KIND_POINTER);
    type1 = type0->u.t_pointer.type;
    assert ((type1->kind == TYPE_KIND_STRUCT)
	    || (type1->kind == TYPE_KIND_UNION));

    eval_exp (ast->u.child [0]);
    vm.p_to_i ();

    field = lookup_field (type1, ast->u.child [1]->u.string);
    if (field == NULL) {
	xci_error_args (ast->u.child [1],
			"no such member in structure or union", 
			ast->u.child [1]->u.string);
    }
    vm.push_i (field->offset);
    vm.add_i ();
    vm.i_to_p ();
}

static struct StructUnionField *
lookup_field (struct Type *type, char *field_name)
{
    int    i, field_num;
    struct StructUnionField **field;

    assert ((type->kind == TYPE_KIND_STRUCT)
	    || (type->kind == TYPE_KIND_STRUCT));

    field_num = type->u.t_struct_union.num;
    field     = type->u.t_struct_union.field;

    for (i = 0; i < field_num; i++) {
	if (!strcmp (field [i]->name, field_name)) {
	    return field [i];
	}
    }
    return NULL;
}

static void
exp_push_addr (void *addr, struct Type *type)
{
    if (addr == NULL) {
	xci_error_noast ("NULL dereferenced");
    }
    switch (type->kind) {
    case TYPE_KIND_POINTER:    vm.push_p (*(void **) addr);  break;
    case TYPE_KIND_PRIM:
	switch (type->u.t_prim.ptype) {
	case PRIM_TYPE_CHAR:   vm.push_c (*(char *) addr);   break;
	case PRIM_TYPE_INT:    vm.push_i (*(int *) addr);    break;
	case PRIM_TYPE_FLOAT:  /* falling through */
	case PRIM_TYPE_DOUBLE: vm.push_d (*(double *) addr); break;
	case PRIM_TYPE_VOID:   /* falling through */
	default:               assert (0);                   break;
	}
	break;
    case TYPE_KIND_STRUCT:     /* falling through */
    case TYPE_KIND_UNION:      vm.push_b (addr, type->size); break;
    case TYPE_KIND_ENUM:       /* falling through */
    default:                   assert (0);                   break;
    }
}

static struct AST *
eval_initializer_list (struct AST *ast, struct Type *type, void *addr)
{
    struct StructUnionField **field;
    struct Type *member_type;
    int num_field;
    int i;

    assert ((ast->ast_type == AST_initializer_list_single)
	    || (ast->ast_type == AST_initializer_list_pair));

    if (type->kind == TYPE_KIND_QUAL) {
	type = type->u.t_qual.type;
    }

    if (type->kind == TYPE_KIND_STRUCT) {
	num_field = type->u.t_struct_union.num;
	field = type->u.t_struct_union.field;
    } else if (type->kind == TYPE_KIND_ARRAY) {
	num_field = type->u.t_array.num;
	assert (num_field >= 0);
	member_type = type->u.t_array.elem_type;
    } else {
	assert (0);
    }

    for (i = 0;; i++) {
	struct AST *ast_init;
	void   *member_addr;

	if ((ast->ast_type != AST_initializer_list_single)
	    && (ast->ast_type != AST_initializer_list_pair)) {
	    /* no more initializer */
	    return ast;
	}

	switch (ast->ast_type) {
	case AST_initializer_list_single: ast_init = ast->u.child [0]; break;
	case AST_initializer_list_pair:   ast_init = ast->u.child [1]; break;
	default:                          assert (0);                  break;
	}

	if (type->kind == TYPE_KIND_STRUCT) {
	    while ((i < num_field) && (field [i]->name == NULL)) {
		i++; /* skip unnamed bit-field members */
	    }
    	}
	if (i >= num_field) {
	    /* no more member to be initialized in this (sub) aggregate */
	    return ast;
	}

	if (type->kind == TYPE_KIND_STRUCT) {
	    member_type = field [i]->type;
	    member_addr = addr + field [i]->offset;
	} else if (type->kind == TYPE_KIND_ARRAY) {
	    member_addr = addr + i * get_type_size (member_type);
	} else {
	    assert (0);
	}

	if (member_type->kind == TYPE_KIND_QUAL) {
	    member_type = member_type->u.t_qual.type;
	}

	if (((member_type->kind == TYPE_KIND_STRUCT)
	     || (member_type->kind == TYPE_KIND_ARRAY))
	    && (ast_init->ast_type == AST_initializer_assign)) {
	    /* in the case that braces are omitted for subaggregate */
	    ast = eval_initializer_list (ast, member_type, member_addr);
	    /* ast = ast->parent is already done in the above function,
	       so we just do continue here. */
	    continue;
	} else {
	    eval_initializer (ast_init, member_type, member_addr);
    	}

	ast = ast->parent;
    }
}

static void
push_arguments_for_builtin (struct AST *ast, struct Type *type)
{
    int    i;
    void   *sp;
    struct AST *ast_bottom, *ast_ellipsis;

    assert ((ast->ast_type == AST_argument_expression_list_single)
	    || (ast->ast_type == AST_argument_expression_list_pair));
    assert (type->kind == TYPE_KIND_FUNCTION);

    /* first search the bottom */
    for (ast_bottom = ast;
	 ast_bottom->ast_type != AST_argument_expression_list_single;
	 ast_bottom = ast_bottom->u.child [0])
	; /* empty body */

    /* pushing arguments in order. if the ellipsis parameters exist,
       the order is changed as follows:
       (1) first push the ellipsis arguments in order.
       (2) second push the fixed arguments in order.
       (3) finally push the first address of the ellipsis arguments;
       	   this will be used as the parameter typed with va_list. */
    if (type->u.t_function.is_ellipsis) {
	sp = vm.get_sp ();
	ast_ellipsis = ast_bottom;

    	/* skip the fixed arguments */
	for (i = 0; i < type->u.t_function.arg_num; i++) {
	    ast_ellipsis = ast_ellipsis->parent;
    	}

	for (; ast_ellipsis != ast->parent;
	       ast_ellipsis = ast_ellipsis->parent) {
	    switch (ast_ellipsis->ast_type) {
	    case AST_argument_expression_list_single:
		eval_exp (ast_ellipsis->u.child [0]);
		break;
	    case AST_argument_expression_list_pair:
		eval_exp (ast_ellipsis->u.child [1]);
		break;
	    default:
		assert (0);
		break;
	    }
    	}
    }

    for (i = 0; i < type->u.t_function.arg_num; i++) {
	switch (ast_bottom->ast_type) {
	case AST_argument_expression_list_single:
	    eval_exp (ast_bottom->u.child [0]);
	    break;
	case AST_argument_expression_list_pair:
	    eval_exp (ast_bottom->u.child [1]);
	    break;
	default:
	    assert (0);
	    break;
	}
	ast_bottom = ast_bottom->parent;
    }

    if (type->u.t_function.is_ellipsis) {
	vm.push_p (sp);
    }
}

static void
push_arguments_for_normal  (struct AST *ast, struct Type *type)
{
    int    i;
    struct AST *ast_bottom, *ast_ellipsis;

    /* first search the bottom */
    for (ast_bottom = ast;
	 ast_bottom->ast_type != AST_argument_expression_list_single;
	 ast_bottom = ast_bottom->u.child [0])
	; /* empty body */

    assert ((ast->ast_type == AST_argument_expression_list_single)
	    || (ast->ast_type == AST_argument_expression_list_pair));
    assert (type->kind == TYPE_KIND_FUNCTION);

    /* push ellipsis arguments in reverse order */
    if (type->u.t_function.is_ellipsis) {
	ast_ellipsis = ast_bottom;

    	/* skip the fixed arguments */
	for (i = 0; i < type->u.t_function.arg_num - 1; i++) {
	    ast_ellipsis = ast_ellipsis->parent;
    	}

	for (; ast != ast_ellipsis; ast = ast->u.child [0]) {
	    switch (ast->ast_type) {
	    case AST_argument_expression_list_single:
		eval_exp (ast->u.child [0]);
		break;
	    case AST_argument_expression_list_pair:
		eval_exp (ast->u.child [1]);
		break;
	    default:
		assert (0);
		break;
	    }
    	}
    }

    /* push fixed arguments in reverse order */
    for (i = 0; i < type->u.t_function.arg_num; i++) {
	switch (ast->ast_type) {
	case AST_argument_expression_list_single:
	    eval_exp (ast->u.child [0]);
	    break;
	case AST_argument_expression_list_pair:
	    eval_exp (ast->u.child [1]);
	    break;
	default:
	    assert (0);
	    break;
	}
	ast = ast->u.child [0];
    }
    assert (ast->parent->ast_type == AST_argument_expression_list_single);
}
/* ---------------------------------------------------------------------- */
