/*
    XCI: eXperimental C Interpreter for ANSI C
    -- Copyright (c) 2000-2001, gondow@jaist.ac.jp
    $Id: symbol.c,v 1.3 2002/06/09 16:17:14 gondow Exp $
*/ 

#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <string.h>
#include "AST.h"
#include "xci.h"
#include "type.h"
#include "symbol.h"
#include "vm.h"
#include "eval_exp.h"
#include "attr.h"

/* ---------------------------------------------------------------------- */
struct SymTable sym_table = {NULL, NULL, NULL, -1, 0, NULL};
/* ---------------------------------------------------------------------- */
static struct Symbol *sym_copy (struct Symbol *sym);
static void   sym_free (struct Symbol *sym);
static void   symstack_free (struct SymStack *symstack);

static struct Symbol *sym_redeclaration_check (struct Symbol *sym);
static struct Symbol *sym_lookup_at_file (char const * const name,
					  const enum NameSpace ns);
static struct Symbol *sym_lookup_at_block (char const * const name,
					   const enum NameSpace ns);
static struct Symbol *sym_lookup_at_innermost_block (char const * const name,
						     const enum NameSpace ns);

static void   sym_dump       (struct Symbol *sym, int is_verbose);
static char   *namelevel2str (enum NameLevel namelevel);
static char   *namespace2str (enum NameSpace namespace);
static char   *symstat2str   (int stat);
#if 0
static char   *find_declarator_id (struct AST *ast);
#endif
static void sym_entry_label  (struct AST const *ast_label);
/* ---------------------------------------------------------------------- */
struct Symbol *
sym_lookup (char const * const name, const enum NameSpace ns)
{
    struct Symbol *sym;

    sym = sym_lookup_at_block (name, ns);

    if (sym == NULL) {
	sym = sym_lookup_at_file (name, ns);
    }

    /* link the symbol with NULL address if VM is running */
    if ((vm.status == VM_STAT_RUNNING) && (sym != NULL)
	&& (sym->addr == NULL) && (sym->status & SYM_STAT_GLOBAL)) {
	struct Symbol *sym2 = sym_lookup_at_all_files (name, ns);
	if (sym2 != NULL)
	    sym->addr = sym2->addr;
    }
    return sym;
}

struct Symbol *
sym_lookup_at (char const * const name, enum NameSpace const ns,
	       enum NameLevel const nl)
{
    switch (nl) {
    case NL_FILE:      return sym_lookup_at_file     (name, ns);
    case NL_BLOCK:     return sym_lookup_at_block    (name, ns);
    default:           assert (0); return NULL;
    }
}

/*
  When handling external declaration, we need to find its definition
  in other files or in the same file, that is, link the name.
  This function is used to do this.
  
  Note: we assume that the symbol entry the declaration 1 (see below)
  be overriden by the definition 2. This is legal if outside of functions.
      extern int x; // 1
      static int x; // 2
 */

struct Symbol *
sym_lookup_at_all_files (char const * const name, const enum NameSpace ns)
{
    struct Symbol  *sym, *result = NULL;
    struct SymList *symlist = sym_table.file;

    /* search all files for name */
    for (symlist = sym_table.file; symlist != NULL; symlist = symlist->next) {
	for (sym = symlist->symbol; sym != NULL; sym = sym->next) {
	    if (!strcmp (sym->name, name)
		&& sym->namespace == ns
		&& sym->namelevel == NL_FILE
		&& (sym->status & SYM_STAT_DEFINITION)
		&& (sym->status & (SYM_STAT_GLOBAL))) {
		if (result != NULL) {
		    xci_error_args (sym->ast,
				    "duplicate definitions: %s\n", name);
		}
		result = sym;
	    }
	}
    }

    /* NULL is returned when not found */
    return result;
}

static struct Symbol *
sym_lookup_at_file (char const * const name, const enum NameSpace ns)
{
    struct Symbol *sym;

    for (sym = sym_table.curr_file->symbol; sym != NULL; sym = sym->next) {
	if (!strcmp (sym->name, name) && sym->namespace == ns) {
	    return sym;
	}
    }
    return NULL;
}

static struct Symbol *
sym_lookup_at_block (char const * const name, const enum NameSpace ns)
{
    struct SymStack *symstack;
    struct Symbol   *sym;
    int             i;

    if (sym_table.func_sp < 0) return NULL;

    symstack = sym_table.func_stack [sym_table.func_sp];

    for (i = symstack->block_sp; i >= 0; i--) {
	for (sym = symstack->block_stack [i]; sym != NULL; sym = sym->next) {
	    if (!strcmp (sym->name, name) && sym->namespace == ns) {
		return sym;
	    }
	}
    }
    return NULL;
}

static struct Symbol *
sym_lookup_at_innermost_block (char const * const name,
			       const enum NameSpace ns)
{
    struct SymStack *symstack;
    struct Symbol   *sym;
    int             i;

    symstack = sym_table.func_stack [sym_table.func_sp];
    i = symstack->block_sp;
    for (sym = symstack->block_stack [i]; sym != NULL; sym = sym->next) {
	if (!strcmp (sym->name, name) && sym->namespace == ns) {
	    return sym;
	}
    }
    return NULL;
}

struct Symbol *
sym_redeclaration_check (struct Symbol *sym)
{
    struct Symbol   *sym2;

    assert (sym->namespace != NS_LABEL);
    assert (sym->namespace != NS_MEMBER);

    /* check if already entried or not */
    if (sym->namelevel == NL_BLOCK) {
	sym2 = sym_lookup_at_innermost_block (sym->name, sym->namespace);
    } else {
	sym2 = sym_lookup_at (sym->name, sym->namespace, sym->namelevel);
    }

    /* sym is new entry */
    if (sym2 == NULL) return NULL;

    /* sym is redefinition */
    if ((sym->status & SYM_STAT_DEFINITION)
	&& (sym2->status & SYM_STAT_DEFINITION)) {
	xci_error_args (sym->ast, "redefinition of %s", sym->name);
    }

    /* type is different than declared before */
    if (!is_type_compatible (sym->type, sym2->type)
	|| !is_sym_stat_compatible (sym->status, sym2->status)) {
	xci_error_args (sym->ast,
			"conflicting type definition of '%s'", sym->name);
    }

    /* sym needs to be completed by overriding the previous entry */
    if ((sym->status & SYM_STAT_DEFINITION)
	|| ((sym->status & SYM_STAT_TENTATIVE)
	    && (sym2->status & SYM_STAT_DECLARATION))) {

	/* completing incomplete type in the old symbol */
	*sym2->type    = *sym->type;

	/* copying (overriding) symbol entries */
	sym2->addr     = sym->addr;
	sym2->filename = sym->filename;
	sym2->status   = sym->status;
	sym2->ast      = sym->ast;
	assert (sym2->namespace == sym->namespace);
	assert (sym2->namelevel == sym->namelevel);
#if 0
	/* some types are reused, and there is no way to easily
	   identify them under the current implementation.
	   That is why I commented out this line. */
	efree (sym->type);
#endif
	efree (sym);
    }
	
    return sym2;
}

struct Symbol *
sym_entry (struct Symbol * const sym)
{
    struct SymStack *symstack;
    struct Symbol   *prev_sym;
    
    assert (sym != NULL);

#if 1 /* for debug */
    if (!strcmp (sym->name, "symdump")) {
	symtable_dump (0);
    } else if (!strcmp (sym->name, "symdump_verbose")) {
	symtable_dump (1);
    }
#endif

    prev_sym = sym_redeclaration_check (sym);

    /* prev_sym is declared before as tentative, incomplete or declaration */
    if (prev_sym != NULL) {
	return prev_sym;
    }

    switch (sym->namelevel) {
    case NL_FILE:
	sym->next = sym_table.curr_file->symbol;
	sym_table.curr_file->symbol = sym;
	break;
    case NL_BLOCK:
	symstack = sym_table.func_stack [sym_table.func_sp];
	sym->next = symstack->block_stack [symstack->block_sp];
	symstack->block_stack [symstack->block_sp] = sym;
	break;
    default:
	assert (0);
	break;
    }
    return sym;
}

/*
  In type.c, only sym_entry_byname is used to entry symbol names.
  This function also allocate static objects.
 */

struct Symbol *
sym_entry_byname (char const        *const name,
		  struct AST        *ast,    /* used to represent context */
		  struct Type       *type,
		  enum StorageClass storage,
		  enum NameSpace    ns)
{
    int            sym_stat;
    enum NameLevel nl;
    void           *addr;
    struct Symbol  *sym;

    if (ns == NS_TAG) {
	assert (type->kind != TYPE_KIND_QUAL);
    }
    sym_stat = create_sym_stat (ast, type, name, storage);

    switch (decl_context) {
    case DECL_CONTEXT_GLOBAL: nl = NL_FILE;  break;
    case DECL_CONTEXT_LOCAL:  nl = NL_BLOCK; break;
    default:                  assert (0);    break;
    }

    if (sym_stat & SYM_STAT_ARGUMENT) {
	nl = NL_BLOCK;
    }

    switch (ast->ast_type) {
    case AST_function_definition_spec:
    case AST_function_definition_nospec:
	addr = ast;
	ast->type = type;
	break;

    case AST_struct_or_union_specifier_tag:
    case AST_struct_or_union_specifier_incomplete:
    case AST_enum_specifier_tag:
    case AST_parameter_declaration_dec:
    case AST_declaration_list_single:
    case AST_declaration_list_pair:
    case AST_declarator_direct:
    case AST_IDENTIFIER:
	addr = NULL;
	break;

    case AST_enumerator_id:
    case AST_enumerator_id_init:
	addr = emalloc (vm.VM_SIZEOF_INT);
	break;

    case AST_init_declarator_dec:
    case AST_init_declarator_dec_init:
	switch (decl_context) {
	case DECL_CONTEXT_GLOBAL:
	    if ((sym_stat & (SYM_STAT_DEFINITION | SYM_STAT_TENTATIVE))
		&& !(sym_stat & SYM_STAT_INCOMPLETE)) {
		addr = ecalloc (1, get_type_size (type));
	    } else {
		addr = NULL; /* including function prototype declaration */
	    }
	    break;

	case DECL_CONTEXT_LOCAL:
	    if ((sym_stat & SYM_STAT_STATIC) 
		&& !(sym_stat & SYM_STAT_INCOMPLETE)) {
		addr = ecalloc (1, get_type_size (type));
	    } else {
		/* auto variables are allocated dynamically by interpreter */
		addr = NULL;
	    }
	    break;

	default:
	    assert (0);
	    break;
	}
	break;

    default:
	assert (0);
	break;
    }

    sym = sym_create (name, ast, type, addr, filename, ns, nl, sym_stat);
    sym = sym_entry (sym);
    return sym;
}

struct Symbol *
sym_create (char const * const name,
	    struct AST const   *ast,
	    struct Type        *type,
	    void    	       *addr,
	    char    	       *filename,
	    enum NameSpace     namespace,
	    enum NameLevel     namelevel,
	    int                sym_stat)
{
    struct Symbol *sym = emalloc (sizeof (struct Symbol));

    sym->name      = name;
    sym->ast       = (struct AST *) ast;
    sym->type      = type;
    sym->addr      = addr;
    sym->filename  = filename;
    sym->status    = sym_stat;
    sym->namespace = namespace ;
    sym->namelevel = namelevel;
    sym->next      = NULL;

    return sym;
}

/*
  Register all symbols in the current (innermost) block
  into 'sym_table.static_block' with the key 'ast'.
  This is done *before* the code runs; when it runs and enters into
  the block, the symbols will be copied into sym_table's block_stack:
  
    	sym_table.func_stack [sym_table.func_sp]
	->block_stack [symstack->block_sp]

  and they will be free-ed when exiting.

  Then, register the total size for local variables declared in the block.
 */
void
sym_entry_static_block (struct AST *ast)
{
    struct StaticBlock *static_block;
    struct SymStack    *symstack;
    struct Symbol      *sym;
    int                offset;

    /* the key 'ast' must be compound_statement. */
    assert ((ast->ast_type == AST_compound_statement_decn)
	    || (ast->ast_type == AST_compound_statement_nodecn));

    /* get the current symbols */
    symstack = sym_table.func_stack [sym_table.func_sp];
    sym      = symstack->block_stack [symstack->block_sp];

    static_block = emalloc (sizeof (struct StaticBlock));
    static_block->key    = ast;

    /* reverse the entries */
    {
	struct Symbol *prev = NULL, *next;
	while (sym != NULL) {
	    next      = sym->next;
	    sym->next = prev;
	    prev      = sym;
	    sym       = next;
    	}
	static_block->symbol = prev;
    }

    /* NULL is assigned here to prevent sym_end_block () from
       freeing these symbols. Due to this side effect, this function
       is supposed to be called just before exiting the block. */
    symstack->block_stack [symstack->block_sp] = NULL;


    /* insert the static_block to the top of sym_table.static_block. */
    static_block->next = sym_table.static_block;
    sym_table.static_block = static_block;

    offset = 0;
    for (sym = static_block->symbol; sym != NULL; sym = sym->next) {
	if (sym->status & SYM_STAT_ARGUMENT)
	    continue;
	offset += get_type_size (sym->type);
    }
    set_block_sp_offset (ast, offset);
}

void
sym_copy_static_block (struct AST const * const ast)
{
    struct StaticBlock *p, *static_block;
    struct SymStack    *symstack;
    struct Symbol      *symbol;

    assert ((ast->ast_type == AST_compound_statement_decn)
	    || (ast->ast_type == AST_compound_statement_nodecn));

    /* get the current symbols */
    symstack = sym_table.func_stack [sym_table.func_sp];
    symbol   = symstack->block_stack [symstack->block_sp];

    /* this function is supposed to be called just after entering the block */
    assert (symbol == NULL);

    /* search sym_table.local_static first, which should succeed.
       (you think we should use hash here? yeah, you are right.) */
    for (p = sym_table.static_block; p != NULL; p = p->next) {
	if (p->key == ast) {
	    static_block = p;
	    break;
	}
    }

    assert (p != NULL); /* searching must not be failed. */

    /* copy symbols in static_block and store them as the current symbols */
    symbol =  sym_copy (static_block->symbol);
    symstack->block_stack [symstack->block_sp] = symbol;
}

/* this function is statically used, that is, in compile-time. */
void
sym_begin_file (char const * const filename)
{
    struct SymList *symlist;

    /* search filename first */
    for (symlist = sym_table.file; symlist != NULL; symlist = symlist->next) {
	if (!strcmp(filename, symlist->filename))
	    break;
    }

    /* if not found, create new entry for filename */
    if (symlist == NULL) {
	symlist = emalloc (sizeof (struct SymList));
	symlist->filename = filename;
	symlist->symbol = NULL;
	symlist->next = sym_table.file;
	sym_table.file = symlist;
    }

    sym_table.curr_file = symlist;
}

void
sym_end_file (void)
{
    /* do nothing */
}

void
sym_begin_function (void)
{
    struct SymStack *symstack;

    assert (sym_table.func_sp < sym_table.func_size);

    if (sym_table.func_sp >= sym_table.func_size - 1) {
	sym_table.func_stack = 
	    erealloc (sym_table.func_stack,
		      sizeof (struct SymStack *) * ++sym_table.func_size);
    }

    symstack = emalloc (sizeof (struct SymStack));
    symstack->block_stack = NULL;
    symstack->block_sp    = -1;
    symstack->block_size  = 0;

    symstack->label        = NULL;
    symstack->switch_stack = NULL;
    symstack->switch_sp    = -1;
    symstack->switch_size  = 0;

    sym_table.func_stack [++sym_table.func_sp] = symstack;
}

void
sym_end_function (void)
{
    assert (sym_table.func_sp >= 0);
    symstack_free (sym_table.func_stack [sym_table.func_sp]);
    sym_table.func_stack [sym_table.func_sp] = NULL;
    sym_table.func_sp--;
}

void
sym_begin_block (void)
{
    struct SymStack *symstack = sym_table.func_stack [sym_table.func_sp];

    assert (symstack->block_sp < symstack->block_size);

    if (symstack->block_sp >= symstack->block_size - 1) {
	symstack->block_stack = (struct Symbol **) 
	    erealloc (symstack->block_stack,
		      sizeof (struct Symbol *) * ++symstack->block_size);
    }
    symstack->block_stack [++symstack->block_sp] = NULL;
}

void
sym_end_block (void)
{
    struct SymStack *symstack = sym_table.func_stack [sym_table.func_sp];

    assert (symstack->block_sp >= 0);

    sym_free (symstack->block_stack [symstack->block_sp]);
    symstack->block_sp--;
}

void
sym_begin_funcall  (struct AST const * const ast)
{
    struct SymList *symlist;

    assert ((ast->ast_type == AST_function_definition_spec)
	    || (ast->ast_type == AST_function_definition_nospec));

    for (symlist = sym_table.file; symlist != NULL; symlist = symlist->next) {
	if (!strcmp(ast->filename, symlist->filename))
	    break;
    }
    assert (symlist != NULL);
    sym_table.curr_file = symlist;
    sym_begin_function ();
    sym_table.func_stack [sym_table.func_sp]->curr_file = symlist;
}

void
sym_end_funcall    (void)
{
    sym_end_function ();
    if (sym_table.func_sp >= 0) {
	sym_table.curr_file
	    = sym_table.func_stack [sym_table.func_sp]->curr_file;
    }
}

void
assign_address_to_local_sym (struct Type *ret_type)
{
    struct SymStack    *symstack;
    struct Symbol      *sym;
    void               *arg_addr, *lvar_addr;

    symstack = sym_table.func_stack [sym_table.func_sp];
    sym      = symstack->block_stack [symstack->block_sp];

    arg_addr = vm.get_bp () - vm.VM_SIZEOF_VOID_P * 2; /* skip bp and pc.ast */
    lvar_addr = vm.get_sp ();

    /* skip space for return value */
    if (ret_type != NULL && !is_void_type (ret_type)) {
	arg_addr -= ret_type->size;
    }

    /* we assume that symbols are listed in order */
    for (; sym != NULL; sym = sym->next) {
	if (sym->status & SYM_STAT_ARGUMENT) {  /* sym is a formal parameter */
	    arg_addr -= get_type_size (sym->type);
	    sym->addr = arg_addr;
    	} else { /* sym is a local variable */
	    sym->addr = lvar_addr;
	    lvar_addr += get_type_size (sym->type);
    	}
    }
}

void
initialize_local_vars (void)
{
    struct SymStack *symstack = sym_table.func_stack [sym_table.func_sp];
    struct Symbol   *sym      = symstack->block_stack [symstack->block_sp];

    /* we assume that symbols are listed in order */
    for (; sym != NULL; sym = sym->next) {
	if (sym->status & SYM_STAT_ARGUMENT)
	    continue;
	if (sym->ast->ast_type != AST_init_declarator_dec_init)
	    continue;
	eval_initializer (sym->ast->u.child [1], sym->type, sym->addr);
    }
}

void
sym_entry_all_labels (struct AST const *ast)
{
    switch (ast->ast_type) {
    case AST_statement_list_opt_null:
    case AST_statement_jump:
    case AST_statement_expression:
    case AST_expression_statement_direct:
    case AST_jump_statement_return:
    case AST_jump_statement_break:
    case AST_jump_statement_continue:
    case AST_jump_statement_goto:
	break;

    case AST_statement_labeled:
    case AST_statement_compound:
    case AST_statement_selection:
    case AST_statement_iteration:
    case AST_statement_list_opt_direct:
    case AST_statement_list_single:
    case AST_labeled_statement_default:
    case AST_compound_statement_nodecn:
    case AST_iteration_statement_do_while:
	sym_entry_all_labels (ast->u.child [0]);
	break;

    case AST_labeled_statement_case:
    case AST_compound_statement_decn:
    case AST_selection_statement_if:
    case AST_iteration_statement_while:
    case AST_selection_statement_switch:
	sym_entry_all_labels (ast->u.child [1]);
	break;

    case AST_statement_list_pair:
	sym_entry_all_labels (ast->u.child [0]);
	sym_entry_all_labels (ast->u.child [1]);
	break;

    case AST_selection_statement_if_else:
	sym_entry_all_labels (ast->u.child [1]);
	sym_entry_all_labels (ast->u.child [2]);
	break;

    case AST_iteration_statement_for:
	sym_entry_all_labels (ast->u.child [3]);
	break;

    case AST_labeled_statement_label:
	sym_entry_label (ast);
	sym_entry_all_labels (ast->u.child [1]);
	break;

    default:
	assert (0);
	break;
    }
}

struct AST *
sym_lookup_label  (char const * const name)
{
    struct SymStack  *symstack = sym_table.func_stack [sym_table.func_sp];
    struct Symbol    *sym;

    for (sym = symstack->label; sym != NULL; sym = sym->next) {
	if (!strcmp (sym->name, name)) {
	    return sym->addr;
	}
    }
    return NULL;
}

void
sym_begin_switch (void)
{
    struct SymStack *symstack = sym_table.func_stack [sym_table.func_sp];

    assert (symstack->switch_sp < symstack->switch_size);
    if (symstack->switch_sp >= symstack->switch_size - 1) {
	symstack->switch_stack = erealloc (symstack->switch_stack,
					   sizeof (struct Attribute *)
					   * ++symstack->switch_size);
    }

    symstack->switch_stack [++symstack->switch_sp]
	= create_attr_sw_table (NULL, NULL, 0);
}

void
sym_end_switch (struct AST *ast_switch)
{
    struct SymStack *symstack = sym_table.func_stack [sym_table.func_sp];

    assert (ast_switch->ast_type == AST_selection_statement_switch);
    assert (symstack->switch_sp >= 0);

    set_block_sw_table (ast_switch,
			symstack->switch_stack [symstack->switch_sp]);
    symstack->switch_sp--;
}

void
sym_entry_case    (struct AST *ast_case)
{
    struct SymStack *symstack;
    struct Attribute *attr;

    assert (ast_case->ast_type == AST_labeled_statement_case);

    symstack = sym_table.func_stack [sym_table.func_sp];
    if (symstack->switch_sp < 0) {
	xci_error (ast_case, "'case' not within switch statement");
    }

    attr = symstack->switch_stack [symstack->switch_sp];
    assert (attr->kind == ATTR_SWITCH_TABLE);

    attr->u.a_sw_table.ast_case = erealloc (attr->u.a_sw_table.ast_case,
					    sizeof (struct AST *)
					    * ++attr->u.a_sw_table.case_num);

    attr->u.a_sw_table.ast_case [attr->u.a_sw_table.case_num - 1] = ast_case;
}

void
sym_entry_default (struct AST *ast_default)
{
    struct SymStack *symstack;
    struct Attribute *attr;

    assert (ast_default->ast_type == AST_labeled_statement_default);

    symstack = sym_table.func_stack [sym_table.func_sp];
    if (symstack->switch_sp < 0) {
	xci_error (ast_default, "'default' not within switch statement");
    }

    attr = symstack->switch_stack [symstack->switch_sp];
    assert (attr->kind == ATTR_SWITCH_TABLE);
    if (attr->u.a_sw_table.ast_default != NULL) {
	xci_error (ast_default, "two or more 'default:'");
    }
    attr->u.a_sw_table.ast_default = ast_default;
}

/* ---------------------------------------------------------------------- */
/* We assume that 'sym' has no loop and the tail is NULL-terminated. */
static struct Symbol *
sym_copy (struct Symbol *sym)
{
    struct Symbol *from, **to, *to_top;

    to = &to_top;
    for (from = sym; from != NULL; from = from->next) {
	*to = emalloc (sizeof (struct Symbol));
	**to = *from;
	to = &(*to)->next;
    }
    *to = NULL;
    return to_top;
}

static void
sym_free (struct Symbol *sym)
{
    struct Symbol *next;

    if (sym == NULL) return;

    next = sym->next;
    efree (sym);
    sym_free (next);
}

static void
symstack_free (struct SymStack *symstack)
{
    int i;

    if (symstack == NULL) return;
    for (i = 0; i <= symstack->block_sp; i++) {
    	sym_free (symstack->block_stack [i]);
    }
    efree (symstack->block_stack);
    efree (symstack);
}

/* for debugging */
void
symtable_dump (int const is_verbose)
{
    struct SymList     *symlist;
    struct SymStack    *symstack;
    struct Symbol      *sym;
    struct StaticBlock *sb;
    int                i, j;

    /* dump all symbols in all files */
    for (symlist = sym_table.file; symlist != NULL; symlist = symlist->next) {
	fprintf (stderr, "===== file: %s =====\n", symlist-> filename);
	for (sym = symlist->symbol; sym != NULL; sym = sym->next) {
	    sym_dump (sym, is_verbose);
	}
    }

    /* dump all symbols in all blocks of all stack frames */
    for (i = 0; i <= sym_table.func_sp; i++) {
	fprintf (stderr, "===== func level: %d =====\n", i);
	symstack = sym_table.func_stack [i];

	/* dump labels */
	fprintf (stderr, "===== labels =====\n");
	for (sym = symstack->label; sym != NULL; sym = sym->next) {
	    sym_dump (sym, is_verbose);
    	}

	for (j = 0; j <= symstack->block_sp; j++) {
	    fprintf (stderr, "===== block level: %d =====\n", j);
	    for (sym = symstack->block_stack [j]; sym != NULL;
						  sym = sym->next) {
		sym_dump (sym, is_verbose);
	    }
	}
    }

    /* dump all static block symbols */

    fprintf (stderr, "===== static block =====\n");
    for (sb = sym_table.static_block; sb != NULL; sb = sb->next) {
	assert ((sb->key->ast_type == AST_compound_statement_decn)
		|| (sb->key->ast_type == AST_compound_statement_nodecn));
	fprintf (stderr, "===== key: %p =====\n", sb->key);
	for (sym = sb->symbol; sym != NULL; sym = sym->next) {
	    sym_dump (sym, is_verbose);
	}
    }
}

static void
sym_dump (struct Symbol *sym, int is_verbose)
{
    fprintf (stderr, "%10s, %10p, %s, %s, %s\n",
	     sym->name,
	     sym->addr,
	     namespace2str (sym->namespace),
	     namelevel2str (sym->namelevel),
	     symstat2str   (sym->status));

#if 1
    if (sym->type == NULL) {
	fprintf (stderr, "VALUE: (type null)\n");
    } else if (sym->addr == NULL) {
	fprintf (stderr, "VALUE: (addr null)\n");
    } else if (is_integral_type (sym->type)) {
	fprintf (stderr, "VALUE: %d\n", *(int *)sym->addr);
    } else if (is_floating_type (sym->type)) {
	fprintf (stderr, "VALUE: %f\n", *(double *)sym->addr);
    } else if (sym->type->kind == TYPE_KIND_POINTER) {
	fprintf (stderr, "VALUE: %p\n", *(void **)sym->addr);
    } else {
	fprintf (stderr, "VALUE: XXX\n");    /* 蔲 */
    }
#endif

    if ((sym->type != NULL) && is_verbose) {
	type_dump (sym->type, 0);
    }
}

static char *
namelevel2str (enum NameLevel namelevel)
{
    switch (namelevel) {
    case NL_FILE:        return "file  ";
    case NL_BLOCK:       return "block ";
    default:             assert (0); return NULL;
    }
}

static char *
namespace2str (enum NameSpace namespace)
{
    switch (namespace) {
    case NS_NORMAL:      return "normal";
    case NS_LABEL:       return "label ";
    case NS_TAG:         return "tag   ";
    case NS_MEMBER:      return "member";
    default:             assert (0); return NULL;
    }
}

/*
  symstat2str stores bit information 'stat' of enum SymbolStatus
  into 'buf' as string like the following.
      "   :INC:   :DEF:   :   :   :   :   :"
  */ 
  
static char *
symstat2str (int stat)
{
    static char buf [BUFSIZ];

    buf [0] = '\0';
#if defined (PAD_EMPTY)
    assert (0);
#endif
#define PAD_EMPTY    strcat (buf, "   :")
    (stat & SYM_STAT_TENTATIVE)     ? strcat (buf, "TEN:") : PAD_EMPTY ;
    (stat & SYM_STAT_INCOMPLETE)    ? strcat (buf, "INC:") : PAD_EMPTY ;
    (stat & SYM_STAT_DECLARATION)   ? strcat (buf, "DEC:") : PAD_EMPTY ;
    (stat & SYM_STAT_DEFINITION)    ? strcat (buf, "DEF:") : PAD_EMPTY ;
    (stat & SYM_STAT_GLOBAL)        ? strcat (buf, "GLO:") : PAD_EMPTY ;
    (stat & SYM_STAT_ENUM_CONST)    ? strcat (buf, "ENU:") : PAD_EMPTY ;
    (stat & SYM_STAT_TYPEDEF)       ? strcat (buf, "TYP:") : PAD_EMPTY ;
    (stat & SYM_STAT_STATIC)        ? strcat (buf, "STA:") : PAD_EMPTY ;
    (stat & SYM_STAT_ARGUMENT)      ? strcat (buf, "ARG:") : PAD_EMPTY ;
    (stat & SYM_STAT_BUILTIN)       ? strcat (buf, "BLT:") : PAD_EMPTY ;
#undef PAD_EMPTY
    return buf;
}

#if 0
static char *
find_declarator_id (struct AST *ast)
{
    switch (ast->ast_type) {
    case AST_declarator_direct:
    case AST_function_declarator_direct:
	return find_declarator_id (ast->u.child[1]);
    
    case AST_direct_declarator_id:
	return ast->u.child[0]->u.string;

    case AST_direct_declarator_paren:
    case AST_direct_declarator_array:
    case AST_direct_declarator_func_old:
    case AST_direct_function_declarator_func_old:
    case AST_direct_declarator_func_new:
    case AST_direct_function_declarator_func_new:
	return find_declarator_id (ast->u.child[0]);

    default:
	assert (0);
	return NULL;
    }
}
#endif

static void
sym_entry_label   (struct AST const *ast_label)
{
    struct SymStack *symstack = sym_table.func_stack [sym_table.func_sp];
    struct Symbol   *sym;
    char   const    *name;

    assert (ast_label->ast_type == AST_labeled_statement_label);
    name = ast_label->u.child [0]->u.string;
    sym = sym_create (name, ast_label, NULL, (void *) ast_label, filename,
		      NS_LABEL, NL_BLOCK, SYM_STAT_NONE);
    sym->next = symstack->label;
    symstack->label = sym;
}
