/*
    XCI: eXperimental C Interpreter for ANSI C
    -- Copyright (c) 2000-2001, gondow@jaist.ac.jp
    $Id: symbol.h,v 1.2 2002/06/09 16:17:15 gondow Exp $
*/ 

#ifndef XCI_SYMBOL_H
#define XCI_SYMBOL_H
#include "type.h"
#include "attr.h"

enum NameSpace {
    NS_NORMAL,	  /* objects, functions, typedef names, and enum constants */
    NS_LABEL,	  /* labels */
    NS_TAG,       /* tags of structures, unions, and enumerations */
    NS_MEMBER     /* members of each structure or union individually (?) */
};

enum NameLevel {
    NL_FILE,            /* global variables and functions */
    NL_BLOCK            /* local variables and formal parameters */
};

enum SymbolStatus {
    SYM_STAT_NONE            = 0x000U,
    SYM_STAT_TENTATIVE       = 0x001U,	/* tentative definition */
    SYM_STAT_INCOMPLETE      = 0x002U,  /* incomplete type  */
    SYM_STAT_DECLARATION     = 0x004U,
    SYM_STAT_DEFINITION      = 0x008U,
    SYM_STAT_GLOBAL          = 0x010U,
    SYM_STAT_ENUM_CONST      = 0x020U,
    SYM_STAT_TYPEDEF         = 0x040U,
    SYM_STAT_STATIC          = 0x080U,  /* static local variable */
    SYM_STAT_ARGUMENT        = 0x100U,
    SYM_STAT_BUILTIN         = 0x200U
};

struct Symbol {
    char const	     *name;        /* symbol name */
    struct Type      *type;        /* symbol type (some shared) */
    struct AST       *ast;         /* AST where the symbol being declared */
    void    	     *addr;        /* address of variable, function, etc. */
    char    	     *filename;    /* filename that symbol belongs to */
    enum NameSpace   namespace;    /* name space for symbol */
    enum NameLevel   namelevel;    /* local, internal, or external */
    int              status;       /* combination of enum SymbolStatus */
    struct Symbol    *next;    	   /* pointer to next entry */
};

struct SymStack {
    struct Symbol   **block_stack; /* array to be extended dynamically */
    int             block_sp;	   /* stack top pointer */
    int             block_size;    /* size of currently allocated stack */
    struct SymList  *curr_file;    /* file symbol lists for this function */

    struct Symbol   *label;        /* list of all labels in the function */

    struct Attribute **switch_stack;
    int             switch_sp;
    int             switch_size;
};

struct SymList {
    char const      *filename;     /* filename that symbols belong to */
    struct Symbol   *symbol;       /* list for symbols */
    struct SymList  *next;         /* points to the next SymList */
};

struct StaticBlock {
    struct AST         *key;	      /* points compound_statement AST */
    struct Symbol      *symbol;       /* list for symbols */
    struct StaticBlock *next;         /* pointer to the next Local */
};

struct SymTable {
    struct SymList     *file;         /* symbol lists for internal linkage */
    struct SymList     *curr_file;    /* current symbol list for file */
    struct SymStack    **func_stack;  /* array of SymStack * for functions */
    int                func_sp;       /* stack pointer for func */
    int                func_size;     /* stack size for func */
    struct StaticBlock *static_block; /* local symbols stored statically */
};

/* ---------------------------------------------------------------------- */

struct Symbol *sym_lookup (char const * const name, const enum NameSpace ns);
struct Symbol *sym_lookup_at (char const * const name,
			      enum NameSpace const ns,
			      enum NameLevel const nl);
struct Symbol *sym_lookup_at_all_files (char const * const name,
					const enum NameSpace ns);
struct Symbol *sym_entry (struct Symbol * const sym);
struct Symbol *sym_entry_byname (char const        *const name,
				 struct AST        *ast,
				 struct Type       *type,
				 enum StorageClass storage,
				 enum NameSpace    ns);

void   sym_entry_static_block (struct AST *ast);
void   sym_copy_static_block  (struct AST const * const ast);

/* static file scope */
void   sym_begin_file     (char const * const filename);
void   sym_end_file       (void);

/* static function scope */
void   sym_begin_function (void);
void   sym_end_function   (void);

/* static/dynamic block scope */
void   sym_begin_block    (void);
void   sym_end_block      (void);

/* dynamic function scope */
void   sym_begin_funcall  (struct AST const * const ast);
void   sym_end_funcall    (void);

void   assign_address_to_local_sym (struct Type *ret_type);
void   initialize_local_vars (void);

/* for constructing switch table */
void   sym_begin_switch (void);
void   sym_end_switch   (struct AST *ast_switch);

void   sym_entry_case    (struct AST *ast_case);
void   sym_entry_default (struct AST *ast_default);

/* for fixing goto destination */
void       sym_entry_all_labels (struct AST const *ast);
struct AST *sym_lookup_label    (char const * const name);

struct Symbol *sym_create (char const * const name,
			   struct AST   const *ast,
			   struct Type        *type,
			   void               *addr,
			   char               *filename,
			   enum NameSpace     namespace,
			   enum NameLevel     namelevel,
			   int                sym_stat);

void   symtable_dump  (int const is_verbose);
/* ---------------------------------------------------------------------- */
extern struct SymTable sym_table;
/* ---------------------------------------------------------------------- */
#endif /* XCI_SYMBOL_H */
