/*
    XCI: eXperimental C Interpreter for ANSI C
    -- Copyright (c) 2000-2001, gondow@jaist.ac.jp
    $Id: type.c,v 1.2 2002/06/06 14:47:32 gondow Exp $
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "AST.h"
#include "xci.h"
#include "symbol.h"
#include "type.h"
#include "vm.h"
#include "eval_exp.h"

/* ---------------------------------------------------------------------- */
enum DeclContext decl_context = DECL_CONTEXT_GLOBAL;

/*
  Frequently used types. The function init_types () initialize
  this variable 'types'.
 */

static struct {
    struct Type *t_void;       /* void */
    struct Type *t_char;       /* char */
    struct Type *t_uchar;      /* unsigned char */
    struct Type *t_int;        /* int */
    struct Type *t_short;      /* short */
    struct Type *t_long;       /* long */
    struct Type *t_uint;       /* unsigned int */
    struct Type *t_ushort;     /* unsigned short */
    struct Type *t_ulong;      /* unsigned long */
#ifdef XCI_LONGLONG
    struct Type *t_longlong;   /* long long */
    struct Type *t_ulonglong;  /* unsigned long long */
#endif /* XCI_LONGLONG */
    struct Type *t_float;      /* float */
    struct Type *t_double;     /* double */
    struct Type *t_longdouble; /* long double */
    struct Type *t_char_p;     /* (char *) */
    struct Type *t_char_pp;    /* (char **) */
    struct Type *t_void_p;     /* (void *) */
    /*for default function declaration */
    struct Type *t_int_func;   /* int ()(...) */
} types;

enum { MAX_TYPE_DUMP_DEPTH = 5 };
/* ---------------------------------------------------------------------- */
static void   process_external_declaration_func (struct AST *ast);
static void   process_external_declaration_dec  (struct AST *ast);
static void   process_compound_statement (struct AST *ast);
static void   process_inner_blocks (struct AST *ast);

static int    create_sym_stat_outside (struct AST *ast, struct Type *type,
				       char const * const id,
				       enum StorageClass storage);
static int    create_sym_stat_inside  (struct AST *ast, struct Type *type,
				       char const * const id,
				       enum StorageClass storage);
static struct Type *create_declaration_specifiers_type
    	    	    	    	      (struct AST *ast,
				       enum StorageClass *storage_p);
static struct Type *find_and_create_principal_type (struct AST *ast);
static void   register_other_modifiers (struct AST *ast, struct Type *type,
					enum StorageClass *storage_p,
					int *qual_p);
static void   set_prim_type_size (struct Type *type);
static void   check_prim_modifiers (struct AST *ast, struct Type *type);

static struct Type *create_declarator_type
    	    	    	(struct AST *ast, struct Type *type, char **id_p);
static struct Type *create_abs_declarator_type
    	    	    	(struct AST *ast, struct Type *type);

static struct Type *create_prim_type (enum PrimType ptype);
static struct Type *create_struct_or_union_type (struct AST *ast);
static struct StructUnionField *create_struct_union_field
    	    	    	    	     (struct AST *ast, struct Type *type);
static struct Type *create_enum_type (struct AST *ast);
static struct Type *create_pointer_type (struct Type *type);
static int    find_type_qualifier (struct AST *ast);
static struct Type *create_array_type (struct Type *type, int num);
static struct Type *create_function_type (struct Type *type);
static void   register_new_style_arg_type (struct AST *ast, struct Type *type);
static void   register_old_style_arg_type (struct AST *ast, struct Type *type);
static void   register_id_list  (struct AST *ast, struct Type *type);

static struct Type *create_plain_type ();
static struct Type *create_qualifier_type
                 (struct AST *ast, enum TypeQualifier qual, struct Type *type);
static struct Type *create_qualifier_type_nocheck
                 (enum TypeQualifier qual, struct Type *type);

static char   *storage2str (enum StorageClass storage);
static char   *kind2str    (enum TypeKind kind);
static char   *qual2str    (int qual);
static char   *prim2str    (enum PrimType prim);
static char   *mod2str     (int type_mod);

static int    type_align      (struct Type *type);
static int    roundup_align   (int addr, int align);
static int    rounddown_align (int addr, int align);

static void   init_types ();
static void   type_analyze_find_exp (struct  AST *ast);
static void   type_analyze_exp_leaf (struct AST *ast);
static void   type_analyze_exp_node (struct AST *ast);
static struct Type *type_usual_arithmetic_conversion (struct AST *ast0,
						      struct AST *ast1);
static struct Type *type_integral_promotion (struct Type *type);
static struct Type *type_argument_promotion_ast  (struct AST  *ast);
static struct Type *type_argument_promotion_type (struct Type *type);
static void   type_pointer_generation_ast        (struct AST *ast);
static struct Type *type_pointer_generation_type (struct Type *type);

static int    is_type_identical  (struct Type *type1, struct Type *type2);

static int    is_lvalue                   (struct AST  *ast);
static int    is_modifiable_lvalue        (struct AST  *ast);
static int    is_string_literal           (struct AST  *ast);
static int    is_const_qualified_in_field (struct Type *type);
static int    is_void_pointer_type        (struct Type *type);
static int    is_char_pointer_type        (struct Type *type);
static int    is_char_array_type          (struct Type *type);

static struct AST *find_non_direct_parent (struct AST *ast, int *nth_p);

static int    is_type_assignable (struct AST *ast_right, 
				  struct Type *type_left,
				  struct Type *type_right);
static int    is_constant_zero   (struct AST *ast); 

static struct Type *type_analyze_add      (struct AST *ast0, struct AST *ast1);
static struct Type *type_analyze_sub      (struct AST *ast0, struct AST *ast1);
static struct Type *type_analyze_mult_div (struct AST *ast0, struct AST *ast1);
static struct Type *type_analyze_mod      (struct AST *ast0, struct AST *ast1);
static struct Type *type_analyze_shift    (struct AST *ast0, struct AST *ast1);
static struct Type *type_analyze_bop      (struct AST *ast0, struct AST *ast1);

static struct Type *get_type_no_skip (struct AST *ast);
static int         get_type_qual     (struct AST *ast);

static void   type_analyze_arg (struct AST *ast, struct Type *type);
static int    is_function_designator_id (struct AST *ast);

static int    type_analyze_initializer (struct AST *ast, struct Type *type);
static struct AST *type_analyze_initializer_list (struct AST *ast,
						  struct Type *type,
						  int *num_p);
static void   complete_array_type (struct Symbol *sym, int num);
/* ---------------------------------------------------------------------- */

/*
  By processing a given abstract syntax tree 'ast', this function
  constructs type information that represented as the structure type
  'struct Type'.
  As side effects, symbols are registered into the symbol table.
 */

void
type_analyze (struct AST *ast_root)
{
    struct AST *ast;

    assert ((ast_root->ast_type == AST_translation_unit_single)
	    || (ast_root->ast_type == AST_translation_unit_pair));

    init_types ();

    /* first search the bottom */
    for (ast = ast_root;
	 ast->ast_type != AST_translation_unit_single;
	 ast = ast->u.child [0]) ; /* empty body */
    
    while (1) {
	struct AST *ast_ext_decn;

	switch (ast->ast_type) {
	case AST_translation_unit_single:
	    ast_ext_decn = ast->u.child [0];
	    break;
	case AST_translation_unit_pair:
	    ast_ext_decn = ast->u.child [1];
	    break;
	default:
	    assert (0);
	    break;
	}

	switch (ast_ext_decn->ast_type) {
	case AST_external_declaration_func:
	    process_external_declaration_func (ast_ext_decn->u.child [0]);
	    break;
	case AST_external_declaration_dec:
	    process_external_declaration_dec (ast_ext_decn->u.child [0]);
	    break;
	default:
	    assert (0);
	    break;
	}

	if (ast == ast_root) {
	    break;
	}
	ast = ast->parent;
    }
}

int
is_type_compatible (struct Type *type1, struct Type *type2)
{
    /* xxx: not yet implemented */
    return 1;
}

int
is_sym_stat_compatible (int stat1, int stat2)
{
    /* xxx: not yet implemented */
    return 1;
}

int
create_sym_stat (struct AST *ast, struct Type *type,
		 char const * const id,
		 enum StorageClass storage)
{
    int sym_stat;

    switch (ast->ast_type){
    case AST_function_definition_spec:
    case AST_function_definition_nospec:
	sym_stat = SYM_STAT_DEFINITION;
	if ((storage == STORAGE_NONE) || (storage == STORAGE_EXTERN)) {
	    sym_stat |= SYM_STAT_GLOBAL;
	}
	break;

    case AST_struct_or_union_specifier_tag:
    case AST_struct_or_union_specifier_incomplete:
	if (type->size < 0) {
	    sym_stat = SYM_STAT_INCOMPLETE;
	} else {
	    sym_stat = SYM_STAT_DEFINITION;
	}
	break;
    case AST_enumerator_id:
    case AST_enumerator_id_init:
	sym_stat = SYM_STAT_DEFINITION | SYM_STAT_ENUM_CONST;
	break;

    case AST_enum_specifier_tag:
	sym_stat = SYM_STAT_DEFINITION;
	break;

    case AST_parameter_declaration_dec:
    case AST_declarator_direct:
	sym_stat = SYM_STAT_ARGUMENT | SYM_STAT_DEFINITION;
	break;

    case AST_init_declarator_dec:
    case AST_init_declarator_dec_init:
	switch (decl_context) {
	case DECL_CONTEXT_GLOBAL:
	    sym_stat = create_sym_stat_outside (ast, type, id, storage);
	    break;
	case DECL_CONTEXT_LOCAL:
	    sym_stat = create_sym_stat_inside (ast, type, id, storage);
	    break;
	default:
	    assert (0);
	    break;
	}
	break;
    case AST_IDENTIFIER:
	assert ((type->kind == TYPE_KIND_POINTER)
		&& (type->u.t_pointer.type->kind == TYPE_KIND_FUNCTION));
	sym_stat = SYM_STAT_DECLARATION | SYM_STAT_GLOBAL;
	break;
    default:
	assert (0);
	break;
    }

    return sym_stat;
}

void
type_dump (struct Type *type, int depth)
{
    int i;

    /* to avoid dumping a recursive type infinitely */
    if (depth > MAX_TYPE_DUMP_DEPTH) {
	fprintf (stderr, "%*ssnipped; max depth (%d) exceeded\n",
		 depth, "", MAX_TYPE_DUMP_DEPTH);
	return;
    }

    fprintf (stderr, "%*s%s (size=%d): ",
	     depth, "", kind2str (type->kind), type->size);

    switch (type->kind) {
    case TYPE_KIND_QUAL:
	fprintf (stderr, "%s ", qual2str (type->u.t_qual.qualifier));
	type_dump (type->u.t_qual.type, depth);
	break;
    case TYPE_KIND_PRIM:
	fprintf (stderr, "%s %s\n", prim2str (type->u.t_prim.ptype),
		 mod2str (type->u.t_prim.modifier));
	break;
    case TYPE_KIND_ENUM:
	fprintf (stderr, "%s: ", type->u.t_enum.tag);
	for (i = 0; i < type->u.t_enum.num; i++) {
	    fprintf (stderr, "%s ", type->u.t_enum.elem [i]);
	}
	fprintf (stderr, "\n");
	break;
    case TYPE_KIND_STRUCT:
    case TYPE_KIND_UNION:
	fprintf (stderr, "%s:\n", type->u.t_struct_union.tag);
	for (i = 0; i < type->u.t_struct_union.num; i++) {
	    struct StructUnionField *field;
	    field = type->u.t_struct_union.field [i];
	    if (field->is_bit_field) {
		fprintf (stderr, "%*s=>%s (offset=%d, boffset=%d, bsize=%d)\n",
			 depth, "", field->name, field->offset,
			 field->bit_offset, field->bit_size);
	    } else {
		fprintf (stderr, "%*s=>%s (offset=%d)\n",
			 depth, "", field->name, field->offset);
	    }
	    type_dump (field->type, depth + 1);
	}
	break;
    case TYPE_KIND_FUNCTION:
	fprintf (stderr, "%*sis_ellipsis=%d, is_old_type=%d\n",
		 depth, "", type->u.t_function.is_ellipsis,
		 type->u.t_function.is_old_type);
	fprintf (stderr, "%*s=>return\n", depth, "");
	type_dump (type->u.t_function.ret_type, depth + 1);
	for (i = 0; i < type->u.t_function.arg_num; i++) {
	    fprintf (stderr, "%*s=>arg [%d]: %s\n", depth, "",
		     i, type->u.t_function.arg_name [i]);
	    type_dump (type->u.t_function.arg_type [i], depth + 1);
	}
	break;
    case TYPE_KIND_POINTER:
	fprintf (stderr, "\n");
	type_dump (type->u.t_pointer.type, depth + 1);
	break;
    case TYPE_KIND_ARRAY:
	fprintf (stderr, "elem_num=%d\n", type->u.t_array.num);
	type_dump (type->u.t_array.elem_type, depth + 1);
	break;
    default:
	assert (0);
	break;
    }
}

int
get_type_size (struct Type *type)
{
    if (type->kind == TYPE_KIND_QUAL) {
	return get_type_size (type->u.t_qual.type);
    } else {
	return type->size;
    }
}

struct Type *
get_type (struct AST *ast)
{
    struct Type *type;

    type = (ast->prom_type != NULL) ? ast->prom_type : ast->type;
    if (type == NULL) {
	return NULL;
    }
    if (type->kind == TYPE_KIND_QUAL) {
	type = type->u.t_qual.type;
    }
    return type;
}

int
is_void_ast       (struct AST *ast)
{
    return is_void_type (get_type (ast));
}

int
is_void_type      (struct Type *type)
{
    return ((type->kind == TYPE_KIND_PRIM)
	    && (type->u.t_prim.ptype == PRIM_TYPE_VOID));
}

int
is_integral_ast   (struct AST *ast)
{
    return is_integral_type (get_type (ast));
}

int
is_integral_type   (struct Type *type)
{
    /* xxx: what about bit-field? */
    return ((type->kind == TYPE_KIND_ENUM)
	    || ((type->kind == TYPE_KIND_PRIM) 
		&& ((type->u.t_prim.ptype == PRIM_TYPE_CHAR)
		    || (type->u.t_prim.ptype == PRIM_TYPE_INT))));
}

int
is_floating_ast   (struct AST *ast)
{
    return is_floating_type (get_type (ast));
}

int
is_floating_type   (struct Type *type)
{
    return ((type->kind == TYPE_KIND_PRIM)
	    && ((type->u.t_prim.ptype == PRIM_TYPE_FLOAT)
		|| (type->u.t_prim.ptype == PRIM_TYPE_DOUBLE)));
}

int
is_arithmetic_ast (struct AST *ast)
{
    return (is_integral_ast (ast) || is_floating_ast (ast));
}

int
is_arithmetic_type (struct Type *type)
{
    return (is_integral_type (type) || is_floating_type (type));
}

int
is_type_equivalent (struct Type *type1, struct Type *type2)
{
    int i;

    /* 'identical' implies 'equivalent' */
    if (is_type_identical (type1, type2))
	return 1;

    /* xxx: before comparing, all type qualifiers are eliminated here.
       this feature is one of cutting corners. if you want to modify
       this part, please remember the following:
       the assignment of 'int * -> const int *' is OK, but
       that of 'int ** -> const int **' is not allowed. */
    if (type1->kind == TYPE_KIND_QUAL) type1 = type1->u.t_qual.type;
    if (type2->kind == TYPE_KIND_QUAL) type2 = type2->u.t_qual.type;

    /* different kind implies different type.
       please note that 'const int' and 'int' are different type,
       so if you want to regard them as equivalent then you have to
       explicitly remove type-qualifier in advance. */
    if (type1->kind != type2->kind)
	return 0;

    assert (type1->kind == type2->kind);
    switch (type1->kind) {
    case TYPE_KIND_PRIM:
	return ((type1->u.t_prim.ptype == type2->u.t_prim.ptype)
		&& (type1->u.t_prim.modifier == type2->u.t_prim.modifier));
    case TYPE_KIND_ENUM:
	return !strcmp (type1->u.t_enum.tag, type2->u.t_enum.tag);
    case TYPE_KIND_STRUCT:
    case TYPE_KIND_UNION:
	return !strcmp (type1->u.t_struct_union.tag,
			type2->u.t_struct_union.tag);
    case TYPE_KIND_FUNCTION:
	if (type1->u.t_function.is_ellipsis
	    != type2->u.t_function.is_ellipsis) return 0;
	if (type1->u.t_function.is_old_type
	    != type2->u.t_function.is_old_type) return 0;
	if (type1->u.t_function.arg_num
	    != type2->u.t_function.arg_num)     return 0;
	if (!is_type_equivalent (type1->u.t_function.ret_type,
				 type2->u.t_function.ret_type))  return 0;
	for (i = 0; i < type1->u.t_function.arg_num; i++) {
	    if (!is_type_equivalent (type1->u.t_function.arg_type [i],
				     type2->u.t_function.arg_type [i])) {
		return 0;
	    }
	}
	/* we do not check argument names */
	return 1;
    case TYPE_KIND_POINTER:
	return is_type_equivalent (type1->u.t_pointer.type,
				   type2->u.t_pointer.type);
    case TYPE_KIND_ARRAY:
	return ((type1->u.t_array.num == type2->u.t_array.num)
		&& is_type_equivalent (type1->u.t_array.elem_type,
				       type2->u.t_array.elem_type));
    case TYPE_KIND_QUAL:
	return ((type1->u.t_qual.qualifier == type2->u.t_qual.qualifier)
		&& is_type_equivalent (type1->u.t_qual.type,
				       type2->u.t_qual.type));
    default:
	assert (0);
	return 0;
    }
}


/*
  Set the field 'ast->type'.and 'ast->prom_type'.
  ast->type is the original type that 'ast' represents,
  while ast->prom_type is the type that ast->type is promoted into
  due to integral-promotion, pointer-generaion, usual-arithmetic-conversion,
  default-argument-promotion, and so on. (I hate implicit conversions in C!)
 */

void
type_analyze_exp (struct AST *ast)
{
    assert (ast != NULL);
    assert (is_AST_expression (ast));

    if (is_AST_LEAF ((struct AST const *) ast)) {
	type_analyze_exp_leaf (ast);
    } else {
	/* internal nodes in AST */
	type_analyze_exp_node (ast);
    }
}

void
check_main_type   (void)
{
    struct Symbol *sym;
    struct Type   *type;

    sym = sym_lookup_at_all_files (MAIN_NAME, NS_NORMAL);
    if (sym == NULL) {
	xci_error_noast ("no such function: main\n");
    }
    type = sym->type;
    assert (type != NULL);

    if (type->kind != TYPE_KIND_FUNCTION) {
	xci_error_noast ("main is not function\n");
    }

    if (!is_type_equivalent (type->u.t_function.ret_type, types.t_int)) {
	xci_error_noast ("return type of main is not int");
    }
    
    if (type->u.t_function.arg_num == 2) {
	if (!is_type_equivalent (type->u.t_function.arg_type [0],
				 types.t_int)) {
	    xci_error_noast ("first argument of main is not int");
    	}
	if (!is_type_equivalent (type->u.t_function.arg_type [1],
				 types.t_char_pp)) {
	    xci_error_noast ("second argument of main is not (char **)");
    	}
    } else if (type->u.t_function.arg_num != 0) {
	xci_error_noast ("the number of main's arguments must be 0 or 2");
    }
}
/* ---------------------------------------------------------------------- */
static void
process_external_declaration_func (struct AST *ast)
{
    struct Type	  *type;
    struct AST    *ast_decr, *ast_decn_list_opt, *ast_comp_stmt;
    char          *id;
    enum StorageClass  storage = STORAGE_NONE;

    sym_begin_function ();
    sym_begin_block ();

    assert ((ast->ast_type == AST_function_definition_spec)
	    || (ast->ast_type == AST_function_definition_nospec));

    switch (ast->ast_type) {
    case AST_function_definition_spec:
	type = create_declaration_specifiers_type (ast->u.child [0], &storage);
	ast_decr          = ast->u.child [1];
	ast_decn_list_opt = ast->u.child [2];
	ast_comp_stmt     = ast->u.child [3];
	break;
    case AST_function_definition_nospec:
	type = create_prim_type (PRIM_TYPE_INT);
	storage = STORAGE_NONE;
	ast_decr          = ast->u.child [0];
	ast_decn_list_opt = ast->u.child [1];
	ast_comp_stmt     = ast->u.child [2];
	break;
    default:
	assert (0);
	break;
    }

    assert (ast_decr->ast_type == AST_function_declarator_direct);

    type = create_declarator_type (ast_decr, type, &id); 

    assert (type->kind == TYPE_KIND_FUNCTION);

    if (ast_decn_list_opt->ast_type == AST_declaration_list_opt_direct) {
	assert (type->u.t_function.is_old_type);
	register_old_style_arg_type (ast_decn_list_opt->u.child [0], type);
    }

    /* register function name */
    (void) sym_entry_byname (id, ast, type, storage, NS_NORMAL);

    /* register all labels in function */
    sym_entry_all_labels (ast_comp_stmt);

    decl_context = DECL_CONTEXT_LOCAL; /* global variable */

    process_compound_statement (ast_comp_stmt);
    sym_entry_static_block (ast_comp_stmt);
    sym_end_block ();
    sym_end_function ();

    decl_context = DECL_CONTEXT_GLOBAL; /* global variable */
}

static void
process_external_declaration_dec (struct AST *ast)
{
    struct Type   *type;
    struct AST    *ast_init_decr_list, *ast1;
    enum StorageClass  storage = STORAGE_NONE;

    decl_context = DECL_CONTEXT_GLOBAL; /* global variable */
    assert (ast->ast_type == AST_declaration_direct);

    type = create_declaration_specifiers_type (ast->u.child [0], &storage);

    if (ast->u.child[1]->ast_type == AST_init_declarator_list_opt_null) {
	return;
    }

    assert (ast->u.child [1]->ast_type == AST_init_declarator_list_opt_direct);
    ast_init_decr_list = ast->u.child [1]->u.child [0];
    assert ((ast_init_decr_list->ast_type == AST_init_declarator_single)
	    || (ast_init_decr_list->ast_type == AST_init_declarator_pair));

    /* first search the bottom AST_init_declarator_single 
       to process declarators in depth-first */

    for (ast1 = ast_init_decr_list;
	 ast1->ast_type != AST_init_declarator_single;
	 ast1 = ast1->u.child [0]) ; /* empty body */
    
    assert (ast1->ast_type == AST_init_declarator_single);

    while (1) {
	struct AST    *ast_init_decr;
	struct Type   *new_type;
	char          *id;
	struct Symbol *sym;

	switch (ast1->ast_type) {
	case AST_init_declarator_single:
	    ast_init_decr = ast1->u.child [0];
	    break;
	case AST_init_declarator_pair:
	    ast_init_decr = ast1->u.child [1];
	    break;
	default:
	    assert (0);
	    break;
	}

	assert (ast_init_decr->u.child [0]->ast_type == AST_declarator_direct);

	new_type = create_declarator_type (ast_init_decr->u.child [0],
					   type, &id); 

	/* register declarations outside of functions  */
	sym = sym_entry_byname (id, ast_init_decr, new_type, storage,
				NS_NORMAL);

	/* initialize the variable if initializer exists */
	if (ast_init_decr->ast_type == AST_init_declarator_dec_init) {
	    int num; /* the number of outermost initializers in braces, or
			the length of string. e.g.,
			num = 3 for int a[] = {{10, 20}, {30, 40}, {50, 60}};
			num = 6 for char s[] = "hello";
			num = 0 for int b = 10; */
	    num = type_analyze_initializer (ast_init_decr->u.child [1],
					    sym->type);
	    complete_array_type (sym, num);
	    sym->addr = ecalloc (1, get_type_size (sym->type));
	    eval_initializer (ast_init_decr->u.child [1],
			      sym->type, sym->addr);
    	}

	if (ast1 == ast_init_decr_list)
	    break;
	ast1 = ast1->parent;
    }
}

static void
process_compound_statement (struct AST *ast)
{
    struct AST *ast_decn_list, *ast1, *ast_comp_stmt;

    assert (ast->ast_type == AST_compound_statement_decn
	    || ast->ast_type == AST_compound_statement_nodecn);

    switch (ast->ast_type) {
    case AST_compound_statement_decn:
	ast_comp_stmt = ast->u.child [1];
	break;
    case AST_compound_statement_nodecn:
	ast_comp_stmt = ast->u.child [0];
	break;
    default:
	assert (0);
	break;
    }

    if (ast->ast_type == AST_compound_statement_nodecn) {
	goto inner_block;
    }

    /* entry all local variables in this block to the symbol table */
    assert (ast->ast_type == AST_compound_statement_decn);
    ast_decn_list = ast->u.child [0];

    assert (ast_decn_list->ast_type == AST_declaration_list_single
	    || ast_decn_list->ast_type == AST_declaration_list_pair);


    /* first search the bottom */
    ast1 = ast_decn_list;
    while (ast1->ast_type != AST_declaration_list_single) {
	ast1 = ast1->u.child [0];
    }

    /* process each declaration */
    while (1) {
	struct AST     *ast_decn, *ast2;
	struct Type    *type;
	enum StorageClass storage = STORAGE_NONE;

	switch (ast1->ast_type) {
	case AST_declaration_list_single:
	    ast_decn = ast1->u.child [0];
	    break;
	case AST_declaration_list_pair:
	    ast_decn = ast1->u.child [1];
	    break;
	default:
	    assert (0);
	    break;
	}

	type = create_declaration_specifiers_type (ast_decn->u.child [0],
						   &storage);

	if (ast_decn->u.child [1]->ast_type
	    == AST_init_declarator_list_opt_null) {
	    break;
	}

	/* first search the bottom */
	for (ast2 = ast_decn->u.child [1]->u.child [0];
	     ast2->ast_type != AST_init_declarator_single;
	     ast2 = ast2->u.child [0]) ; /* empty body */

	assert (ast2->ast_type == AST_init_declarator_single);

	/* process each declarator */
	while (1) {
	    struct AST    *ast_init_decr;
	    struct Type   *new_type;
	    char          *id;
	    struct Symbol *sym;

	    switch (ast2->ast_type) {
	    case AST_init_declarator_single:
		ast_init_decr = ast2->u.child [0];
		break;
	    case AST_init_declarator_pair:
		ast_init_decr = ast2->u.child [1];
		break;
	    default:
		assert (0);
		break;
	    }

	    assert (ast_init_decr->u.child [0]->ast_type
		    == AST_declarator_direct);

	    new_type = create_declarator_type (ast_init_decr->u.child [0],
					       type, &id); 

	    /* register declarations within functions */
	    sym = sym_entry_byname (id, ast_init_decr, new_type, storage,
				    NS_NORMAL);

	    /* assign a given initial value into the variable if it exists */
	    if (ast_init_decr->ast_type == AST_init_declarator_dec_init) {
		int num;
		num = type_analyze_initializer (ast_init_decr->u.child [1],
						sym->type);
		complete_array_type (sym, num);
		if (sym->status & SYM_STAT_STATIC) {
		    sym->addr = ecalloc (1, get_type_size (sym->type));
		    eval_initializer (ast_init_decr->u.child [1],
				      sym->type, sym->addr);
    	    	}
	    }

	    if (ast2 == ast_decn->u.child [1]->u.child [0])
		break;
	    ast2 = ast2->parent;
	}

	if (ast1 == ast_decn_list)
	    break;
	ast1 = ast1->parent;
    }

  inner_block:
    /* process inner blocks */
    type_analyze_find_exp (ast_comp_stmt);
    process_inner_blocks (ast_comp_stmt);
}

static void
process_inner_blocks (struct AST *ast)
{
    switch (ast->ast_type) {
    case AST_statement_list_opt_null:
    case AST_statement_expression:
    case AST_jump_statement_return:
	break;

    case AST_jump_statement_goto:
	set_goto_dest (ast);
	break;

    case AST_jump_statement_continue:
	set_continue_exit (ast);
	break;

    case AST_jump_statement_break:
	set_break_exit (ast);
	break;

    case AST_statement_list_opt_direct:
    case AST_statement_list_single:
    case AST_statement_selection:
    case AST_statement_iteration:
    case AST_statement_labeled:
    case AST_statement_jump:
    case AST_iteration_statement_do_while:
	process_inner_blocks (ast->u.child [0]);
	break;

    case AST_labeled_statement_default:
	sym_entry_default (ast);
	process_inner_blocks (ast->u.child [0]);
	break;

    case AST_statement_list_pair:
	process_inner_blocks (ast->u.child [0]);
	process_inner_blocks (ast->u.child [1]);
	break;

    case AST_selection_statement_if:
    case AST_iteration_statement_while:
    case AST_labeled_statement_label:
	process_inner_blocks (ast->u.child [1]);
	break;

    case AST_selection_statement_switch:
	sym_begin_switch ();
	process_inner_blocks (ast->u.child [1]);
	sym_end_switch (ast);
	break;

    case AST_labeled_statement_case:
	sym_entry_case (ast);
	process_inner_blocks (ast->u.child [1]);
	break;

    case AST_selection_statement_if_else:
	process_inner_blocks (ast->u.child [1]);
	process_inner_blocks (ast->u.child [2]);
	break;

    case AST_iteration_statement_for:
	process_inner_blocks (ast->u.child [3]);
	break;

    case AST_statement_compound:
	sym_begin_block ();
	process_compound_statement (ast->u.child [0]);
	sym_entry_static_block (ast->u.child [0]);
	sym_end_block ();
	break;

    default:
	assert (0);
	break;
    }
}

static int
create_sym_stat_outside (struct AST *ast, struct Type *type,
			 char const * const id, enum StorageClass storage)
{
    int sym_stat = SYM_STAT_NONE;

    if ((storage == STORAGE_AUTO)||(storage == STORAGE_REGISTER)) {
	xci_error (ast, "invalid storage class specifier");
    }

    if (type->kind == TYPE_KIND_QUAL) {
	return create_sym_stat_outside (ast, type->u.t_qual.type,
					id, storage);
    }

    if (storage == STORAGE_TYPEDEF) {
	sym_stat |= SYM_STAT_TYPEDEF;
    }

    if ((type->kind == TYPE_KIND_ARRAY)
	|| (type->kind == TYPE_KIND_STRUCT)
	|| (type->kind == TYPE_KIND_UNION)) {
	if (type->size < 0) {
	    sym_stat |= SYM_STAT_INCOMPLETE;
	}
    }

    if (type->kind == TYPE_KIND_FUNCTION) {  /* function declaration */
	sym_stat |= SYM_STAT_DECLARATION;
	if (storage == STORAGE_NONE || storage == STORAGE_EXTERN) {
	    sym_stat |= SYM_STAT_GLOBAL;
	}
    } else {                                 /* global variable */
	switch (ast->ast_type) {
	case AST_init_declarator_dec_init:
	    switch (storage) {
	    case STORAGE_NONE:
		sym_stat |= SYM_STAT_GLOBAL | SYM_STAT_DEFINITION;
		break;
	    case STORAGE_STATIC:
		sym_stat |= SYM_STAT_DEFINITION;
		break;
	    case STORAGE_EXTERN:
		fprintf (stderr, "Warning: variable '%s' is initialized, "
			      "but declarated extern", id);
		sym_stat |= SYM_STAT_GLOBAL | SYM_STAT_DEFINITION;
		break;
	    default:
		assert (0);
		break;
	    }
	    break;
	case AST_init_declarator_dec:
	    switch (storage) {
	    case STORAGE_NONE:
		sym_stat |= SYM_STAT_GLOBAL | SYM_STAT_TENTATIVE;
		break;
	    case STORAGE_STATIC:
		sym_stat |= SYM_STAT_TENTATIVE;
		break;
	    case STORAGE_EXTERN:
		sym_stat |= SYM_STAT_DECLARATION | SYM_STAT_GLOBAL;
		break;
	    case STORAGE_TYPEDEF:
		sym_stat |= SYM_STAT_TYPEDEF;
		break;
	    default:
		assert (0);
		break;
	    }
	    break;
	default:
	    assert (0);
	    break;
	}
    }
    return sym_stat;
}

static int
create_sym_stat_inside (struct AST *ast, struct Type *type,
			char const * const id,
			enum StorageClass storage)
{
    int sym_stat = SYM_STAT_NONE;

    if (type->kind == TYPE_KIND_QUAL) {
	return create_sym_stat_inside (ast, type->u.t_qual.type, id, storage);
    }

    assert (type->kind != TYPE_KIND_QUAL);

    if ((type->kind == TYPE_KIND_FUNCTION)
	&& ((storage == STORAGE_AUTO)
	    || (storage == STORAGE_REGISTER)
	    || (storage == STORAGE_STATIC))) {
	xci_error (ast, "invalid storage class specifier");
    }

    if (storage == STORAGE_TYPEDEF) {
	sym_stat |= SYM_STAT_TYPEDEF;
    }

    if ((type->kind == TYPE_KIND_ARRAY)
	|| (type->kind == TYPE_KIND_STRUCT)
	|| (type->kind == TYPE_KIND_UNION)) {
	if (type->size < 0) {
	    sym_stat |= SYM_STAT_INCOMPLETE;
	}
    }

    if (type->kind == TYPE_KIND_FUNCTION) {  /* function declaration */
	sym_stat |= SYM_STAT_DECLARATION | SYM_STAT_GLOBAL;
    } else {                                 /* local variable */
	switch (storage) {
	case STORAGE_AUTO:
	case STORAGE_REGISTER:
	case STORAGE_NONE:
	    sym_stat |= SYM_STAT_DEFINITION;
	    break;
	case STORAGE_STATIC:
	    sym_stat |= SYM_STAT_STATIC | SYM_STAT_DEFINITION;
	    break;
	case STORAGE_EXTERN:
	    sym_stat |= SYM_STAT_DECLARATION | SYM_STAT_GLOBAL;
	    if (ast->ast_type == AST_init_declarator_dec_init) {
		xci_error_args (ast, "variable '%s' is initialized, "
				"but declarated extern", id);
	    }
	    break;
	case STORAGE_TYPEDEF:
	    sym_stat |= SYM_STAT_TYPEDEF;
	    break;
	default:
    	    /* ubN typedef int foo; ƃG[ł */
	    fprintf (stderr, "id = %s\n", id);
	    assert (0);
	    break;
	}
    }
    return sym_stat;
}

static struct Type *
create_declaration_specifiers_type (struct AST *ast,
				    enum StorageClass *storage_p)
{
    struct Type  *type, *new_type;
    int          qualifier = TYPE_QUAL_NONE;

    assert (is_AST_declaration_specifiers (ast)
	    || is_AST_specifier_qualifier_list (ast));

    type = find_and_create_principal_type (ast);
    if (type == NULL) {
	type = create_prim_type (PRIM_TYPE_INT);
    }

    register_other_modifiers (ast, type, storage_p, &qualifier);
    new_type = create_qualifier_type (ast, qualifier, type);
    check_prim_modifiers (ast, type);
    set_prim_type_size (type);
    return new_type;
}

/*
  looks for 'void', 'char', 'int', 'float', 'double', 
  'struct', 'union', or 'enum' from declaration_specifiers.
 */
static struct Type *
find_and_create_principal_type (struct AST *ast)
{
    struct Type *type0, *type1;
    struct Symbol *sym;

    switch (ast->ast_type) {
    case AST_declaration_specifers_storage_pair:
    case AST_declaration_specifers_spec_pair:
    case AST_declaration_specifers_qual_pair:
    case AST_specifier_qualifier_list_spec_pair:
    case AST_specifier_qualifier_list_qual_pair:
	type0 = find_and_create_principal_type (ast->u.child[0]);
	type1 = find_and_create_principal_type (ast->u.child[1]);
	if (type0 != NULL && type1 != NULL) {
    	    xci_error (ast, "two or more principal types in declaration");
    	}
	return (type0 != NULL ? type0 : type1);

    case AST_declaration_specifers_storage:
    case AST_declaration_specifers_spec:
    case AST_declaration_specifers_qual:
    case AST_specifier_qualifier_list_spec:
    case AST_specifier_qualifier_list_qual:
	return find_and_create_principal_type (ast->u.child[0]);

    case AST_storage_class_specifier_auto:
    case AST_storage_class_specifier_register:
    case AST_storage_class_specifier_static:
    case AST_storage_class_specifier_extern:
    case AST_storage_class_specifier_typedef:
    case AST_type_specifier_short:
    case AST_type_specifier_long:
    case AST_type_specifier_signed:
    case AST_type_specifier_unsigned:
    case AST_type_qualifier_const:
    case AST_type_qualifier_volatile:
	return NULL;

    case AST_type_specifier_void:
	return create_prim_type (PRIM_TYPE_VOID);
    case AST_type_specifier_char:
	return create_prim_type (PRIM_TYPE_CHAR);
    case AST_type_specifier_int:
	return create_prim_type (PRIM_TYPE_INT);
    case AST_type_specifier_float:
	return create_prim_type (PRIM_TYPE_FLOAT);
    case AST_type_specifier_double:
	return create_prim_type (PRIM_TYPE_DOUBLE);

    case AST_type_specifier_struct_or_union:
	return create_struct_or_union_type (ast->u.child [0]);
    case AST_type_specifier_enum:
	return create_enum_type (ast->u.child [0]);

    case AST_type_specifier_TYPEDEF_NAME:
	sym = sym_lookup (ast->u.child [0]->u.string, NS_NORMAL);
	return sym->type;
    default:
	assert (0);
	return NULL;
    }
}

static void
register_other_modifiers (struct AST *ast, struct Type *type,
			  enum StorageClass *storage_p, int *qual_p)
{
    enum StorageClass old_storage;

    if (type->kind == TYPE_KIND_QUAL) {
	register_other_modifiers (ast, type->u.t_qual.type,
				  storage_p, qual_p);
	return;
    }
    assert (type->kind != TYPE_KIND_QUAL);

    switch (ast->ast_type) {
    case AST_declaration_specifers_storage_pair:
    case AST_declaration_specifers_spec_pair:
    case AST_declaration_specifers_qual_pair:
    case AST_specifier_qualifier_list_spec_pair:
    case AST_specifier_qualifier_list_qual_pair:
	/* forwarding */
	register_other_modifiers (ast->u.child[0], type, storage_p, qual_p);
	register_other_modifiers (ast->u.child[1], type, storage_p, qual_p);
	return;

    case AST_declaration_specifers_storage:
    case AST_declaration_specifers_spec:
    case AST_declaration_specifers_qual:
    case AST_specifier_qualifier_list_spec:
    case AST_specifier_qualifier_list_qual:
	/* forwarding */
	register_other_modifiers (ast->u.child[0], type, storage_p, qual_p);
	return;

    case AST_type_specifier_void:
    case AST_type_specifier_char:
    case AST_type_specifier_int:
    case AST_type_specifier_float:
    case AST_type_specifier_double:
    case AST_type_specifier_struct_or_union:
    case AST_type_specifier_enum:
    case AST_type_specifier_TYPEDEF_NAME:
	/* do nothing */
	return;

    case AST_storage_class_specifier_auto:
    case AST_storage_class_specifier_register:
    case AST_storage_class_specifier_static:
    case AST_storage_class_specifier_extern:
    case AST_storage_class_specifier_typedef:
	old_storage = *storage_p;

	switch (ast->ast_type) {
	case AST_storage_class_specifier_auto:
	    *storage_p = STORAGE_AUTO;
	    break;
	case AST_storage_class_specifier_register:
	    *storage_p = STORAGE_REGISTER;
	    break;
	case AST_storage_class_specifier_static:
	    *storage_p = STORAGE_STATIC;
	    break;
	case AST_storage_class_specifier_extern:
	    *storage_p = STORAGE_EXTERN;
	    break;
	case AST_storage_class_specifier_typedef:
	    *storage_p = STORAGE_TYPEDEF;
	    break;
	default: 
	    assert (0);
	    break;
    	}

	if (old_storage != STORAGE_NONE) {
	    xci_error_args (ast, "two or more storage classes: `%s' and `%s'",
			    storage2str (*storage_p),
			    storage2str (old_storage));
    	}
	return;

    case AST_type_specifier_short:
	if (type->kind != TYPE_KIND_PRIM)
	    xci_error (ast, "short specified for non-primitive type");
	if (type->u.t_prim.modifier & TYPE_MOD_SHORT) 
	    xci_error (ast, "two or more short");
	if (type->u.t_prim.modifier & TYPE_MOD_LONG)
	    xci_error (ast, "both long and short specified");
	if (type->kind != TYPE_KIND_PRIM)
	    xci_error (ast, "short specified for non-primitive type");
	type->u.t_prim.modifier |= TYPE_MOD_SHORT;
	return;

    case AST_type_specifier_long:
	if (type->kind != TYPE_KIND_PRIM)
	    xci_error (ast, "long specified for non-primitive type");
#ifdef XCI_LONGLONG
	if (type->u.t_prim.modifier & TYPE_MOD_LONGLONG)
	    xci_error (ast, "three or more long");
#else
	if (type->u.t_prim.modifier & TYPE_MOD_LONG)
	    xci_error (ast, "two or more long");
#endif /* XCI_LONGLONG */
	if (type->u.t_prim.modifier & TYPE_MOD_SHORT)
	    xci_error (ast, "both long and short specified");
	if (type->kind != TYPE_KIND_PRIM)
	    xci_error (ast, "long specified for non-primitive type");
#ifdef XCI_LONGLONG
	if (type->u.t_prim.modifier & TYPE_MOD_LONG) { /* long long */
	    type->u.t_prim.modifier |= TYPE_MOD_LONGLONG;
	    type->u.t_prim.modifier &= ~TYPE_MOD_LONG;
	} else { /* long */
	    type->u.t_prim.modifier |= TYPE_MOD_LONG;
	}
#else
	type->u.t_prim.modifier |= TYPE_MOD_LONG;
#endif /* XCI_LONGLONG */
	return;

    case AST_type_specifier_signed:
	if (type->kind != TYPE_KIND_PRIM)
	    xci_error (ast, "signed specified for non-primitive type");
	if (type->u.t_prim.modifier & TYPE_MOD_SIGNED) 
	    xci_error (ast, "two or more signed");
	if (type->u.t_prim.modifier & TYPE_MOD_UNSIGNED)
	    xci_error (ast, "both signed and unsigned specified");
	if (type->kind != TYPE_KIND_PRIM)
	    xci_error (ast, "signed specified for non-primitive type");
	type->u.t_prim.modifier |= TYPE_MOD_SIGNED;
	return;

    case AST_type_specifier_unsigned:
	if (type->kind != TYPE_KIND_PRIM)
	    xci_error (ast, "unsigned specified for non-primitive type");
	if (type->u.t_prim.modifier & TYPE_MOD_UNSIGNED) 
	    xci_error (ast, "two or more unsigned");
	if (type->u.t_prim.modifier & TYPE_MOD_SIGNED)
	    xci_error (ast, "both signed and unsigned specified");
	if (type->kind != TYPE_KIND_PRIM)
	    xci_error (ast, "unsigned specified for non-primitive type");
	type->u.t_prim.modifier |= TYPE_MOD_UNSIGNED;
	return;

    case AST_type_qualifier_const:
	if (*qual_p & TYPE_QUAL_CONST) {
	    xci_error (ast, "two or more const");
	}
	*qual_p |= TYPE_QUAL_CONST;
	return;

    case AST_type_qualifier_volatile:
	if (*qual_p & TYPE_QUAL_VOLATILE)
	    xci_error (ast, "two or more volatile");
	*qual_p |= TYPE_QUAL_VOLATILE;
	return;

    default:
	assert (0);
	break;
    }
}

static void
set_prim_type_size (struct Type *type)
{
    assert (type->kind != TYPE_KIND_QUAL);
    if (type->kind != TYPE_KIND_PRIM)
	return;

    type->size = vm.size (type);
}

static void
check_prim_modifiers (struct AST *ast, struct Type *type)
{
    int type_mod;

    assert (type->kind != TYPE_KIND_QUAL);
    if (type->kind != TYPE_KIND_PRIM)
	return;

    type_mod = type->u.t_prim.modifier;
    switch (type->u.t_prim.ptype) {
    case PRIM_TYPE_VOID:
	if (type_mod != TYPE_MOD_NONE)
	    xci_error (ast, "invalid type modifier(s) for void");
	break;
    case PRIM_TYPE_CHAR:
	if (type_mod & (TYPE_MOD_SHORT | TYPE_MOD_LONG))
	    xci_error (ast, "invalid short or long for char");
	break;
    case PRIM_TYPE_INT:
	/* do nothing */
	break;
    case PRIM_TYPE_FLOAT:
	if (type_mod & (TYPE_MOD_SHORT | TYPE_MOD_LONG))
	    xci_error (ast, "invalid short or long for float");
	if (type_mod & (TYPE_MOD_SIGNED | TYPE_MOD_UNSIGNED))
	    xci_error (ast, "invalid signed or unsigned for float");
	break;
    case PRIM_TYPE_DOUBLE:
	if (type_mod & TYPE_MOD_SHORT)
	    xci_error (ast, "invalid short for double");
	if (type_mod & (TYPE_MOD_SIGNED | TYPE_MOD_UNSIGNED))
	    xci_error (ast, "invalid signed or unsigned for double");
	break;
    default:
	assert (0);
	break;
    }
}

static struct Type *
create_prim_type (enum PrimType ptype)
{
    struct Type *type;

    type = create_plain_type ();

    /* setting default values */
    type->kind              = TYPE_KIND_PRIM;
    type->size              = 0;
    type->u.t_prim.modifier = TYPE_MOD_NONE;
    type->u.t_prim.ptype    = ptype;

    return type;
}

static struct Type *
create_struct_or_union_type (struct AST *ast)
{
    struct Type   *type;
    struct AST    *ast_struct_decn_list, *ast1;
    int           field_num, i, offset, max_size;
    char          *tag;
    struct Symbol *sym;

    assert ((ast->ast_type == AST_struct_or_union_specifier_notag)
	    || (ast->ast_type == AST_struct_or_union_specifier_tag)
	    || (ast->ast_type == AST_struct_or_union_specifier_incomplete));

    type = create_plain_type ();

    type->size = -1;
    type->u.t_struct_union.num   = 0;
    type->u.t_struct_union.field = NULL;

    switch (ast->u.child[0]->ast_type) {
    case AST_struct_or_union_struct: type->kind = TYPE_KIND_STRUCT; break;
    case AST_struct_or_union_union:  type->kind = TYPE_KIND_UNION ; break;
    default:                         assert (0);                    break;
    }

    switch (ast->ast_type) {
    case AST_struct_or_union_specifier_incomplete:
	tag =  ast->u.child [1]->u.string;
	type->u.t_struct_union.tag = tag;
	sym = sym_lookup (tag, NS_TAG);
	if (sym == NULL) {
	    /* entry incomplete struct/union tag */
	    sym = sym_entry_byname (tag, ast, type, STORAGE_NONE, NS_TAG);
	    /*
	      struct foo { int a; };
	      {
	      	  struct foo; <= this declaration should be entried,
		              <= but is not in this implementation.
	      }
	     */
	}
	return sym->type;
    case AST_struct_or_union_specifier_tag:
	tag = ast->u.child [1]->u.string;
	type->u.t_struct_union.tag = tag;
	sym = sym_entry_byname (tag, ast, type, STORAGE_NONE, NS_TAG);
	type = sym->type;
	ast_struct_decn_list = ast->u.child [2];
	break;
    case AST_struct_or_union_specifier_notag:
	ast_struct_decn_list = ast->u.child [1];
	type->u.t_struct_union.tag = NULL;
	break;
    default:
	assert (0);
	break;
    }

    {
	enum AST_Type atype = ast_struct_decn_list->ast_type;
	assert ((atype == AST_struct_declaration_list_pair)
		|| (atype == AST_struct_declaration_list_single));
    }

    /* count fields first */
    field_num = 0;
    for (ast1 = ast_struct_decn_list; ; ast1= ast1->u.child[0]) {
	struct AST *ast_struct_decr_list, *ast2;
	switch (ast1->ast_type) {
	case AST_struct_declaration_list_pair:
	    ast_struct_decr_list = ast1->u.child[1]->u.child[1];
	    break;
	case AST_struct_declaration_list_single:
	    ast_struct_decr_list = ast1->u.child[0]->u.child[1];
	    break;
	default:
	    assert (0);
	    break;
	}
	{
	    enum AST_Type atype = ast_struct_decr_list->ast_type;
	    assert ((atype == AST_struct_declarator_list_single)
		    || (atype == AST_struct_declarator_list_pair));
	}
	for (ast2 = ast_struct_decr_list; ; ast2 = ast2->u.child[0]) {
	    field_num++;
	    if (ast2->ast_type == AST_struct_declarator_list_single)
		break;
	}
	if (ast1->ast_type == AST_struct_declaration_list_single)
	    break;
    }

    /* set fields */
    type->u.t_struct_union.num   = field_num;
    type->u.t_struct_union.field =
	emalloc (sizeof (struct Struct_Union_Field *) * field_num);

    i = field_num - 1;
    for (ast1 = ast_struct_decn_list; ; ast1= ast1->u.child[0]) {
	struct AST *ast_struct_decn, *ast2;
	struct Type *field_type;
	enum StorageClass storage = STORAGE_NONE;

	switch (ast1->ast_type) {
	case AST_struct_declaration_list_pair:
	    ast_struct_decn = ast1->u.child[1];
	    break;
	case AST_struct_declaration_list_single:
	    ast_struct_decn = ast1->u.child[0];
	    break;
	default:
	    assert (0);
	    break;
	}

	field_type = create_declaration_specifiers_type 
	    	    	(ast_struct_decn->u.child [0], &storage);

	for (ast2 = ast_struct_decn->u.child [1]; ; ast2 = ast2->u.child[0]) {
	    struct AST *ast_struct_decr;
	    switch (ast2->ast_type) {
	    case AST_struct_declarator_list_single:
		ast_struct_decr = ast2->u.child[0];
		break;
	    case AST_struct_declarator_list_pair:
		ast_struct_decr = ast2->u.child[1];
		break;
	    default:
		assert (0);
		break;
	    }
	    type->u.t_struct_union.field [i--] = 
		create_struct_union_field (ast_struct_decr, field_type);
	    if (ast2->ast_type == AST_struct_declarator_list_single)
		break;
	}
	if (ast1->ast_type == AST_struct_declaration_list_single)
	    break;
    }

    /* set offset and size */
    switch (type->kind) {
    case TYPE_KIND_STRUCT:
	offset = 0;
	for (i = 0; i < field_num; i++) {
	    struct StructUnionField *field;
	    field = type->u.t_struct_union.field [i];
	    /* the current implementation does not pack 
	       several bit-fields into the same word. */
	    field->offset = roundup_align (offset,
					   type_align (field->type));
	    offset = field->offset + field->type->size;
	}
	type->size = roundup_align (offset, type_align (NULL));
	break;
    case TYPE_KIND_UNION:
	max_size = 0;
	for (i = 0; i < field_num; i++) {
	    struct StructUnionField *field;
	    field = type->u.t_struct_union.field [i];
	    if (field->type->size > max_size)
		max_size = field->type->size;
	}
	type->size = max_size;
	break;
    default:
	assert (0);
	break;
    }

    /* register complete type definition, if necessary */
    {
	char *tag = type->u.t_struct_union.tag;
	if (tag != NULL) {
	    sym = sym_entry_byname (tag, ast, type, STORAGE_NONE, NS_TAG);
	    assert (sym->type == type);
	}
    }
    return type;
}

static struct StructUnionField *
create_struct_union_field (struct AST *ast, struct Type *type)
{
    struct StructUnionField *field;
    char   *field_name;

    assert ((ast->ast_type == AST_struct_declarator_direct)
	    || (ast->ast_type == AST_struct_declarator_bitfield_noname)
	    || (ast->ast_type == AST_struct_declarator_bitfield));

    field = emalloc (sizeof (struct StructUnionField));

    switch (ast->ast_type) {
    case AST_struct_declarator_bitfield_noname:
	field->type         = type;
	field->name         = ANONYMOUS_FIELD;
	field->offset       = 0;
	field->is_bit_field = 1;
	field->bit_offset   = 0;

	type_analyze_exp (ast->u.child [0]);
	eval_constant (ast->u.child [0], types.t_int, &field->bit_size);
	if (field->bit_size < 0) {
	    xci_error (ast, "negative width unnamed bit-field");
	}
	if (field->bit_size > vm.size (types.t_int) * 8) {
	    xci_error (ast, "max bit-field width exceeded");
	}
	break;

    case AST_struct_declarator_bitfield:
	field->type         = create_declarator_type (ast->u.child [0],
						      type, &field_name);
	field->name         = field_name;
	field->offset       = 0;
	field->is_bit_field = 1;
	field->bit_offset   = 0;
	type_analyze_exp (ast->u.child [1]);
	eval_constant (ast->u.child [1], types.t_int, &field->bit_size);
	if (field->bit_size == 0) {
	    xci_error_args (ast, "zero width bit-field: %s", field_name);
	} else if (field->bit_size < 0) {
	    xci_error_args (ast, "negative width bit-field: %s", field_name);
	} else if (field->bit_size > vm.size (types.t_int) * 8) {
	    xci_error_args (ast, "max bit-field width exceeded: %s",
			    field_name);
	}
	break;

    case AST_struct_declarator_direct:
	field->type         = create_declarator_type (ast->u.child [0],
						      type, &field_name);
	field->name         = field_name;
	field->offset       = 0;
	field->is_bit_field = 0;
	break;
    default:
	assert (0); break;
    }

    return field;
}

static struct Type *
create_enum_type (struct AST *ast)
{
    struct AST    *ast_enum_list, *ast2, *ast_enum;
    struct Type   *type;
    int           i, enum_num;
    unsigned int  enum_val;
    struct Symbol *sym;
    char          *tag;

    assert ((ast->ast_type == AST_enum_specifier_untag)
	    || (ast->ast_type == AST_enum_specifier_tag)
	    || (ast->ast_type == AST_enum_specifier_dec));

    switch (ast->ast_type) {
    case AST_enum_specifier_untag:
	tag = NULL;
	ast_enum_list = ast->u.child [0];
	break;
    case AST_enum_specifier_tag:
	tag = ast->u.child [0]->u.string;
	ast_enum_list = ast->u.child [1];
	break;
    case AST_enum_specifier_dec:
	tag = ast->u.child [0]->u.string;
	sym = sym_lookup (tag, NS_TAG);
	if (sym == NULL) {
	    xci_error_args (ast, "no such enum tag: ", tag);
	}
	return sym->type;
    default:
	assert (0);
	break;
    }

    type = create_plain_type ();

    /* setting default values */
    type->kind = TYPE_KIND_ENUM;
    type->size = vm.size (types.t_int);

    assert ((ast_enum_list->ast_type == AST_enumerator_list_single)
	    || (ast_enum_list->ast_type == AST_enumerator_list_pair));

    /* count enum elements first */
    enum_num = 1; /* count the last AST_enumerator_list_single first */
    for (ast2 = ast_enum_list;
	 ast2->ast_type == AST_enumerator_list_pair;
	 ast2 = ast2->u.child [0]) {
	enum_num++;
    }

    type->u.t_enum.tag   = tag;
    type->u.t_enum.num   = enum_num;
    type->u.t_enum.elem  = emalloc (sizeof (char *) * enum_num);
    type->u.t_enum.value = emalloc (sizeof (int ) * enum_num);

    /* set enum elements */
    assert (ast2->ast_type == AST_enumerator_list_single);
    enum_val = 0;
    for (i = 0; i< enum_num; i++, ast2 = ast2->parent) {
	char *id;

	switch (ast2->ast_type) {
	case AST_enumerator_list_single: ast_enum = ast2->u.child [0]; break;
	case AST_enumerator_list_pair:   ast_enum = ast2->u.child [1]; break;
	default:                         assert (0); break;
	}

	id = ast_enum->u.child [0]->u.string;
	type->u.t_enum.elem [i] = id;

	switch (ast_enum->ast_type) {
	case AST_enumerator_id:
	    /* do nothing */
	    break;
	case AST_enumerator_id_init:
	    type_analyze_exp (ast_enum->u.child [1]);
	    eval_constant (ast_enum->u.child [1], types.t_int, &enum_val);
	    break;
	default:
	    assert (0);
	    break;
	}

	/* register enum constant */
	sym = sym_entry_byname (id, ast_enum, type, STORAGE_NONE, NS_NORMAL);
	assert (sym->type == type);
	/* register enum value (and increment next enumeration) */
	*(int *)sym->addr = enum_val;
	type->u.t_enum.value [i] = enum_val;
	enum_val++;
    }
    assert (ast2 == ast_enum_list->parent);

    /* register enum tag */
    if (ast->ast_type == AST_enum_specifier_tag) {
	sym = sym_entry_byname (tag, ast, type, STORAGE_NONE, NS_TAG);
	assert (sym->type == type);
    }

    return type;
}

static struct Type *
create_declarator_type (struct AST *ast, struct Type *type, char **id_p)
{
    struct Type *new_type;
    struct AST  *ast_pointer;
    int         num, qualifier;

    switch (ast->ast_type) {
    case AST_declarator_direct:
    case AST_function_declarator_direct:
	if (ast->u.child[0]->ast_type == AST_pointer_opt_null)
	    return create_declarator_type (ast->u.child[1], type, id_p);
	assert (ast->u.child[0]->ast_type == AST_pointer_opt_direct);
	ast_pointer = ast->u.child[0]->u.child[0];
	assert ((ast_pointer->ast_type == AST_pointer_single)
		|| (ast_pointer->ast_type == AST_pointer_pair));
	new_type = type;
	while (1) {
	    new_type  = create_pointer_type (new_type);
	    qualifier = find_type_qualifier (ast_pointer->u.child[0]);
	    new_type  = create_qualifier_type (ast_pointer->u.child[0],
					       qualifier, new_type);

	    if (ast_pointer->ast_type == AST_pointer_single)
		break;
	    ast_pointer = ast_pointer->u.child[1];
	}
	return create_declarator_type (ast->u.child[1], new_type, id_p);
    
    case AST_direct_declarator_paren:
	return create_declarator_type (ast->u.child[0], type, id_p);

    case AST_direct_declarator_id:
	if (id_p != NULL)
	    *id_p = ast->u.child[0]->u.string;
	return type;

    case AST_direct_declarator_array:
	switch (ast->u.child[1]->ast_type) {
	case AST_constant_expression_opt_null:
	    num = -1; /* indicating incomplete type */
	    break;
	case AST_constant_expression_opt_direct:
	    type_analyze_exp (ast->u.child [1]->u.child [0]);
	    eval_constant (ast->u.child [1]->u.child [0], types.t_int, &num);
	    break;
	default:
	    assert (0);
	    break;
	}
	new_type = create_array_type (type, num);
	return create_declarator_type (ast->u.child [0], new_type, id_p);

    case AST_direct_declarator_func_new:
    case AST_direct_function_declarator_func_new:
	new_type = create_function_type (type);
	register_new_style_arg_type (ast->u.child [1], new_type);
	return create_declarator_type (ast->u.child [0], new_type, id_p);

    case AST_direct_declarator_func_old:
    case AST_direct_function_declarator_func_old:
	new_type = create_function_type (type);
	new_type->u.t_function.is_old_type = 1;

	if (ast->ast_type == AST_direct_function_declarator_func_old) {
	    /* old-style function parameter definition */
	    register_id_list (ast->u.child [1], new_type);
	} else { /* proto type dec. */
	    if (ast->u.child [1]->ast_type != AST_identifier_list_opt_null) {
		xci_error (ast, "old-style arguments must be empty "
			   "in prototype declaration");
	    }
	    /* 'extern int foo ()' == 'extern int foo (...)' */
	    new_type->u.t_function.is_ellipsis = 1;
	    new_type->u.t_function.arg_num     = 0;
	}
	return create_declarator_type (ast->u.child[0], new_type, id_p);

    default:
	assert (0);
	return NULL;
    }
}

static struct Type *
create_abs_declarator_type (struct AST *ast, struct Type *type)
{
    struct Type *new_type;
    struct AST  *ast_pointer, *ast_const_exp;
    int         num, qualifier;

    switch (ast->ast_type) {
    case AST_abstract_declarator_opt_null:
	return type;

    case AST_abstract_declarator_opt_direct:
	return create_abs_declarator_type (ast->u.child [0], type);

    case AST_abstract_declarator_pointer:
	ast_pointer = ast->u.child[0];
	assert ((ast_pointer->ast_type == AST_pointer_single)
		|| (ast_pointer->ast_type == AST_pointer_pair));
	new_type = type;
	while (1) {
	    new_type  = create_pointer_type (new_type);
	    qualifier = find_type_qualifier (ast_pointer->u.child[0]);
	    new_type  = create_qualifier_type (ast_pointer->u.child[0],
					       qualifier, new_type);

	    if (ast_pointer->ast_type == AST_pointer_single)
		break;
	    ast_pointer = ast_pointer->u.child[1];
	}
	return new_type;

    case AST_abstract_declarator_abs:
	if (ast->u.child[0]->ast_type == AST_pointer_opt_null)
	    return create_abs_declarator_type (ast->u.child[1], type);
	ast_pointer = ast->u.child [0]->u.child [0];
	assert ((ast_pointer->ast_type == AST_pointer_single)
		|| (ast_pointer->ast_type == AST_pointer_pair));
	new_type = type;
	while (1) {
	    new_type  = create_pointer_type (new_type);
	    qualifier = find_type_qualifier (ast_pointer->u.child[0]);
	    new_type  = create_qualifier_type (ast_pointer->u.child[0],
					       qualifier, new_type);
	    if (ast_pointer->ast_type == AST_pointer_single)
		break;
	    ast_pointer = ast_pointer->u.child[1];
	}
	return create_abs_declarator_type (ast->u.child[1], type);

    case AST_direct_abstract_declarator_paren:
	return create_abs_declarator_type (ast->u.child[0], type);

    case AST_direct_abstract_declarator_array_nil:
    case AST_direct_abstract_declarator_array_pair:
	switch (ast->ast_type) {
	case AST_direct_abstract_declarator_array_nil:
	    ast_const_exp = ast->u.child [0];
	    break;
	case AST_direct_abstract_declarator_array_pair:
	    ast_const_exp = ast->u.child [1];
	    break;
	default:
	    assert (0);
	    break;
	}

	switch (ast_const_exp->ast_type) {
	case AST_constant_expression_opt_null:
	    num = -1; /* indicating incomplete type */
	    break;
	case AST_constant_expression_opt_direct:
	    type_analyze_exp (ast_const_exp->u.child [0]);
	    eval_constant (ast_const_exp->u.child [0], types.t_int, &num);
	    break;
	default:
	    assert (0);
	    break;
	}
	new_type = create_array_type (type, num);
	switch (ast->ast_type) {
	case AST_direct_abstract_declarator_array_nil:
	    return new_type;
	case AST_direct_abstract_declarator_array_pair:
	    return create_abs_declarator_type (ast->u.child [0], new_type);
	default:
	    assert (0);
	    return NULL;
	}

    case AST_direct_abstract_declarator_func_nil:
	new_type = create_function_type (type);
	register_new_style_arg_type (ast->u.child [0], new_type);
	return new_type;

    case AST_direct_abstract_declarator_func_pair:
	new_type = create_function_type (type);
	register_new_style_arg_type (ast->u.child [1], new_type);
	return create_abs_declarator_type (ast->u.child [0], new_type);

    default:
	assert (0);
	return NULL;
    }
}

static struct Type *
create_pointer_type (struct Type *type)
{
    struct Type *new_type;

    new_type = create_plain_type ();

    new_type->kind      = TYPE_KIND_POINTER;
    new_type->size      = vm.VM_SIZEOF_VOID_P;
    new_type->u.t_pointer.type = type;

    return new_type;
}

static int
find_type_qualifier (struct AST *ast)
{
    int qual1, qual2;

    assert ((ast->ast_type == AST_type_qualifier_list_opt_nil)
	    || (ast->ast_type == AST_type_qualifier_list_opt_direct)
	    || (ast->ast_type == AST_type_qualifier_list_single)
	    || (ast->ast_type == AST_type_qualifier_list_pair));
    switch (ast->ast_type) {
    case AST_type_qualifier_list_opt_nil:
	return TYPE_QUAL_NONE;
    case AST_type_qualifier_list_opt_direct:
	return find_type_qualifier (ast->u.child[0]);
    case AST_type_qualifier_list_single:
	switch (ast->u.child[0]->ast_type) {
	case AST_type_qualifier_const:    return TYPE_QUAL_CONST;
	case AST_type_qualifier_volatile: return TYPE_QUAL_VOLATILE;
	default:                          assert (0); break;
	}
	assert (0);
	return TYPE_QUAL_NONE;
    case AST_type_qualifier_list_pair:
	qual1 = find_type_qualifier (ast->u.child[0]);
	switch (ast->u.child[1]->ast_type) {
	case AST_type_qualifier_const:
	    qual2 = TYPE_QUAL_CONST;
	    break;
	case AST_type_qualifier_volatile:
	    qual2 = TYPE_QUAL_VOLATILE;
	    break;
	default:
	    assert (0);
	    break;
	}
	if ((qual1 & TYPE_QUAL_CONST) && (qual2 & TYPE_QUAL_CONST)) {
	    xci_error (ast, "two or more const");
	}
	if ((qual1 & TYPE_QUAL_VOLATILE) && (qual2 & TYPE_QUAL_VOLATILE)) {
	    xci_error (ast, "two or more volatile");
	} 
	return (qual1 | qual2);
    default:
	assert (0);
	return TYPE_QUAL_NONE;
    }
}

static struct Type *
create_array_type (struct Type *type, int num)
{
    struct Type *new_type;

    new_type = create_plain_type ();

    new_type->kind      = TYPE_KIND_ARRAY;
    new_type->size      = get_type_size (type) * num;
    new_type->u.t_array.elem_type = type;
    new_type->u.t_array.num  = num;

    return new_type;
}

static struct Type *
create_function_type (struct Type *type)
{
    struct Type *new_type;

    new_type = create_plain_type ();

    new_type->kind      = TYPE_KIND_FUNCTION;
    new_type->size      = -1; /* sizeof may not be applied to function */
    new_type->u.t_function.ret_type     = type;
    new_type->u.t_function.is_ellipsis  = 0;
    new_type->u.t_function.is_old_type  = 0;
    new_type->u.t_function.arg_num      = 0;
    new_type->u.t_function.arg_name     = NULL;
    new_type->u.t_function.arg_type     = NULL;

    return new_type;
}

static void
register_new_style_arg_type (struct AST *ast, struct Type *type)
{
    int        arg_num, i, j;
    struct AST *ast_parameter_list;
    int        is_func_def;

    assert (type->kind != TYPE_KIND_QUAL);

    if (ast->ast_type == AST_parameter_type_list_opt_null) {
	return;
    } else if (ast->ast_type == AST_parameter_type_list_opt_direct) {
	register_new_style_arg_type (ast->u.child [0], type);
	return;
    }

    assert ((ast->ast_type == AST_parameter_type_list_direct)
	    || (ast->ast_type == AST_parameter_type_list_ellipsis));

    is_func_def = (ast->parent->ast_type
		   == AST_direct_function_declarator_func_new);

    if (ast->ast_type == AST_parameter_type_list_ellipsis) {
	type->u.t_function.is_ellipsis = 1;
    }

    type->u.t_function.is_old_type = 0;

    /* count arguments first and search the bottom */
    arg_num = 1; /* count the last AST_parameter_list_single in advance */
    for (ast_parameter_list = ast->u.child [0]; 
	 ast_parameter_list->ast_type == AST_parameter_list_pair;
	 ast_parameter_list = ast_parameter_list->u.child [0]) {
	assert ((ast_parameter_list->ast_type == AST_parameter_list_pair)
		||(ast_parameter_list->ast_type == AST_parameter_list_single));
	arg_num++;
    }

    type->u.t_function.arg_name = emalloc (sizeof (char *) * arg_num);
    type->u.t_function.arg_type = emalloc (sizeof (struct Type *) * arg_num);

    assert (ast_parameter_list->ast_type == AST_parameter_list_single);

    for (i=0; i<arg_num; i++) {
	struct AST         *ast_parameter_dec;
	struct Type        *arg_type;
	char               *arg_name;
	enum StorageClass storage = STORAGE_NONE;

	switch (ast_parameter_list->ast_type) {
	case AST_parameter_list_single:
	    ast_parameter_dec = ast_parameter_list->u.child [0];
	    break;
	case AST_parameter_list_pair:
	    ast_parameter_dec = ast_parameter_list->u.child [1];
	    break;
	default:
	    assert (0);
	    break;
    	}

	switch (ast_parameter_dec->ast_type) {
	case AST_parameter_declaration_dec:
	    arg_type = create_declaration_specifiers_type
	    	    	    (ast_parameter_dec->u.child [0], &storage);
	    arg_type = create_declarator_type
		    	    (ast_parameter_dec->u.child [1],
			     arg_type, &arg_name);
	    /* argument promotion (T []) -> (T *), T () -> (*T) () */
	    arg_type = type_pointer_generation_type (arg_type);
	    type->u.t_function.arg_name [i] = arg_name;
	    type->u.t_function.arg_type [i] = arg_type;

	    /* check if the same name exists */
	    for (j = 0; j < i; j++) {
		if (type->u.t_function.arg_name [j] == NULL) {
		    continue;
		}
		if (!strcmp (type->u.t_function.arg_name [j], arg_name)) {
		    xci_error_args (ast, "duplicate argument identifier '%s'",
				    arg_name);
		}
	    }

	    /* register this argument to the symbol table */
	    if (is_func_def) {
		(void) sym_entry_byname (arg_name, ast_parameter_dec,
					 arg_type, storage, NS_NORMAL);
	    }
	    break;

	case AST_parameter_declaration_abs:
	    arg_type = create_declaration_specifiers_type
	    	    	    (ast_parameter_dec->u.child [0], &storage);
	    arg_type = create_abs_declarator_type
		    	    (ast_parameter_dec->u.child [1], arg_type);
	    type->u.t_function.arg_name [i]   = NULL;
	    type->u.t_function.arg_type [i] = arg_type;
	    break;
	default:
	    assert (0);
	    break;
	}

	/* check the storage class */
	if ((storage != STORAGE_NONE) && (storage != STORAGE_REGISTER)) {
	    xci_error_args (ast,
			    "illegal storage class `%s' for parameter `%s'",
			    storage2str (storage), arg_name);
	}

	/* set the next parameter */
	ast_parameter_list = ast_parameter_list->parent;
    }

    assert (ast_parameter_list == ast);

    type->u.t_function.arg_num = arg_num;

    /* handling void parameter */
    if (arg_num == 1) {
	struct Type *first_arg_type = type->u.t_function.arg_type [0];

	if (first_arg_type->kind == TYPE_KIND_QUAL) {
	    first_arg_type = first_arg_type->u.t_qual.type;
	}
	assert (first_arg_type->kind != TYPE_KIND_QUAL);

	if (first_arg_type->kind == TYPE_KIND_PRIM
	    && first_arg_type->u.t_prim.ptype == PRIM_TYPE_VOID) {
	    type->u.t_function.arg_num = 0;
	    efree (type->u.t_function.arg_name);
	    efree (type->u.t_function.arg_type);
	    type->u.t_function.arg_name = NULL;
	    type->u.t_function.arg_type = NULL;
	}
    }

    /* check if all parameter names are given for function definition */
    if (is_func_def) {
	for (i = 0; i < type->u.t_function.arg_num; i++) {
	    if (type->u.t_function.arg_name [i] == NULL) {
		xci_error_args (ast, "parameter (%d) name omitted", i);
	    }
	}
    }
}

static void
register_id_list (struct AST *ast, struct Type *type)
{
    struct AST *ast_id_list;
    int         id_num, i;
    int         is_func_def;

    assert (type->kind != TYPE_KIND_QUAL);
    assert (ast->ast_type == AST_identifier_list_opt_direct
	    || ast->ast_type == AST_identifier_list_opt_null);
    assert (type->kind == TYPE_KIND_FUNCTION);

    if (ast->ast_type == AST_identifier_list_opt_null) {
	type->u.t_function.arg_num  = 0;
	type->u.t_function.arg_name = NULL;
	type->u.t_function.arg_type = NULL;
	return;
    }

    is_func_def = (ast->parent->ast_type
		   == AST_direct_function_declarator_func_old);

    /* first search the bottom and count id(entifier)s */
    id_num = 1;
    for (ast_id_list = ast->u.child [0];
	 ast_id_list->ast_type == AST_identifier_list_pair;
	 ast_id_list = ast_id_list->u.child [0]) {
	id_num ++;
    }
    assert (ast_id_list->ast_type == AST_identifier_list_single);

    type->u.t_function.arg_num  = id_num;
    type->u.t_function.arg_name = emalloc (sizeof (char *) * id_num);
    type->u.t_function.arg_type = NULL;

    for (i=0; i<id_num; i++) {
	if (i == 0) {
	    assert (ast_id_list->ast_type == AST_identifier_list_single);
	    type->u.t_function.arg_name [i] =
		ast_id_list->u.child [0]->u.string;
	    if (!is_func_def) {
		xci_error_args (ast,
				"parameter(s) declared without type(s): %s\n",
				type->u.t_function.arg_name [i]); 
	    }
	} else {
	    assert (ast_id_list->ast_type == AST_identifier_list_pair);
	    type->u.t_function.arg_name [i] = 
		ast_id_list->u.child [1]->u.string;
	}
	ast_id_list = ast_id_list->parent;
    }

    assert (ast_id_list == ast);
}

static void
register_old_style_arg_type (struct AST *ast, struct Type *type)
{
    struct AST        *ast1;
    char              **arg_name;
    struct Type       **arg_type;
    struct AST        **arg_ast;
    enum StorageClass *arg_storage;
    int               arg_num, i;
    
    assert (type->kind != TYPE_KIND_QUAL);

    assert ((ast->ast_type == AST_declaration_list_single)
	    || (ast->ast_type == AST_declaration_list_pair));

    assert ((ast->parent->parent->ast_type == AST_function_definition_spec)
	    || (ast->parent->parent->ast_type == AST_function_definition_nospec));

    arg_name    = type->u.t_function.arg_name;
    arg_num     = type->u.t_function.arg_num;
    arg_type    = ecalloc (arg_num, sizeof (struct Type *));
    arg_ast     = ecalloc (arg_num, sizeof (struct AST *));
    arg_storage = emalloc (sizeof (enum StorageClass) * arg_num);

    /* first search the bottom */
    ast1 = ast;
    while (ast1->ast_type != AST_declaration_list_single) {
	ast1 = ast1->u.child [0];
    }

    /* process each declaration */
    while (1) {
	struct AST   *ast_decn, *ast2;
	struct Type  *type;
	enum StorageClass storage = STORAGE_NONE;

	switch (ast1->ast_type) {
	case AST_declaration_list_single:
	    ast_decn = ast1->u.child [0];
	    break;
	case AST_declaration_list_pair:
	    ast_decn = ast1->u.child [1];
	    break;
	default:
	    assert (0);
	    break;
	}
	type = create_declaration_specifiers_type (ast_decn->u.child [0],
						   &storage);

	/* check the storage class */
	if ((storage != STORAGE_NONE) && (storage != STORAGE_REGISTER)) {
	    xci_error_args (ast, "illegal storage class `%s' for parameter",
			    storage2str (storage));
	}

	if (ast_decn->u.child [1]->ast_type
	    == AST_init_declarator_list_opt_null) {
	    break;
	}

	/* first search the bottom */
	for (ast2 = ast_decn->u.child [1]->u.child [0];
	     ast2->ast_type != AST_init_declarator_single;
	     ast2 = ast2->u.child [0]) ; /* empty body */

	assert (ast2->ast_type == AST_init_declarator_single);

	/* process each declarator */
	while (1) {
	    struct AST    *ast_init_decr;
	    struct Type   *new_type;
	    char          *id;

	    switch (ast2->ast_type) {
	    case AST_init_declarator_single:
		ast_init_decr = ast2->u.child [0];
		break;
	    case AST_init_declarator_pair:
		ast_init_decr = ast2->u.child [1];
		break;
	    default:
		assert (0);
		break;
	    }

	    assert (ast_init_decr->u.child [0]->ast_type
		    == AST_declarator_direct);

	    new_type = create_declarator_type (ast_init_decr->u.child [0],
					       type, &id); 

	    if ((storage != STORAGE_NONE) && (storage != STORAGE_REGISTER)) {
		xci_error_args (ast, "invalid storage class '%s'", id);
	    }

	    if (ast_init_decr->ast_type == AST_init_declarator_dec_init) {
		xci_error_args (ast, "parameter '%s' is initialized", id);
	    }

	    for (i = 0; i < arg_num; i++) {
		if (!strcmp (arg_name [i], id)) {
		    if (arg_type [i] != NULL) {
			xci_error_args (ast, "redeclaration of %s", id);
		    }
		    /* argument promotion (T []) -> (T *), T () -> (*T) () */
		    new_type = type_pointer_generation_type (new_type);
		    arg_type    [i] = new_type;
		    arg_ast     [i] = ast_init_decr->u.child [0];
		    arg_storage [i] = storage;
		    break;
		}

	    }

	    /* not found */
	    if (i == arg_num) {
		xci_error_args (ast, "no such parameter: %s\n", id);
	    }

	    if (ast2 == ast_decn->u.child [1]->u.child [0])
		break;
	    ast2 = ast2->parent;
	}

	if (ast1 == ast)
	    break;
	ast1 = ast1->parent;
    }

    /* set 'int' for each argument that has no declaration */
    for (i = 0; i < arg_num; i++) {
	if (arg_type [i] == NULL) {
	    arg_type    [i] = types.t_int;
	    arg_ast     [i] = ast;
	    arg_storage [i] = STORAGE_NONE;
	}
    }

    type->u.t_function.arg_type = arg_type;

    /* register old-style parameters to the symbol table */
    for (i=0; i<arg_num; i++) {
	(void) sym_entry_byname (arg_name [i], arg_ast [i], arg_type [i],
				 arg_storage [i], NS_NORMAL);
    }

    efree (arg_storage);
}

static struct Type *
create_plain_type ()
{
    struct Type *type;
    type = emalloc (sizeof (struct Type));

    return type;
}

static struct Type *
create_qualifier_type (struct AST *ast, enum TypeQualifier qual,
		       struct Type *type)
{
    if (qual == TYPE_QUAL_NONE) {
	return type;
    }

    if (type->kind == TYPE_KIND_QUAL) {
	/* we do not need to create qualifier_type since it already exists */
	/* check duplicate const or volatile */
	if ((qual & TYPE_QUAL_CONST)
	    && (type->u.t_qual.qualifier & TYPE_QUAL_CONST)) {
	    xci_error (ast, "two or more const");
	}
	if ((qual & TYPE_QUAL_VOLATILE)
	    && (type->u.t_qual.qualifier & TYPE_QUAL_VOLATILE)) {
	    xci_error (ast, "two or more volatile");
	}
	
	type->u.t_qual.qualifier |= qual;
	return type;
    } else {
	return create_qualifier_type_nocheck (qual, type);
    }
}

static struct Type *
create_qualifier_type_nocheck (enum TypeQualifier qual, struct Type *type)
{
    struct Type *new_type;
    new_type = create_plain_type ();
    new_type->kind = TYPE_KIND_QUAL;
    /* 
       new_type->size = type->size; is incorrect,
       because 'type' may be incomplete; therefore it will be
       completed later. This means that when obtaining the size of type,
       we always need to check if the type is TYPE_KIND_QUAL or not.
       This inconvenience is solved using the function get_type_size ().
       */
    new_type->size = -1;
    new_type->u.t_qual.qualifier = qual;
    new_type->u.t_qual.type      = type;
    return new_type;
}

static char *
storage2str (enum StorageClass storage)
{
    switch (storage) {
    case STORAGE_NONE:	       return "none";
    case STORAGE_REGISTER:     return "register";
    case STORAGE_AUTO:         return "auto";
    case STORAGE_STATIC:       return "static";
    case STORAGE_EXTERN:       return "extern";
    case STORAGE_TYPEDEF:      return "typedef";
    default:                   assert (0); return NULL;
    }
}

static char *
kind2str (enum TypeKind kind)
{
    switch (kind) {
    case TYPE_KIND_PRIM:     return "prim   ";
    case TYPE_KIND_ENUM:     return "enum   ";
    case TYPE_KIND_STRUCT:   return "struct ";
    case TYPE_KIND_UNION:    return "union  ";
    case TYPE_KIND_FUNCTION: return "func   ";
    case TYPE_KIND_POINTER:  return "pointer";
    case TYPE_KIND_ARRAY:    return "array  ";
    case TYPE_KIND_QUAL:     return "qual   ";
    default:	    	     assert (0); return NULL;
    }
}

static char *
qual2str (int qual)
{
    switch (qual) {
    case TYPE_QUAL_NONE:                       return "  ";
    case TYPE_QUAL_CONST:                      return "C ";
    case TYPE_QUAL_VOLATILE:                   return " V";
    case TYPE_QUAL_CONST | TYPE_QUAL_VOLATILE: return "CV";
    default:                                   assert (0); return NULL;
    }
}

static char *
prim2str (enum PrimType prim)
{
    switch (prim) {
    case PRIM_TYPE_VOID:    return "void";
    case PRIM_TYPE_CHAR:    return "char";
    case PRIM_TYPE_INT:     return "int";
    case PRIM_TYPE_FLOAT:   return "float";
    case PRIM_TYPE_DOUBLE:  return "double";
    default:                assert (0); return NULL;
    }
}

static char *
mod2str (int type_mod)
{
    static char buf [BUFSIZ];

    buf [0] = '\0';

    if (type_mod & TYPE_MOD_SIGNED) {
	strcat (buf, "signed   ");
    } else if (type_mod & TYPE_MOD_UNSIGNED) {
	strcat (buf, "unsigned ");
    } else {
	strcat (buf, "         ");
    }

    if (type_mod & TYPE_MOD_LONG) {
	strcat (buf, "long ");
    } else if (type_mod & TYPE_MOD_SHORT) {
	strcat (buf, "short");
    } else {
	strcat (buf, "     ");
    }
    return buf;
}

static int
type_align (struct Type *type)
{
    if (type == NULL) return 4;

    if (type->kind == TYPE_KIND_PRIM) {
	if (type->u.t_prim.ptype & PRIM_TYPE_CHAR) 
	    return 1;
#if 0
	if ((type->u.t_prim.ptype & PRIM_TYPE_INT)
	    && (type->u.t_prim.modifier & TYPE_MOD_SHORT)) 
	    return 2;
#endif
    }
    return 4;
}

static int
roundup_align (int addr, int align)
{
    return ((addr + align - 1) & -align);
}

static int
rounddown_align (int addr, int align)
{
    return (addr & -align);
}

/* 
   Initialize the structure 'types'; this must be called exactly once.
   I do not like to use structure initializer, since I think it is error-prone.
   */

static void
init_types ()
{
    static int called_times = 0;

    if (called_times++ != 0) {
	/* already initialized */
	return;
    }

    /* void */
    types.t_void       = create_prim_type (PRIM_TYPE_VOID);
    assert (types.t_void->kind == TYPE_KIND_PRIM);
    set_prim_type_size (types.t_void);

    /* char */
    types.t_char       = create_prim_type (PRIM_TYPE_CHAR);
    set_prim_type_size (types.t_char);

    /* unsigned char */
    types.t_uchar      = create_prim_type (PRIM_TYPE_CHAR);
    types.t_uchar->u.t_prim.modifier
	               = TYPE_MOD_UNSIGNED;
    set_prim_type_size (types.t_uchar);

    /* int */
    types.t_int        = create_prim_type (PRIM_TYPE_INT);
    set_prim_type_size (types.t_int);

    /* short */
    types.t_short      = create_prim_type (PRIM_TYPE_INT);
    types.t_short->u.t_prim.modifier
                       = TYPE_MOD_SHORT;
    set_prim_type_size (types.t_short);

    /* long */
    types.t_long       = create_prim_type (PRIM_TYPE_INT);
    types.t_long->u.t_prim.modifier
                       = TYPE_MOD_LONG;
    set_prim_type_size (types.t_long);

    /* unsigned int */
    types.t_uint       = create_prim_type (PRIM_TYPE_INT);
    types.t_uint->u.t_prim.modifier
                       = TYPE_MOD_UNSIGNED;
    set_prim_type_size (types.t_uint);

    /* unsigned short */
    types.t_ushort     = create_prim_type (PRIM_TYPE_INT);
    types.t_ushort->u.t_prim.modifier
                       = TYPE_MOD_UNSIGNED | TYPE_MOD_SHORT;
    set_prim_type_size (types.t_ushort);

    /* unsigned long */
    types.t_ulong      = create_prim_type (PRIM_TYPE_INT);
    types.t_ulong->u.t_prim.modifier
                       = TYPE_MOD_UNSIGNED | TYPE_MOD_LONG;
    set_prim_type_size (types.t_ulong);

#ifdef XCI_LONGLONG
    types.t_longlong   = create_prim_type (PRIM_TYPE_INT);
    types.t_longlong->u.t_prim.modifier
                       = TYPE_MOD_LONGLONG;
    set_prim_type_size (types.t_longlong);

    types.t_ulonglong  = create_prim_type (PRIM_TYPE_INT);
    types.t_ulonglong->u.t_prim.modifier
                       = TYPE_MOD_LONGLONG;
    set_prim_type_size (types.t_ulonglong);
#endif /* XCI_LONGLONG */

    /* float */
    types.t_float      = create_prim_type (PRIM_TYPE_FLOAT);
    set_prim_type_size (types.t_float);

    /* double */
    types.t_double     = create_prim_type (PRIM_TYPE_DOUBLE);
    set_prim_type_size (types.t_double);

    /* long double */
    types.t_longdouble = create_prim_type (PRIM_TYPE_DOUBLE);
    types.t_longdouble->u.t_prim.modifier
                       = TYPE_MOD_LONG;
    set_prim_type_size (types.t_longdouble);

    /* (char *) */
    types.t_char_p     = create_pointer_type (types.t_char);

    /* (char **) */
    types.t_char_pp    = create_pointer_type (types.t_char_p);

    /* (void *) */
    types.t_void_p     = create_pointer_type (types.t_void);

    /* int (*)(...) */
    {
	types.t_int_func = create_function_type (types.t_int);
	types.t_int_func->u.t_function.is_ellipsis  = 1;
	types.t_int_func->u.t_function.is_old_type  = 1;
    }
}

/*
  Visit a given AST in pre-order, and apply 'type_analyze_exp'
  to expression nodes .
 */

static void
type_analyze_find_exp (struct  AST *ast)
{
    /* skip inner blocks */
    if (ast->ast_type == AST_statement_compound) {
	return;
    }

    /* skip goto statement */
    if (ast->ast_type == AST_jump_statement_goto) {
	return;
    }

    /* expression found, so analyze it! */
    if (is_AST_expression (ast)) {
	type_analyze_exp (ast);
	return;
    }

    assert (!is_AST_expression (ast));

    if (is_AST_LEAF ((struct AST const *) ast)) {
	/* do nothing for leaves */
	return;
    } else {  /* analyze internal nodes in AST */
	int i;
	for (i = 0; i < ast->num_child; i++) {
	    type_analyze_find_exp (ast->u.child [i]);
	}
    }
}

static void
type_analyze_exp_leaf (struct AST *ast)
{
    struct Symbol *sym;

    switch (ast->ast_type) {
    case AST_ENUMERATION_CONSTANT:
    case AST_CHARACTER_CONSTANT:
	ast->type = types.t_int;
	break;

    case AST_STRING:
	ast->type = types.t_char_p;
	break;

    case AST_IDENTIFIER:
	/* this identifier designates a field name of structure or union. */
	if (((ast->parent->ast_type == AST_postfix_expression_dot)
	     || (ast->parent->ast_type == AST_postfix_expression_arrow))
	    && (ast->nth == 1)) {
	    int    i, num;
	    struct StructUnionField **field;
	    char   *name = ast->u.string;
	    struct Type *type_brother = get_type (ast->parent->u.child [0]);
	    
	    if (ast->parent->ast_type == AST_postfix_expression_arrow) {
		if (type_brother->kind != TYPE_KIND_POINTER) {
		    xci_error (ast->parent->u.child [0],
			       "invalid type of -> operand: pointer required");
		}
		type_brother = type_brother->u.t_pointer.type;
		if (type_brother->kind == TYPE_KIND_QUAL)
		    type_brother = type_brother->u.t_qual.type;
	    }
	    if ((type_brother->kind != TYPE_KIND_STRUCT)
		&& (type_brother->kind != TYPE_KIND_UNION)) {
		xci_error (ast, "invalid type of . or -> operand;"
			   " struct or union type required");
	    }
	    num   = type_brother->u.t_struct_union.num;
	    field = type_brother->u.t_struct_union.field;
	    for (i = 0; i < num; i++) {
		if (!strcmp (name, field [i]->name)) {
		    ast->type = field [i]->type;
		    break;
		}
	    }
	    if (i >= num) {
		xci_error_args (ast, "no such field: %s", name);
	    }
	} else {
	    /* lookup the identifier from symbol table. */
	    sym = sym_lookup (ast->u.string, NS_NORMAL);
	    if (sym == NULL) {
		/* if it is undeclared function, entry 'extern int func ();' */
		if (is_function_designator_id (ast)) {
		    sym = sym_create (ast->u.string, ast, types.t_int_func,
				      NULL, filename, NS_NORMAL, NL_FILE,
				      SYM_STAT_DECLARATION | SYM_STAT_GLOBAL);
		    sym_entry (sym);
		} else if (ast->parent->ast_type
			   == AST_labeled_statement_label) {
		    /* do nothing; this identifier represents goto label. */
		    return;
		} else {
		    xci_error_args (ast, "no such identifier: %s",
				    ast->u.string);
		}
	    }
	    ast->type = sym->type;
	}
	break;

    case AST_INTEGER_CONSTANT:
	if ((ast->status & INTEGER_CONSTANT_UNSIGNED)
	    && (ast->status & INTEGER_CONSTANT_LONG)) {      /* UL */
	    ast->type = types.t_ulong;
	} else if (ast->status & INTEGER_CONSTANT_LONG) {    /*  L */
	    if (ast->u.int_ul <= vm.VM_LONG_MAX) {
		ast->type = types.t_long;
	    } else {
		ast->type = types.t_ulong;
	    }
	} else if (ast->status & INTEGER_CONSTANT_UNSIGNED) { /* U  */
	    if (ast->u.int_ul <= vm.VM_UINT_MAX) {
		ast->type = types.t_uint;
	    } else {
		ast->type = types.t_ulong;
	    }
	} else {                                       /* no suffix */
	    if (ast->u.int_ul <= vm.VM_INT_MAX) {
		ast->type = types.t_int;
	    } else if (((ast->status & INTEGER_CONSTANT_OCTAL)
			|| (ast->status & INTEGER_CONSTANT_HEXADECIMAL))
		       && (ast->u.int_ul <= vm.VM_UINT_MAX)) {
		ast->type = types.t_uint;
	    } else if (ast->u.int_ul <= vm.VM_LONG_MAX) {
		ast->type = types.t_long;
	    } else {
		ast->type = types.t_ulong;
	    }
	}
	break;

    case AST_FLOATING_CONSTANT:
	if ((ast->status & FLOATING_CONSTANT_FLOAT)
	    && (ast->status & FLOATING_CONSTANT_LONG)) {
	    xci_error (ast, "illegal floating constant suffix: FL");
	} else if (ast->status & FLOATING_CONSTANT_FLOAT) {
	    ast->type = types.t_float;
	} else if (ast->status & FLOATING_CONSTANT_LONG) {
	    ast->type = types.t_longdouble;
	} else {
	    ast->type = types.t_double;
	}
	break;

    case AST_TYPEDEF_NAME:
	assert (0);
	ast->type = NULL;
	break;

    default:
	assert (0);
	ast->type = NULL;
	break;
    }

    type_pointer_generation_ast (ast);
}

static void
type_analyze_exp_node (struct AST *ast)
{
    int i;
    struct Type **type;

    type = emalloc (sizeof (struct Type *) * ast->num_child);

    /* first, visit all the ast's children */
    for (i = 0; i < ast->num_child; i++) {
	if (is_AST_expression (ast->u.child [i])) {
	    type_analyze_exp (ast->u.child [i]);
	} else if (ast->u.child [i]->ast_type == AST_type_name_direct) {
	    /* for cast and sizeof operator */
	    struct AST *ast_i = ast->u.child [i];
	    enum StorageClass storage = STORAGE_NONE;
	    struct Type *new_type;
	    new_type = create_declaration_specifiers_type (ast_i->u.child [0],
							   &storage);
	    if (storage != STORAGE_NONE) {
		xci_error (ast_i, "storage class specified in sizeof or cast");
	    }
	    new_type = create_abs_declarator_type (ast_i->u.child [1],
						   new_type);
	    ast_i->type = new_type;
	    type_pointer_generation_ast (ast_i);
	}
	type [i] = get_type (ast->u.child [i]);

	if (type [i] == NULL) {
	    assert ((!is_AST_expression (ast->u.child [i])
		     && (ast->u.child [i]->ast_type != AST_type_name_direct))
		    || (ast->u.child [i]->ast_type
			== AST_argument_expression_list_single)
		    || (ast->u.child [i]->ast_type
			== AST_argument_expression_list_pair));
	}
    }

    /* then, decide ast's type using the childrens' types */
    switch (ast->ast_type) {
    case AST_constant_expression_direct:
    case AST_expression_direct:
    case AST_assignment_expression_direct:
    case AST_conditional_expression_direct:
    case AST_logical_or_expression_direct:
    case AST_logical_and_expression_direct:
    case AST_inclusive_or_expression_direct:
    case AST_exclusive_or_expression_direct:
    case AST_and_expression_direct:
    case AST_equality_expression_direct:
    case AST_relational_expression_direct:
    case AST_shift_expression_direct:
    case AST_additive_expression_direct:
    case AST_multiplicative_expression_direct:
    case AST_cast_expression_direct:
    case AST_unary_expression_direct:
    case AST_cast_expression_cast:
    case AST_postfix_expression_direct:
    case AST_primary_expression_ID:
    case AST_primary_expression_constant:
    case AST_primary_expression_STRING:
    case AST_primary_expression_paren:
    case AST_constant_INT:
    case AST_constant_CHAR:
    case AST_constant_FLOAT:
    case AST_constant_ENUM:
	ast->type = type [0];
	break;

    case AST_expression_comma:
	ast->type = type [1];
	break;

    case AST_assignment_expression_assign:
	{
	    struct Type *type_right;

	    /* check type of left operand */
	    if (!is_modifiable_lvalue (ast->u.child [0])) {
#if 0
		asm ("int3");
#endif
		is_modifiable_lvalue (ast->u.child [0]);
		xci_error (ast->u.child [0], "left operand in assignment is"
			   " not modifiable lvalue");
	    }
	    /* e1 op= e2 is equivalent to e1 = ((e1) op (e2)) except that
	       e1 is evaluated only once.
	       here, calculate type of (e1) op (e2) into type_right */
	    switch (ast->u.child [1]->ast_type) {
	    case AST_assignment_operator_eq:
		type_right = type [2];
		break;
	    case AST_assignment_operator_mult:
	    case AST_assignment_operator_div:
		type_right = type_analyze_mult_div (ast->u.child [0],
						    ast->u.child [2]);
		break;
	    case AST_assignment_operator_mod:
		type_right = type_analyze_mod      (ast->u.child [0],
						    ast->u.child [2]);
		break;
	    case AST_assignment_operator_add:
		type_right = type_analyze_add      (ast->u.child [0],
						    ast->u.child [2]);
		break;
	    case AST_assignment_operator_sub:
		type_right = type_analyze_sub      (ast->u.child [0],
						    ast->u.child [2]);
		break;
	    case AST_assignment_operator_lshift:
	    case AST_assignment_operator_rshift:
		type_right = type_analyze_shift    (ast->u.child [0],
						    ast->u.child [2]);
		break;
	    case AST_assignment_operator_band:
	    case AST_assignment_operator_beor:
	    case AST_assignment_operator_bior:
		type_right = type_analyze_bop      (ast->u.child [0],
						    ast->u.child [2]);
		break;
	    default:
		assert (0);
		break;
	    }

	    /* finally, we check if the result of (e1) op (e2) 
	       is assignable to left operand. 
	       Note: type [0] is incorrect for 2nd argument, because
	       it may be promoted. */
	    if (!is_type_assignable (ast->u.child [2], ast->u.child [0]->type,
				     type_right)) {
		xci_error (ast, "incompatible type in assignment");
	    }

	    ast->type = ast->u.child [0]->type;
	}
	break;

    case AST_conditional_expression_cond:
	if (is_arithmetic_ast (ast->u.child [1])
	    && is_arithmetic_ast (ast->u.child [2])) {
	    ast->type = type_usual_arithmetic_conversion (ast->u.child [1],
							  ast->u.child [2]);
	} else if (is_type_equivalent (type [1], type [2])) {
	    /* pointer generation promotes array and function 
	       to pointers, so array and function must not come here. */
	    assert (type [1]->kind != TYPE_KIND_ARRAY);
	    assert (type [1]->kind != TYPE_KIND_FUNCTION);
	    ast->type = type [1];
	} else if ((type [1]->kind == TYPE_KIND_POINTER)
		   && is_constant_zero (ast->u.child [2])) {
	    ast->type = type [1];
	} else if ((type [2]->kind == TYPE_KIND_POINTER)
		   && is_constant_zero (ast->u.child [1])) {
	    ast->type = type [2];
	} else if (is_void_pointer_type (type [1])
		   && type [2]->kind == TYPE_KIND_POINTER) {
	    ast->type = types.t_void_p;
	} else if (is_void_pointer_type (type [2])
		   && type [1]->kind == TYPE_KIND_POINTER) {
	    ast->type = types.t_void_p;
	} else {
	    xci_error (ast, "type mismatch in conditional expression");
	}

	/* the result type inherits quailifiers from *both* types */
	{
	    int qual = (get_type_qual (ast->u.child [1])
			| get_type_qual (ast->u.child [2]));
	    ast->type = create_qualifier_type (ast, qual, ast->type);
	}
	break;

    case AST_logical_or_expression_lor:
    case AST_logical_and_expression_land:
	/* each operand must have arithmetic or pointer type */
	if ((type [0]->kind != TYPE_KIND_POINTER)
	    && !is_arithmetic_ast (ast->u.child [0])) {
	    xci_error (ast->u.child [0],
		       "invalid type in the left operand of || or &&; "
		       "arithmetic type required");
	}
	if ((type [1]->kind != TYPE_KIND_POINTER)
	    && !is_arithmetic_ast (ast->u.child [1])) {
	    xci_error (ast->u.child [1],
		       "invalid type in the right operand of || or &&; "
		       "arithmetic type required");
	}
	ast->type = types.t_int;
	break;

    case AST_and_expression_band:
    case AST_inclusive_or_expression_bior:
    case AST_exclusive_or_expression_beor:
	ast->type = type_analyze_bop (ast->u.child [0], ast->u.child [1]);
	break;

    case AST_equality_expression_eq:
    case AST_equality_expression_neq:
	if (!is_arithmetic_ast (ast->u.child [0])
	    && (get_type (ast->u.child [0])->kind != TYPE_KIND_POINTER)) {
	    xci_error (ast->u.child [0],
		       "invalid type in the left operand of == or !=; "
		       "arithmetic or pointer type required");
	}
	if (!is_arithmetic_ast (ast->u.child [1])
	    && (get_type (ast->u.child [1])->kind != TYPE_KIND_POINTER)) {
	    xci_error (ast->u.child [1],
		       "invalid type in the right operand of == or !=; "
		       "arithmetic or pointer type required");
	}

	if ((type [0]->kind == TYPE_KIND_POINTER)
	    && is_constant_zero (ast->u.child [1])) {
	    ast->type = type [0];
	} else if ((type [1]->kind == TYPE_KIND_POINTER)
		   && is_constant_zero (ast->u.child [0])) {
	    ast->type = type [1];
	} else if (is_void_pointer_type (type [0])
		   && type [1]->kind == TYPE_KIND_POINTER) {
	    ast->type = types.t_void_p;
	} else if (is_void_pointer_type (type [1])
		   && type [0]->kind == TYPE_KIND_POINTER) {
	    ast->type = types.t_void_p;
	} else if ((is_arithmetic_ast (ast->u.child [0])
		    && (type [1]->kind == TYPE_KIND_POINTER))
		    || (is_arithmetic_ast (ast->u.child [1])
			&& (type [0]->kind == TYPE_KIND_POINTER))) {
	    xci_error (ast, "arithmetic value and pointer compared");
	} else if ((type [0]->kind == TYPE_KIND_POINTER)
		   && (type [1]->kind == TYPE_KIND_POINTER)
		   && !is_type_equivalent (type [0], type [1])) {
	    if (!is_void_pointer_type (type [0])
		&& !is_void_pointer_type (type [1])) {
		xci_error (ast, "different type pointers compared");
	    }
	}
	if (is_arithmetic_ast (ast->u.child [0])
	    && is_arithmetic_ast (ast->u.child [1])) {
	    type_usual_arithmetic_conversion (ast->u.child [0],
					      ast->u.child [1]);
	}
	ast->type = types.t_int;
	break;

    case AST_relational_expression_less:
    case AST_relational_expression_more:
    case AST_relational_expression_lesseq:
    case AST_relational_expression_moreeq:
	if (!is_arithmetic_ast (ast->u.child [0])
	    && (type [0]->kind != TYPE_KIND_POINTER)) {
	    xci_error (ast->u.child [0],
		       "invalid type in the left operand of <, >, <=, "
		       "or >=; arithmetic or pointer type required");
	}
	if (!is_arithmetic_ast (ast->u.child [1])
	    && (type [1]->kind != TYPE_KIND_POINTER)) {
	    xci_error (ast->u.child [1],
		       "invalid type in the right operand of <, >, <=, "
		       "or >=; arithmetic or pointer type required");
	}
	if ((is_arithmetic_ast (ast->u.child [0])
	     && (type [1]->kind == TYPE_KIND_POINTER))
	    || (is_arithmetic_ast (ast->u.child [1])
		&& (type [0]->kind == TYPE_KIND_POINTER))) {
	    xci_error (ast, "arithmetic value and pointer compared");
	}
	if ((type [0]->kind == TYPE_KIND_POINTER)
	    && (type [1]->kind == TYPE_KIND_POINTER)
	    && !is_type_equivalent (type [0], type [1])) {
	    xci_error (ast, "different type pointers compared");
	}
	if (type [0]->kind != TYPE_KIND_POINTER) {
	    assert (is_arithmetic_ast (ast->u.child [0])
		    && is_arithmetic_ast (ast->u.child [1]));
	    type_usual_arithmetic_conversion (ast->u.child [0],
					      ast->u.child [1]);
	}
	ast->type = types.t_int;
	break;

    case AST_shift_expression_lshift:
    case AST_shift_expression_rshift:
	ast->type = type_analyze_shift (ast->u.child [0], ast->u.child [1]);
	break;

    case AST_additive_expression_add:
	ast->type = type_analyze_add   (ast->u.child [0], ast->u.child [1]);
	break;

    case AST_additive_expression_sub:
	ast->type = type_analyze_sub   (ast->u.child [0], ast->u.child [1]);
	break;

    case AST_multiplicative_expression_mult:
    case AST_multiplicative_expression_div:
	ast->type = type_analyze_mult_div (ast->u.child [0], ast->u.child [1]);
	break;

    case AST_multiplicative_expression_mod:
	ast->type = type_analyze_mod   (ast->u.child [0], ast->u.child [1]);
	break;

    case AST_unary_expression_unary_op:
	switch (ast->u.child [0]->ast_type) {
	case AST_unary_operator_addr:
	    /* operand must be lvalue */
	    if (!is_lvalue (ast->u.child [1])) {
		xci_error (ast->u.child [1], "unary & operand is not lvalue");
	    }
	    ast->type
		= create_pointer_type (get_type_no_skip (ast->u.child [1]));
	    /* type [1] is incorrect because its type modifier was removed. */
	    break;

	case AST_unary_operator_deref:
	    if (type [1]->kind != TYPE_KIND_POINTER) {
		xci_error (ast->u.child [1], 
			   "non-pointer value dereferenced");
	    }
	    ast->type = type [1]->u.t_pointer.type;
	    break;

	case AST_unary_operator_add:
	case AST_unary_operator_sub:
	    if (!is_arithmetic_ast (ast->u.child [1])) {
		xci_error (ast->u.child [1], "invalid type of unary + or -"
			   " operand; arithmetic type required");
	    }
	    ast->type = type_integral_promotion (type [1]);
	    break;

	case AST_unary_operator_comp:
	    if (!is_integral_ast (ast->u.child [1])) {
		xci_error (ast->u.child [1], "invalid type of ~ operand; "
			   "integral type required");
	    }
	    ast->type = type_integral_promotion (type [1]);
	    break;

	case AST_unary_operator_neg:
	    if (!is_arithmetic_ast (ast->u.child [1])
		&& type [1]->kind != TYPE_KIND_POINTER) {
		xci_error (ast->u.child [1], "invalid type of ! operand; "
			   "arithmetic or pointer type required");
	    }
	    ast->type = types.t_int;
	    break;

	default:
	    assert (0);
	    break;
	}
	break;

    case AST_unary_expression_inc:
    case AST_unary_expression_dec:
    case AST_postfix_expression_inc:
    case AST_postfix_expression_dec:
	/* operand must be modifiable lvalue */
	if (!is_modifiable_lvalue (ast->u.child [0])) {
	    xci_error (ast->u.child [0],
		       "operand of ++ or -- not modifiable lvalue");
	}
	if (!is_arithmetic_ast (ast->u.child [0])
	    && type [0]->kind != TYPE_KIND_POINTER) {
	    xci_error (ast->u.child [0], "invalid type of ++ or -- operand; "
		       "arithmetic or pointer type required");
	}
	ast->type = type_integral_promotion (get_type (ast->u.child [0]));
	break;

    case AST_unary_expression_sizeof_exp:
    case AST_unary_expression_sizeof_type:
	if (type [0]->kind == TYPE_KIND_FUNCTION) {
	    xci_error (ast->u.child [0], "sizeof applied to function");
	}
	if (type [0]->size < 0) {
	    xci_error (ast->u.child [0], "sizeof applied to incomplete type");
	}
	/* I should rule out bit-field, but I do not implement it yet. */
	ast->type = types.t_uint; /* this should be size_t in stddef.h */
	break;

    case AST_postfix_expression_array:
	/* one argument must be pointer and the other integral type */
	if (type [0]->kind == TYPE_KIND_POINTER) {
	    if (!is_integral_ast (ast->u.child [1])) {
		xci_error (ast, "subscript (right) is not an integral value");
	    }
	    ast->type = type [0]->u.t_pointer.type;
	} else if (type [1]->kind == TYPE_KIND_POINTER) {
	    if (!is_integral_ast (ast->u.child [0])) {
		xci_error (ast, "subscript (left) is not an integral value");
	    }
	    ast->type = type [1]->u.t_pointer.type;
	} else {
	    xci_error (ast, "no pointer in array");
	}
	break;

    case AST_postfix_expression_funcall:
    case AST_postfix_expression_funcall_noarg:
	if ((type [0]->kind != TYPE_KIND_POINTER)
	    || (type [0]->u.t_pointer.type->kind != TYPE_KIND_FUNCTION)) {
	    xci_error (ast->u.child [0], "not function pointer");
	}
	/* check parameter types */
	type_analyze_arg (ast, type [0]->u.t_pointer.type);
	ast->type = type [0]->u.t_pointer.type->u.t_function.ret_type;
	break;

    case AST_postfix_expression_arrow:
	{
	    struct Type *type_pointed_to;

	    if (type [0]->kind != TYPE_KIND_POINTER) {
		xci_error (ast->u.child [0], "left operand -> is not pointer");
	    }
	    type_pointed_to = type [0]->u.t_pointer.type;
	    if (type_pointed_to->kind == TYPE_KIND_QUAL) {
		type_pointed_to = type_pointed_to->u.t_qual.type;
	    }
	    assert (type_pointed_to->kind != TYPE_KIND_QUAL);
	    if ((type_pointed_to->kind != TYPE_KIND_STRUCT)
		&& (type_pointed_to->kind != TYPE_KIND_UNION)) {
		xci_error (ast->u.child [0], "neither struct nor union");
	    }
	    /* the above two checks imply that ast->u.child [0] is lvalue,
	       so I do not use is_lvalue () here. */

	    ast->type = type [1];
	}
	break;

    case AST_postfix_expression_dot:
	if (!is_lvalue (ast->u.child [0])) {
#if 0
	    asm ("int3");
#endif
	    is_lvalue (ast->u.child [0]);
	    xci_error (ast->u.child [0],
		       "not lvalue in left operand of '.' operator");
	}
	if ((type [0]->kind != TYPE_KIND_STRUCT)
	    && (type [0]->kind != TYPE_KIND_UNION)) {
	    xci_error (ast->u.child [0], "neither struct nor union");
	}
	ast->type = type [1];
	break;

    case AST_argument_expression_list_single:
    case AST_argument_expression_list_pair:
	/* do nothing here */
	break;


    case AST_STRING:
    case AST_IDENTIFIER:
    case AST_ENUMERATION_CONSTANT:
    case AST_INTEGER_CONSTANT:
    case AST_CHARACTER_CONSTANT:
    case AST_FLOATING_CONSTANT:
	/* these terminals (leaves) should be processed in 
	   type_analyze_exp_leaf (), so they must not appear here. */
	/* falling through */
    default:
	assert (0);
	break;
    }

    if (ast->type != NULL) {
	type_pointer_generation_ast (ast);
    }

    efree (type);
}

static int
is_type_identical (struct Type *type1, struct Type *type2)
{
    return (type1 == type2);
}

static struct Type *
type_usual_arithmetic_conversion (struct AST *ast0, struct AST *ast1)
{
    struct Type *type0 = get_type (ast0);
    struct Type *type1 = get_type (ast1);

#if defined (IF_EITHER_RETURN)
    assert (0);
#endif
#define IF_EITHER_RETURN(type, type0, type1)                    \
    {                                                           \
	int result0 = is_type_equivalent (type0, type);         \
	int result1 = is_type_equivalent (type1, type);         \
	if (result0 && result1) return type;                    \
	else if (result0)       return ast1->prom_type = type;  \
	else if (result1)       return ast0->prom_type = type;  \
    }

    IF_EITHER_RETURN (types.t_longdouble, type0, type1)
    IF_EITHER_RETURN (types.t_double,     type0, type1)
    IF_EITHER_RETURN (types.t_float,      type0, type1)

    type0 = ast0->prom_type = type_integral_promotion (type0);
    type1 = ast1->prom_type = type_integral_promotion (type1);

    IF_EITHER_RETURN (types.t_ulong,      type0, type1)

    if (is_type_equivalent (type0, types.t_long)
	 && is_type_equivalent (type1, types.t_uint)) {
	if (vm.VM_UINT_MAX <= vm.VM_LONG_MAX) {
	    return ast1->prom_type = types.t_long;
	} else {
	    return ast0->prom_type = types.t_uint;
	}
    }
    if (is_type_equivalent (type0, types.t_uint)
	&& is_type_equivalent (type1, types.t_long)) {
	if (vm.VM_UINT_MAX <= vm.VM_LONG_MAX) {
	    return ast0->prom_type = types.t_long;
	} else {
	    return ast1->prom_type = types.t_uint;
	}
    }

    IF_EITHER_RETURN (types.t_long,       type0, type1)
    IF_EITHER_RETURN (types.t_uint,       type0, type1)

    return types.t_int;
#undef IF_EITHER_RETURN 
}

/*
  this returns 'int' or 'unsigned int' according to integral promotion rule
  if a given type is 'char' or 'short'. it returns the type as it is otherwise.
 */
static struct Type *
type_integral_promotion (struct Type *type)
{
    /* xxx: In the current implementation, struct Type has no field
       for bit-fields. That is why they do not appear here. */

    if (type->kind == TYPE_KIND_QUAL) {
	struct Type *new_type;

	new_type = type_integral_promotion (type->u.t_qual.type);
	/* preserving type-qualifier */
	new_type = create_qualifier_type_nocheck (type->u.t_qual.qualifier,
						  new_type);
	return new_type;
    }

    if (type->kind == TYPE_KIND_ENUM) {
	return types.t_int;
    }

    if (type->kind != TYPE_KIND_PRIM) {
	return type;
    }

    assert (type->kind == TYPE_KIND_PRIM);

    if (type->u.t_prim.ptype == PRIM_TYPE_CHAR) {
	if (type->u.t_prim.modifier & TYPE_MOD_UNSIGNED) {
	    return ((vm.VM_UCHAR_MAX <= vm.VM_INT_MAX)
		    ? types.t_int : types.t_uint);
	} else {
	    return ((vm.VM_INT_MIN <= vm.VM_CHAR_MIN
		     && vm.VM_CHAR_MAX <= vm.VM_INT_MAX)
		    ? types.t_int : types.t_uint);
	}
    }

    if ((type->u.t_prim.ptype == PRIM_TYPE_INT)
	&& (type->u.t_prim.modifier & TYPE_MOD_SHORT)) {
	if (type->u.t_prim.modifier & TYPE_MOD_UNSIGNED) {
	    return ((vm.VM_USHRT_MAX <= vm.VM_INT_MAX)
		    ? types.t_int : types.t_uint);
	} else {
	    return ((vm.VM_INT_MIN <= vm.VM_SHRT_MIN
		     && vm.VM_SHRT_MAX < vm.VM_INT_MAX)
		    ? types.t_int : types.t_uint);
	}
    }

    return type;
}

/*
  Argument promotion converts type in old-style parameter as follows:
      integral_type -> integral promotion
      float         -> double
  Argument promotion does not take place for new-style parameter.
 */
static struct Type *
type_argument_promotion_ast (struct AST *ast)
{
    /* the outermost type-qualifier of actual arguments discarded */
    struct Type *type, *new_type;

    type = get_type (ast);
    new_type = type_argument_promotion_type (type);
    if (type != new_type) {
	ast->prom_type = new_type;
    }
    return new_type;
}

static struct Type *
type_argument_promotion_type (struct Type *type)
{
    if (is_integral_type (type)) {
	return type_integral_promotion (type);
    } else if ((type->kind == TYPE_KIND_PRIM)
	       && (type->u.t_prim.ptype == PRIM_TYPE_FLOAT)) {
	return types.t_double;
    }
    return type;
}

/*
  Pointer generation converts type as follows:
      array of T --> pointer to T
      function   --> pointer to function
  Pointer generation does not take place if it is the operand of 
  unary &, ++, --, sizeof, ., and assignment operators, for array, and
  if it is the operand of unary &, for function.
 */
static void
type_pointer_generation_ast (struct AST *ast)
{
    int    nth;
    struct AST *ast_up = find_non_direct_parent (ast, &nth);

    /* this function remove any type-qualifier in ast->type when
       storing its promoted type into ast->prom_type.
       in my understanding, this does not matter. */
    switch (get_type (ast)->kind) {
    case TYPE_KIND_ARRAY:
	if ((ast_up->ast_type == AST_unary_expression_unary_op)
	    && (ast_up->u.child [0]->ast_type == AST_unary_operator_addr)
	    && (nth == 1)) {
	    break;         /* skip for unary & */
	}
	if ((ast_up->ast_type == AST_postfix_expression_inc)
	    || (ast_up->ast_type == AST_postfix_expression_dec)
	    || (ast_up->ast_type == AST_unary_expression_inc)
	    || (ast_up->ast_type == AST_unary_expression_dec)) {
	    break;         /* skip for ++, -- */
	}
	if ((ast_up->ast_type == AST_unary_expression_sizeof_exp)
	    || (ast_up->ast_type == AST_unary_expression_sizeof_type)) {
	    break;         /* skip for sizeof */
	}
	if ((ast_up->ast_type == AST_assignment_expression_assign)
	    && (nth == 0)) {
	    break;         /* skip for op= (all assignment operators) */
	}
	if ((ast_up->ast_type == AST_postfix_expression_dot)
	    && (nth == 0)) {
	    break;         /* skip for structure reference */
	}
	ast->prom_type = type_pointer_generation_type (get_type (ast));
	break;
    case TYPE_KIND_FUNCTION:
	if ((ast_up->ast_type == AST_unary_expression_unary_op)
	    && (ast_up->u.child [0]->ast_type == AST_unary_operator_addr)
	    && (nth == 1)) {
	    break;         /* skip for unary & */
	}
	ast->prom_type = type_pointer_generation_type (get_type (ast));
	break;
    default:
	break;
    }
}

static struct Type *
type_pointer_generation_type (struct Type *type)
{
    switch (type->kind) {
    case TYPE_KIND_ARRAY:
	return create_pointer_type (type->u.t_array.elem_type);
    case TYPE_KIND_FUNCTION:
	return create_pointer_type (type);
    default:
	return type;
    }
}

static int
is_lvalue (struct AST *ast)
{
    switch (ast->ast_type) {
    case AST_IDENTIFIER:
	/* identifier must be neither function, enumeration constant
	   nor typedef_name */
	{
	    char *id = ast->u.string;
	    struct Symbol *sym = sym_lookup (id, NS_NORMAL);

	    if (sym == NULL) {
		xci_error_args (ast, "no such identifier: %s\n", id);
	    }
	    if ((sym->status & SYM_STAT_ENUM_CONST)
		|| (sym->status & SYM_STAT_TYPEDEF)) {
		return 0;
	    }

	    if (!is_arithmetic_type (sym->type)
		&& (sym->type->kind != TYPE_KIND_STRUCT)
		&& (sym->type->kind != TYPE_KIND_UNION)
		&& (sym->type->kind != TYPE_KIND_POINTER)) {
		return 0;
	    }
	    return 1;
	}

    case AST_unary_expression_unary_op:
	{
	    struct Type *type_pointed_to, *type1;

	    type1 = get_type (ast->u.child [1]);
	    if (ast->u.child [0]->ast_type != AST_unary_operator_deref) {
		return 0;
	    }
	    if (type1->kind != TYPE_KIND_POINTER) {
		return 0;
	    }
	    type_pointed_to = type1->u.t_pointer.type;
	    return  (is_arithmetic_type (type_pointed_to)
		     || (type_pointed_to->kind == TYPE_KIND_STRUCT)
		     || (type_pointed_to->kind == TYPE_KIND_UNION)
		     || (type_pointed_to->kind == TYPE_KIND_POINTER));
	}

    case AST_postfix_expression_dot:
	return (is_lvalue (ast->u.child [0])
		&& (get_type (ast->u.child [1])->kind != TYPE_KIND_ARRAY));
	
    case AST_postfix_expression_arrow:
	return (get_type (ast->u.child [1])->kind != TYPE_KIND_ARRAY);

    case AST_postfix_expression_array:
	{
	    struct Type *type_pointer, *type_pointed_to;

	    /* expression 'e1 [e2]' is just abbreviation of '((e1) + (e2))',
	       so first calculate the type of '((e1) + (e2))'. */
	    type_pointer = type_analyze_add (ast->u.child [0],
					     ast->u.child [1]);
	    /* the following checking is the same as that of
	       AST_unary_operator_deref. (see 25 lines above) */
	    if (type_pointer->kind != TYPE_KIND_POINTER) {
		return 0;
	    }
	    type_pointed_to = type_pointer->u.t_pointer.type;
	    if (type_pointed_to->kind == TYPE_KIND_QUAL) {
		type_pointed_to = type_pointed_to->u.t_qual.type;
	    }
	    assert (type_pointed_to->kind != TYPE_KIND_QUAL);
	    return (is_arithmetic_type (type_pointed_to)
		    || (type_pointed_to->kind == TYPE_KIND_STRUCT)
		    || (type_pointed_to->kind == TYPE_KIND_UNION)
		    || (type_pointed_to->kind == TYPE_KIND_POINTER));
	}

    case AST_primary_expression_ID:
    case AST_primary_expression_paren:
    case AST_postfix_expression_direct:
    case AST_unary_expression_direct:
    case AST_cast_expression_direct:
    case AST_multiplicative_expression_direct:
    case AST_additive_expression_direct:
    case AST_shift_expression_direct:
    case AST_relational_expression_direct:
    case AST_equality_expression_direct:
    case AST_and_expression_direct:
    case AST_exclusive_or_expression_direct:
    case AST_inclusive_or_expression_direct:
    case AST_logical_and_expression_direct:
    case AST_logical_or_expression_direct:
    case AST_conditional_expression_direct:
    case AST_assignment_expression_direct:
    case AST_expression_direct:
	return is_lvalue (ast->u.child [0]);

    default:
	return 0;
    }
}

static int
is_modifiable_lvalue (struct AST *ast)
{
    struct Type *type = get_type (ast);
    int          qual = get_type_qual (ast);

    /* is_lvalue is a necessary condition */
    if (!is_lvalue (ast)) {
	return 0;
    }

    /* there must be no 'const' type-qualifier */
    if (qual & TYPE_QUAL_CONST) {
	return 0;
    }
    
    /* all members and submembers must have no 'const'. */
    if (type->kind == TYPE_KIND_STRUCT || type->kind == TYPE_KIND_UNION) {
	if (is_const_qualified_in_field (type)) 
	    return 0;
    }
    
    /* neither array nor function is modifiable lvalue */
    if (type->kind == TYPE_KIND_ARRAY || type->kind == TYPE_KIND_FUNCTION) {
	return 0;
    }

    /* there must be no incomplete type */
    /* xxx: this may not be good enough to guarantee that the left operand
       has no incomplete type in its subexpressions. Umm. */
    if (type->size < 0) {
	return 0;
    }

    return 1;
}

static int
is_string_literal (struct AST *ast)
{
    while (!is_AST_LEAF ((struct AST const *) ast))
	ast = ast->u.child [0];
    return (ast->ast_type == AST_STRING);
}

static int
is_const_qualified_in_field (struct Type *type)
{
    int    i, num;
    struct StructUnionField **field;

    assert (type->kind == TYPE_KIND_STRUCT || type->kind == TYPE_KIND_UNION);

    num   = type->u.t_struct_union.num;
    field = type->u.t_struct_union.field;

    /* check all members */
    for (i = 0; i < num; i++) {
	/* check this member's qualifier */
	if ((field [i]->type->kind == TYPE_KIND_QUAL)
	    && (field [i]->type->u.t_qual.qualifier & TYPE_QUAL_CONST)) {
	    return 1;
	}
	/* check submembers recursively */
	if ((field [i]->type->kind == TYPE_KIND_STRUCT)
	    || (field [i]->type->kind == TYPE_KIND_UNION)) {
	    if (is_const_qualified_in_field (field [i]->type)) {
		return 1;
	    }
	}
    }

    /* const field not found */
    return 0;
}

static int
is_void_pointer_type (struct Type *type)
{
    /* true if pointer->(qual->)void */
    if (type->kind != TYPE_KIND_POINTER) {
	return 0;
    }
    type = type->u.t_pointer.type;
    if (type->kind == TYPE_KIND_QUAL) {
	type = type->u.t_qual.type;
    }
    return ((type->kind == TYPE_KIND_PRIM)
	    && (type->u.t_prim.ptype == PRIM_TYPE_VOID));
}

static int
is_char_pointer_type (struct Type *type)
{
    /* true if (qual->)pointer->(qual->)char */
    if (type->kind == TYPE_KIND_QUAL) {
	type = type->u.t_qual.type;
    }
    if (type->kind != TYPE_KIND_POINTER)
	return 0;
    type = type->u.t_pointer.type;
    if (type->kind == TYPE_KIND_QUAL) {
	type = type->u.t_qual.type;
    }
    return ((type->kind == TYPE_KIND_PRIM)
	    && (type->u.t_prim.ptype == PRIM_TYPE_CHAR));
}

static int
is_char_array_type (struct Type *type)
{
    /* true if array ->(qual->)char */
    if (type->kind != TYPE_KIND_ARRAY)
	return 0;
    type = type->u.t_array.elem_type;
    if (type->kind == TYPE_KIND_QUAL) {
	type = type->u.t_qual.type;
    }
    return ((type->kind == TYPE_KIND_PRIM)
	    && (type->u.t_prim.ptype == PRIM_TYPE_CHAR));
}

static struct AST *
find_non_direct_parent (struct AST *ast, int *nth_p)
{
    struct AST *ast_parent = ast->parent;

    (*nth_p) = ast->nth;

    while (1) {
	assert (ast_parent != NULL);

	switch (ast_parent->ast_type) {
	case AST_expression_direct:
	case AST_assignment_expression_direct:
	case AST_conditional_expression_direct:
	case AST_logical_or_expression_direct:
	case AST_logical_and_expression_direct:
	case AST_inclusive_or_expression_direct:
	case AST_exclusive_or_expression_direct:
	case AST_and_expression_direct:
	case AST_equality_expression_direct:
	case AST_relational_expression_direct:
	case AST_shift_expression_direct:
	case AST_additive_expression_direct:
	case AST_multiplicative_expression_direct:
	case AST_cast_expression_direct:
	case AST_unary_expression_direct:
	case AST_postfix_expression_direct:
	case AST_primary_expression_paren:
	case AST_primary_expression_ID:
	    (*nth_p)   = ast_parent->nth;
	    ast_parent = ast_parent->parent;
	    continue;
	default:
	    return ast_parent;
	}
    }
}

static int
is_type_assignable (struct AST *ast_right, struct Type *type_left,
		    struct Type *type_right)
{
    if (type_left->kind == TYPE_KIND_QUAL) {
	type_left = type_left->u.t_qual.type;
    }

    if (type_right->kind == TYPE_KIND_QUAL) {
	type_right = type_right->u.t_qual.type;
    }

    if (is_arithmetic_type (type_left) && is_arithmetic_type (type_right)) {
	return 1;
    }

    if (is_type_equivalent (type_left, type_right)
	&& ((type_right->kind == TYPE_KIND_STRUCT)
	    || (type_right->kind == TYPE_KIND_UNION))) {
	return 1;
    }

    if (is_void_pointer_type (type_left)
	&& (type_right->kind == TYPE_KIND_POINTER)) {
	return 1;
    }

    if (is_void_pointer_type (type_right)
	&& (type_left->kind == TYPE_KIND_POINTER)) {
	return 1;
    }

    if ((type_left->kind == TYPE_KIND_POINTER)
	&& is_constant_zero (ast_right)) {
	return 1;
    }

    if ((type_left->kind == TYPE_KIND_POINTER)
	&& is_type_equivalent (type_left, type_right)) {
	return 1;
    }

    return 0;
}

static int
is_constant_zero (struct AST *ast)
{
    /* not implemented yet; we must evaluate ast and check it is equal to 0. */
    return is_integral_ast (ast) && 1;
}

static struct Type *
type_analyze_add (struct AST *ast0, struct AST *ast1)
{
    struct Type *ret_type;

    if (is_arithmetic_ast (ast0) && is_arithmetic_ast (ast1)) {
	ret_type = type_usual_arithmetic_conversion (ast0, ast1);
    } else if ((get_type (ast0)->kind == TYPE_KIND_POINTER)
	       && !is_void_pointer_type (get_type (ast0))
	       && is_integral_ast (ast1)) {
	ret_type = get_type (ast0);
    } else if ((get_type (ast1)->kind == TYPE_KIND_POINTER)
	       && !is_void_pointer_type (get_type (ast1))
	       && is_integral_ast (ast0)) {
	ret_type = get_type (ast1);
    } else {
	xci_error (ast0, "invalid type of binary + operand");
    }
    return ret_type;
}

static struct Type *
type_analyze_sub (struct AST *ast0,struct AST *ast1)
{
    struct Type *ret_type;

    if (is_arithmetic_ast (ast0) && is_arithmetic_ast (ast1)) {
	ret_type = type_usual_arithmetic_conversion (ast0, ast1);
    } else if ((get_type (ast0)->kind == TYPE_KIND_POINTER)
	       && !is_void_pointer_type (get_type (ast1))
	       && is_integral_ast (ast1)) {
	ret_type = get_type (ast0);
    } else if ((get_type (ast0)->kind == TYPE_KIND_POINTER)
	       && !is_void_pointer_type (get_type (ast1))
	       && is_type_equivalent (get_type (ast0), get_type (ast1))) {
	ret_type = types.t_int;
	/* this is to be ptrdiff_t in stddef.h */
    } else {
	xci_error (ast0, "invalid type of binary - operand");
    }
    return ret_type;
}

static struct Type *
type_analyze_mult_div (struct AST *ast0, struct AST *ast1)
{
    struct Type *ret_type;

    if (!is_arithmetic_ast (ast0)) {
	xci_error (ast0, "invalid type of left operand"
		   " in binary * or /; arithmetic type required");
    }
    if (!is_arithmetic_ast (ast1)) {
	xci_error (ast1, "invalid type of right operand"
		   " in binary * or /; arithmetic type required");
    }

    ret_type =  type_usual_arithmetic_conversion (ast0, ast1);
    return ret_type;
}

static struct Type *
type_analyze_mod (struct AST *ast0, struct AST *ast1)
{
    struct Type *ret_type;

    if (!is_integral_ast (ast0)) {
	xci_error (ast0, "invalid type of left operand"
		   " in %; integral type required");
    }
    if (!is_integral_ast (ast1)) {
	xci_error (ast1, "invalid type of right operand"
		   " in %; integral type required");
    }
    ret_type = type_usual_arithmetic_conversion (ast0, ast1);
    return ret_type;
}

static struct Type *
type_analyze_shift (struct AST *ast0, struct AST *ast1)
{
    struct Type *ret_type;

    if (!is_integral_ast (ast0)) {
	xci_error (ast0, "invalid type in the left operand"
		   " of << or >>; integral type required");
    }
    if (!is_integral_ast (ast1)) {
	xci_error (ast1, "invalid type in the right operand"
		   " of << or >>; integral type required");
    }
    ret_type = type_integral_promotion (get_type (ast0));
    return ret_type;
}

static struct Type *
type_analyze_bop (struct AST *ast0, struct AST *ast1)
{
    struct Type *ret_type;

    /* each operand must have integral type */
    if (!is_integral_ast (ast0)) {
	xci_error (ast0, "invalid type in the left operand of &, |, or ^; "
		   "integral type required");
    }
    if (!is_integral_ast (ast1)) {
	xci_error (ast1, "invalid type in the right operand of &, |, or ^; "
		   "integral type required");
    }
    ret_type = type_usual_arithmetic_conversion (ast0, ast1);
    return ret_type;
}

static void
type_analyze_arg (struct AST *ast, struct Type *type)
{
    struct AST *ast_arg_exp_list;
    struct AST **actual_arg_ast;
    int        actual_arg_num;
    int        i;

    assert ((ast->ast_type == AST_postfix_expression_funcall)
	    || (ast->ast_type == AST_postfix_expression_funcall_noarg));
    assert (type->kind == TYPE_KIND_FUNCTION);

    if (ast->ast_type == AST_postfix_expression_funcall_noarg) {
	/* no actual arguments, so just check the number of formal arguments */
	if (type->u.t_function.arg_num != 0) {
	    xci_error (ast, "too few arguments for function");
	}
	return;
    }

    assert (ast->ast_type == AST_postfix_expression_funcall);

    /* count the number of actual arguments */
    /* count the last AST_argument_expression_list_single first */
    actual_arg_num = 1;
    for (ast_arg_exp_list = ast->u.child [1];
	 ast_arg_exp_list->ast_type == AST_argument_expression_list_pair;
	 ast_arg_exp_list = ast_arg_exp_list->u.child [0]) {
	actual_arg_num++;
    }
    assert (ast_arg_exp_list->ast_type == AST_argument_expression_list_single);

    /* collect actual argument types */
    actual_arg_ast  = emalloc (sizeof (struct AST *)  * actual_arg_num);

    for (i = 0; i < actual_arg_num; i++) {
	switch (ast_arg_exp_list->ast_type) {
	case AST_argument_expression_list_single:
	    actual_arg_ast  [i] = ast_arg_exp_list->u.child [0];
	    break;
	case AST_argument_expression_list_pair:
	    actual_arg_ast  [i] = ast_arg_exp_list->u.child [1];
	    break;
	default:
	    assert (0);
	    break;
	}
	ast_arg_exp_list = ast_arg_exp_list->parent;
    }
    assert (ast_arg_exp_list == ast);

    /* compare the number of formal arguments and that of actual ones. */
    if (type->u.t_function.arg_num > actual_arg_num) {
	xci_error (ast, "too few arguments for function");
    }
    if (!type->u.t_function.is_ellipsis
	&& (type->u.t_function.arg_num < actual_arg_num)) {
	xci_error (ast, "too many arguments for function");
    }

    /* do default argument promotion if old-style arguments */
    if (type->u.t_function.is_old_type) {
	for (i = 0; i < type->u.t_function.arg_num; i++) {
	    actual_arg_ast [i]->prom_type
		= type_argument_promotion_ast (actual_arg_ast [i]);
	}
    }

    /* do default argument promotion for ellipsis  arguments */
    if (type->u.t_function.is_ellipsis) {
	for (i = type->u.t_function.arg_num; i < actual_arg_num; i++) {
	    actual_arg_ast [i]->prom_type
		= type_argument_promotion_ast (actual_arg_ast [i]);
	}
    }

    /* check each actual argument is assignable to formal one */
    for (i = 0; i < type->u.t_function.arg_num; i++) {
	if (!is_type_assignable (actual_arg_ast [i],
				 type->u.t_function.arg_type [i],
				 get_type (actual_arg_ast [i]))) {
#if 1 /* debug */
	    type_dump (type->u.t_function.arg_type [i], 0);
	    type_dump (get_type (actual_arg_ast [i]), 0);
#endif
	    xci_error_args (actual_arg_ast [i],
			    "type mismatch in argument %d", i);
	}
    }
}

static int
type_analyze_initializer (struct AST *ast, struct Type *type)
{
    struct AST *ast_init_list;
    int num;

    assert (type->kind != TYPE_KIND_QUAL);
    assert ((ast->ast_type == AST_initializer_assign)
	    || (ast->ast_type == AST_initializer_list)
	    || (ast->ast_type == AST_initializer_list_comma));

    if (ast->ast_type == AST_initializer_assign) {
	type_analyze_exp (ast->u.child [0]);
	if (is_type_assignable (ast->u.child [0], 
				type, get_type (ast->u.child [0]))) {
	    return 0;
	} else if (is_char_array_type (type)
		   && is_string_literal (ast->u.child [0])) {
	    struct AST *ast_string =ast->u.child [0];
	    while (ast_string->ast_type != AST_STRING)
		ast_string = ast_string->u.child [0];
	    return strlen (ast_string->u.string) + 1;
	} else {
	    xci_error (ast->u.child [0], "invalid initializer: "
		       "single expression must be one of the following: \n"
		       "(1) pointer, arithmetic of assignable type, \n"
		       "(2) structure, or union of the same type, \n"
		       "(3) string literal for char array.");
	}
    }

    assert ((ast->ast_type == AST_initializer_list)
	    || (ast->ast_type == AST_initializer_list_comma));

    /* first search the bottom */
    num = 1; /* count the last AST_initializer_list_single first */
    for (ast_init_list = ast->u.child [0];
	 ast_init_list->ast_type != AST_initializer_list_single;
	 ast_init_list = ast_init_list->u.child [0])
	num++; /* empty body */
    
    assert (ast_init_list->ast_type == AST_initializer_list_single);
    switch (type->kind) {
    case TYPE_KIND_PRIM:
    case TYPE_KIND_ENUM:
	if (num > 1) {
	    xci_error (ast, "too many initializers");
	}
	(void) type_analyze_initializer (ast_init_list->u.child [0], type);
	break;

    case TYPE_KIND_UNION:
	if (num > 1) {
	    xci_error (ast, "too many initializers");
	}
	(void) type_analyze_initializer (ast_init_list->u.child [0], 
					 type->u.t_struct_union.field [0]->type);
	break;

    case TYPE_KIND_STRUCT:
    case TYPE_KIND_ARRAY:
	type_analyze_initializer_list (ast_init_list, type, &num);
	break;

    case TYPE_KIND_FUNCTION:
    case TYPE_KIND_POINTER:
    case TYPE_KIND_QUAL:
    default:
	assert (0);
	break;
    }

    return num;
}

static struct AST *
type_analyze_initializer_list (struct AST *ast, struct Type *type, int *num_p)
{
    struct StructUnionField **field;
    struct Type *member_type;
    int num_field;
    int i;

    assert ((ast->ast_type == AST_initializer_list_single)
	    || (ast->ast_type == AST_initializer_list_pair));

    if (type->kind == TYPE_KIND_QUAL) {
	type = type->u.t_qual.type;
    }

    if (type->kind == TYPE_KIND_STRUCT) {
	num_field = type->u.t_struct_union.num;
	field = type->u.t_struct_union.field;
    } else if (type->kind == TYPE_KIND_ARRAY) {
	num_field = type->u.t_array.num;
	member_type = type->u.t_array.elem_type;
    } else {
	assert (0);
    }

    for (i = 0;; i++) {
	struct AST *ast_init;

	if ((ast->ast_type != AST_initializer_list_single)
	    && (ast->ast_type != AST_initializer_list_pair)) {
	    /* no more initializer */
	    if (num_p != NULL)
		*num_p = i;    /* for incomplete array */
	    return ast;
	}

	switch (ast->ast_type) {
	case AST_initializer_list_single: ast_init = ast->u.child [0]; break;
	case AST_initializer_list_pair:   ast_init = ast->u.child [1]; break;
	default:                          assert (0);                  break;
	}

	if (type->kind == TYPE_KIND_STRUCT) {
	    while ((i < num_field) && (field [i]->name == NULL)) {
		i++; /* skip unnamed bit-field members */
	    }
    	}
	if (i >= num_field && num_field >= 0) {
	    /* no more member to be initialized in this (sub) aggregate */
	    return ast;
	}

	if (type->kind == TYPE_KIND_STRUCT) {
	    member_type = field [i]->type;
	}

	if (member_type->kind == TYPE_KIND_QUAL) {
	    member_type = member_type->u.t_qual.type;
	}

	if (((member_type->kind == TYPE_KIND_STRUCT)
	     || (member_type->kind == TYPE_KIND_ARRAY))
	    && (ast_init->ast_type == AST_initializer_assign)) {
	    /* in the case that braces are omitted for subaggregate */
	    ast = type_analyze_initializer_list (ast, member_type, NULL);
	    /* ast = ast->parent is already done in the above function,
	       so we just do continue here. */
	    continue;
	} else {
	    (void) type_analyze_initializer (ast_init, member_type);
    	}

	ast = ast->parent;
    }
}

static struct Type *
get_type_no_skip (struct AST *ast)
{
    return (ast->prom_type != NULL) ? ast->prom_type : ast->type;
}

static int
get_type_qual  (struct AST *ast)
{
    struct Type *type;
    type = get_type (ast);
    if (type->kind == TYPE_KIND_QUAL) {
	return type->u.t_qual.qualifier;
    } else {
	return TYPE_QUAL_NONE;
    }
}

static int
is_function_designator_id (struct AST *ast)
{
    assert (ast->ast_type == AST_IDENTIFIER);
    
    return ((ast->parent->ast_type == AST_primary_expression_ID)
	    && (ast->parent->parent->ast_type == AST_postfix_expression_direct)
	    && (ast->parent->parent->nth == 0)
	    && ((ast->parent->parent->parent->ast_type
		 == AST_postfix_expression_funcall)
		|| (ast->parent->parent->parent->ast_type ==
		    AST_postfix_expression_funcall_noarg)));
}

/*
  incomplete array is completed by initializer like the following:
      int a[] = {10, 20, 30, 40};
  this function does this!
 */
static void
complete_array_type (struct Symbol *sym, int num)
{
    int elem_size;
    struct Type *type;

    assert (sym != NULL);
    assert (sym->type != NULL);

    /* first check if the given type is to be completed */
    if (sym->type->kind != TYPE_KIND_ARRAY)
	return;
    if (!(sym->status & SYM_STAT_INCOMPLETE))
	return;
    assert (sym->type->size < 0);
    assert (sym->addr == NULL);

    /* then complete it */
    sym->status &= ~SYM_STAT_INCOMPLETE;
    elem_size = get_type_size (sym->type->u.t_array.elem_type);
    /* sym->type may be shared, so duplicate it */
    type = emalloc (sizeof (struct Type));
    *type = *sym->type;
    type->size = elem_size * num;
    type->u.t_array.num = num;
    sym->type = type;
}
/* ---------------------------------------------------------------------- */
