/*
    XCI: eXperimental C Interpreter for ANSI C
    -- Copyright (c) 2000-2001, gondow@jaist.ac.jp
    $Id: typedef.c,v 1.2 2002/06/06 14:47:33 gondow Exp $
*/ 

#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <string.h>
#include "xci.h"
#include "AST.h"
#include "type.h"
#include "typedef.h"
/* ---------------------------------------------------------------------- */
static void typedef_free (struct TypedefName *typedef_name);
static char *typedef_find_name (struct AST *ast);
static void typedef_dump ();
/* ---------------------------------------------------------------------- */
static struct TypedefTable typedef_table = {NULL, -1, 0};
/* ---------------------------------------------------------------------- */

/*
  notice: we do not check duplicated definitions here, because 
  even legal code can introduce duplicated entries in the typedef_table.
  e.g. in the following code, 'foo' and 'x' are doubly entried.

  extern foo ();
  int foo (x, y)
  char x;
  { return 0; }

  we assume static analysis checks them later.
 */

void
typedef_entry (char const *name, enum NameKind kind, int level_adjust)
{
    struct TypedefName *entry;
    int    level = typedef_table.sp - level_adjust;
    /* basically, level_adjust = 0. */

    assert (name != NULL);
    entry = emalloc (sizeof (struct TypedefName));
    entry->name = name;
    entry->kind = kind;
    entry->next = typedef_table.stack [level];
    typedef_table.stack [level] = entry;

#if 0 /* debug */
    if (!strcmp (name, "last")) {
	typedef_dump ();
    }
#endif
}

enum NameKind
typedef_kind (char const *name)
{
    struct TypedefName *p;
    int    i;

    for (i = typedef_table.sp; i >= 0; i--) {
	for (p = typedef_table.stack [i]; p != NULL; p = p->next) {
	    if (!strcmp (name, p->name)) {
		return p->kind; /* found */
	    }
	}
    }
    return NK_UNKNOWN; /* not found */
}

void
typedef_entry_by_declarator (struct AST *ast, int is_typedef)
{
    char *name;

    assert (ast->ast_type == AST_declarator_direct);
    name = typedef_find_name (ast->u.child [1]);
    if (is_typedef) {
	typedef_entry (name, NK_TYPEDEF_NAME, 0);
    } else{
	typedef_entry (name, NK_IDENTIFIER, 0);
    }
}

void
typedef_entry_by_function_declarator (struct AST *ast)
{
    char *name;

    assert (ast->ast_type == AST_function_declarator_direct);
    name = typedef_find_name (ast->u.child [1]);
    typedef_entry (name, NK_IDENTIFIER, 1);
    /*
      here is a tiny hack; level_adjust = 1. this is required because
      function name is registered after parameter registered.
     */
}

void
typedef_entry_by_enumerator (struct AST *ast)
{
    assert (ast->ast_type == AST_IDENTIFIER);
    typedef_entry (ast->u.string, NK_ENUMERATION_CONSTANT, 0);
}

void
typedef_entry_by_identifier_list (struct AST *ast)
{
    assert (ast->ast_type == AST_IDENTIFIER);
    typedef_entry (ast->u.string, NK_IDENTIFIER, 0);
}

void
typedef_scope_begin ()
{
    assert (typedef_table.sp < typedef_table.size);

    if (typedef_table.sp >= typedef_table.size -1) {
	typedef_table.stack = 
	    erealloc (typedef_table.stack,
		      sizeof (struct TypedefName *) * ++typedef_table.size);
    }
    typedef_table.stack [++typedef_table.sp] = NULL;
}

void
typedef_scope_end ()
{
    assert (typedef_table.sp >= 0);
    typedef_free (typedef_table.stack [typedef_table.sp]);
    typedef_table.sp--;
}

/* ---------------------------------------------------------------------- */

static void
typedef_free (struct TypedefName *typedef_name)
{
    struct TypedefName *next;

    if (typedef_name == NULL) return;

    next = typedef_name->next;
    efree (typedef_name);
    typedef_free (next);
}

/*
  This function finds an identifier in a given 'declarator', or
  NULL if not found. E.g., 'foo' is returned for the declarator '*(*foo)()'.
 */
static char *
typedef_find_name (struct AST *ast)
{
    switch (ast->ast_type) {
    case AST_direct_declarator_paren:
    case AST_direct_declarator_array:
    case AST_direct_declarator_func_new:
    case AST_direct_declarator_func_old:
    case AST_direct_function_declarator_func_new:
    case AST_direct_function_declarator_func_old:
	return typedef_find_name (ast->u.child [0]);

    case AST_declarator_direct:
    case AST_function_declarator_direct:
	return typedef_find_name (ast->u.child [1]);

    case AST_direct_declarator_id:
	return ast->u.child [0]->u.string;

    default:
	assert (0);
	return NULL;
    }
}

/* for debugging */
static void
typedef_dump ()
{
    struct TypedefName *p;
    int    i;

    for (i = typedef_table.sp; i >= 0; i--) {
	printf ("===== %d =====\n", i);
	for (p = typedef_table.stack [i]; p != NULL; p = p->next) {
	    printf ("%d: %s\n", p->kind, p->name);
	}
    }
}
