/*
    XCI: eXperimental C Interpreter for ANSI C
    -- Copyright (c) 2000-2001, gondow@jaist.ac.jp
    $Id: vm.c,v 1.3 2002/07/09 11:43:46 gondow Exp $
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>

#include "xci.h"
#include "AST.h"
#include "type.h"
#include "vm.h"

/* ---------------------------------------------------------------------- */
static void   init    (int stack_size);

static void   push_c  (char c);
static char   pop_c   (void);
static char   top_c   (void);
static void   push_i  (int i);
static int    pop_i   (void);
static int    top_i   (void);
static void   push_d  (double d);
static double pop_d   (void);
static double top_d   (void);
static void   push_p  (void *p);
static void*  pop_p   (void);
static void*  top_p   (void);

static void   push_b  (void *from_addr, int size);
static void   pop_b   (void *to_addr,   int size);
static void   top_b   (void *to_addr,   int size);

static void   add_sp  (int size);
static void   sub_sp  (int size);
static void*  get_sp  (void);
static void   set_sp  (void *sp);
static void*  get_bp  (void);
static void   set_bp  (void *bp);

static void   set_pc (void *ast, int nth);

static void   dump    (void);
static int    size    (struct Type *type);

static void   c_to_i  (void);
static void   i_to_c  (void);
static void   i_to_d  (void);
static void   d_to_i  (void);
static void   p_to_i  (void);
static void   i_to_p  (void);

static void   add_i   (void);
static void   sub_i   (void);
static void   mult_i  (void);
static void   div_i   (void);
static void   mod_i   (void);

static void   add_d   (void);
static void   sub_d   (void);
static void   mult_d  (void);
static void   div_d   (void);

static void   l_not   (void);

static void   less_i  (void);
static void   leq_i   (void);
static void   more_i  (void);
static void   meq_i   (void);
static void   eq_i    (void);
static void   neq_i   (void);

static void   less_d  (void);
static void   leq_d   (void);
static void   more_d  (void);
static void   meq_d   (void);
static void   eq_d    (void);
static void   neq_d   (void);

static void   l_sh    (void);
static void   r_sh    (void);
static void   r_ush   (void);
static void   b_and   (void);
static void   b_or    (void);
static void   b_xor   (void);
static void   b_comp  (void);

#if 0
static void   jump    (void);
static void   jump_if (void);

static void   call    (void);
static void   ret     (void);
#endif

static void check_stack_overflow  (int size);
static void check_stack_underflow (int size);
/* ---------------------------------------------------------------------- */
struct VM vm; /* virtual machine interface */
/* ---------------------------------------------------------------------- */
void
vm_init (int stack_size)
{
    vm.init = init;
    vm.init (stack_size);
}

/* ---------------------------------------------------------------------- */
#if 0	 /* for unit test */
void *
emalloc (int size)
{ 
    void *p = malloc (size);
    if (p == NULL) perror ("malloc");
    return p;
}

int
main (void)
{
    vm_init (1024);
    vm.push_i (-5);
    vm.push_i (1);
    vm.r_ush ();
    vm.dump ();
    return 0;
}
#endif
/* ---------------------------------------------------------------------- */

static void
init (int stack_size)
{
    vm.status = VM_STAT_RUNNING;

    /* stack initialization */
    vm.stack_size = stack_size;
    vm.stack = emalloc (vm.stack_size);
    vm.stack_type = emalloc (vm.stack_size);
    vm.sp    = vm.stack;
    vm.bp    = vm.stack;

    vm.size  = size;
    vm.dump  = dump;

    /* stack operators */
    vm.push_c  = push_c;
    vm.pop_c   = pop_c;
    vm.top_c   = top_c;
    vm.push_i  = push_i;
    vm.pop_i   = pop_i;
    vm.top_i   = top_i;
    vm.push_d  = push_d;
    vm.pop_d   = pop_d;
    vm.top_d   = top_d;
    vm.push_p  = push_p;
    vm.pop_p   = pop_p;
    vm.top_p   = top_p;

    vm.push_b  = push_b;
    vm.pop_b   = pop_b;
    vm.top_b   = top_b;

    vm.add_sp  = add_sp;
    vm.sub_sp  = sub_sp;
    vm.get_sp  = get_sp;
    vm.set_sp  = set_sp;

    vm.get_bp  = get_bp;
    vm.set_bp  = set_bp;

    vm.set_pc  = set_pc;

    /* type casting */
    vm.c_to_i = c_to_i;
    vm.i_to_c = i_to_c;
    vm.i_to_d = i_to_d;
    vm.d_to_i = d_to_i;
    vm.p_to_i = p_to_i;
    vm.i_to_p = i_to_p;

    /* arithmetic operators */
    vm.add_i  = add_i;
    vm.sub_i  = sub_i;
    vm.mult_i = mult_i;
    vm.div_i  = div_i;
    vm.mod_i  = mod_i;

    vm.add_d  = add_d;
    vm.sub_d  = sub_d;
    vm.mult_d = mult_d;
    vm.div_d  = div_d;

    vm.l_not  = l_not;

    vm.less_i = less_i;
    vm.leq_i  = leq_i;
    vm.more_i = more_i;
    vm.meq_i  = meq_i;
    vm.eq_i   = eq_i;
    vm.neq_i  = neq_i;

    vm.less_d = less_d;
    vm.leq_d  = leq_d;
    vm.more_d = more_d;
    vm.meq_d  = meq_d;
    vm.eq_d   = eq_d;
    vm.neq_d  = neq_d;

    vm.l_sh   = l_sh;
    vm.r_sh   = r_sh;
    vm.r_ush  = r_ush;
    vm.b_and  = b_and;
    vm.b_or   = b_or;
    vm.b_xor  = b_xor;
    vm.b_comp  = b_comp;

#if 0
    /* jump */
    vm.jump   = jump;
    vm.jump_if = jump_if;

    /* subroutine call/return */
    vm.call   = call;
    vm.ret    = ret;
#endif

    /* limits */
    vm.VM_CHAR_MAX   = CHAR_MAX;
    vm.VM_CHAR_MIN   = CHAR_MIN;
    vm.VM_INT_MAX    = INT_MAX;
    vm.VM_INT_MIN    = INT_MIN;
    vm.VM_LONG_MAX   = INT_MAX;
    vm.VM_LONG_MIN   = INT_MIN;
    vm.VM_SCHAR_MAX  = SCHAR_MAX;
    vm.VM_SCHAR_MIN  = SCHAR_MIN;
    vm.VM_SHRT_MAX   = INT_MAX;
    vm.VM_SHRT_MIN   = INT_MIN;
    vm.VM_UCHAR_MAX  = UCHAR_MAX;
    vm.VM_UINT_MAX   = UINT_MAX;
    vm.VM_ULONG_MAX  = UINT_MAX;
    vm.VM_USHRT_MAX  = UINT_MAX;

    /* sizes */
    vm.VM_SIZEOF_CHAR       = sizeof (char);
    vm.VM_SIZEOF_INT        = 
    vm.VM_SIZEOF_SHORT      = 
    vm.VM_SIZEOF_LONG       = sizeof (int);
    vm.VM_SIZEOF_FLOAT      = 
    vm.VM_SIZEOF_DOUBLE     = 
    vm.VM_SIZEOF_LONGDOUBLE = sizeof (double);
    vm.VM_SIZEOF_VOID_P     = sizeof (void *);
}

static void
dump   (void)
{
    void *p = vm.stack;

    while (p < vm.sp) {
	fprintf (stderr, "%p: %4d: ", p, p - vm.stack);
	switch (vm.stack_type [p - vm.stack]) {
	case ST_CHAR:
	    fprintf (stderr, "char   : %c\n", *(char *)p);
	    p += vm.VM_SIZEOF_CHAR;
	    break;
	case ST_INT:
	    fprintf (stderr, "int    : %d\n", *(int *)p);
	    p += vm.VM_SIZEOF_INT;
	    break;
	case ST_DOUBLE:
	    fprintf (stderr, "double : %f\n", *(double *)p);
	    p += vm.VM_SIZEOF_DOUBLE;
	    break;
	case ST_POINTER:
	    fprintf (stderr, "pointer: %p (%d)\n",
		     *(void **)p, (int)*(void **)p);
	    p += vm.VM_SIZEOF_VOID_P;
	    break;
	case ST_STRING:
	    fprintf (stderr, "string : %s\n", (char *)p);
	    p += strlen ((char *)p) + 1;
	    break;
	case ST_OTHER:
	    fprintf (stderr, "other  : 0x%x\n", (int)*(char *)p);
	    p += vm.VM_SIZEOF_CHAR;
	    break;
	default:
	    fprintf (stderr, "       : 0x%x\n", (int)*(char *)p);
	    p += vm.VM_SIZEOF_CHAR;
	    break;
	}
    }

    fprintf (stderr, "\nsp = %p (%d), bp = %p (%d), pc = %p, %d\n",
	     vm.sp, vm.sp - vm.stack, vm.bp, vm.bp - vm.stack, vm.pc.ast,
	     vm.pc.nth);
}

static int
size    (struct Type *type)
{
    switch (type->kind) {
    case TYPE_KIND_PRIM:
	switch (type->u.t_prim.ptype) {
	case PRIM_TYPE_CHAR:   return vm.VM_SIZEOF_CHAR;
	case PRIM_TYPE_INT:    return vm.VM_SIZEOF_INT;
	case PRIM_TYPE_FLOAT:  return vm.VM_SIZEOF_FLOAT;
	case PRIM_TYPE_DOUBLE: return vm.VM_SIZEOF_DOUBLE;
	case PRIM_TYPE_VOID:   return -1;
	default:               assert (0); return -1;
	}
    case TYPE_KIND_POINTER:
	return vm.VM_SIZEOF_VOID_P;
    default:
	assert (0);
	return -1;
    }
}

static void
check_stack_overflow (int size)
{
    if (vm.sp + size > vm.stack + vm.stack_size) {
	vm.dump ();
	xci_error_noast_args ("stack overflow: sp=%p, base=%p, offset=%d\n",
			      vm.sp, vm.stack, vm.sp - vm.stack);
    }
}

static void
check_stack_underflow (int size)
{
    if (vm.sp - size < vm.stack) {
	xci_error_noast_args ("stack underflow: sp=%p, base=%p, offset=%d\n",
			      vm.sp, vm.stack, vm.sp - vm.stack);
    }
}

static void
push_c  (char c)
{
    vm.stack_type [vm.sp - vm.stack] = ST_CHAR;
    check_stack_overflow (vm.VM_SIZEOF_CHAR);
    *(char *) vm.sp = c;
    vm.sp += vm.VM_SIZEOF_CHAR;
}

static char
pop_c   (void)
{
    check_stack_underflow (vm.VM_SIZEOF_CHAR);
    vm.sp -= vm.VM_SIZEOF_CHAR;
    return *(char *) vm.sp;
}

static char
top_c   (void)
{
    return *(char *) (vm.sp - vm.VM_SIZEOF_CHAR);
}

static void
push_i  (int i)
{
    vm.stack_type [vm.sp - vm.stack] = ST_INT;
    check_stack_overflow (vm.VM_SIZEOF_INT);
    *(int *) vm.sp = i;
    vm.sp += vm.VM_SIZEOF_INT;
}

static int
pop_i   (void)
{
    check_stack_underflow (vm.VM_SIZEOF_INT);
    vm.sp -= vm.VM_SIZEOF_INT;
    return *(int *) vm.sp;
}

static int
top_i   (void)
{
    return *(int *) (vm.sp - vm.VM_SIZEOF_INT);
}

static void
push_d  (double d)
{
    vm.stack_type [vm.sp - vm.stack] = ST_DOUBLE;
    check_stack_overflow (vm.VM_SIZEOF_DOUBLE);
    *(double *) vm.sp = d;
    vm.sp += vm.VM_SIZEOF_DOUBLE;
}

static double
pop_d   (void)
{
    check_stack_underflow (vm.VM_SIZEOF_DOUBLE);
    vm.sp -= vm.VM_SIZEOF_DOUBLE;
    return *(double *) vm.sp;
}

static double
top_d   (void)
{
    return *(double *) (vm.sp - vm.VM_SIZEOF_DOUBLE);
}

static void
push_p  (void *p)
{
    vm.stack_type [vm.sp - vm.stack] = ST_POINTER;
    check_stack_overflow (vm.VM_SIZEOF_VOID_P);
    *(void **) vm.sp = p;
    vm.sp += vm.VM_SIZEOF_VOID_P;
}

static void *
pop_p   (void)
{
    check_stack_underflow (vm.VM_SIZEOF_VOID_P);
    vm.sp -= vm.VM_SIZEOF_VOID_P;
    return *(void **) vm.sp;
}

static void *
top_p   (void)
{
    return *(void **) (vm.sp - vm.VM_SIZEOF_VOID_P);
}

static void
push_b  (void *from_addr, int size)
{
    vm.stack_type [vm.sp - vm.stack] = ST_OTHER;
    check_stack_overflow (size);
    assert (from_addr != NULL);
    memcpy (vm.sp, from_addr, size);
    vm.sp += size;
}

static void
pop_b   (void *to_addr,   int size)
{
    check_stack_underflow (size);
    vm.sp -= size;
    assert (to_addr != NULL);
    memcpy (to_addr, vm.sp, size);
}

static void
top_b   (void *to_addr,   int size)
{
    check_stack_underflow (size);
    assert (to_addr != NULL);
    memcpy (to_addr, vm.sp - size, size);
}

static void
add_sp  (int size)
{
    vm.stack_type [vm.sp - vm.stack] = ST_OTHER;
    check_stack_overflow (size);
    vm.sp += size;
}

static void
sub_sp  (int size)
{
    check_stack_underflow (size);
    vm.sp -= size;
}

static void*
get_sp  (void)
{
    return vm.sp;
}

static void
set_sp  (void *sp)
{
    if (sp < vm.stack || vm.stack + vm.stack_size < sp) {
	xci_error_noast_args ("set_sp: specified sp out of range: %p\n", sp);
    }
    vm.sp = sp;
}

static void *
get_bp  (void)
{
    return vm.bp;
}

static void
set_bp  (void *bp)
{
    if (bp < vm.stack || vm.stack + vm.stack_size < bp) {
	xci_error_noast_args ("set_bp: specified bp out of range: %p\n", bp);
    }
    vm.bp = bp;
}

static void
set_pc (void *ast, int nth)
{
    vm.pc.ast = ast;
    vm.pc.nth = nth;
}

static void
c_to_i  (void)
{
    push_i ((int)    pop_c ());
}

static void
i_to_c  (void)
{
    push_c ((char)   pop_i ());
}

static void
d_to_i  (void)
{
    push_i ((int)    pop_d ());
}

static void
i_to_d  (void)
{
    push_d ((double) pop_i ());
}

static void
p_to_i  (void)
{
    push_i ((int)    pop_p ());
}

static void
i_to_p  (void)
{
    push_p ((int *)  pop_i ());
}

static void
add_i   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 + i1);
}

/* it is incorrect to write push_i (pop_i () - pop_i ()) instead of this,
   since which pop_i is executed first is implementation-dependent. */
static void
sub_i   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 - i1);
}

static void
mult_i   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 * i1);
}

static void
div_i   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 / i1);
}

static void
mod_i   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 % i1);
}

static void
add_d   (void)
{
    double d1 = pop_d ();
    double d2 = pop_d ();
    push_d (d2 + d1);
}

static void
sub_d   (void)
{
    double d1 = pop_d ();
    double d2 = pop_d ();
    push_d (d2 - d1);
}

static void
mult_d  (void)
{
    double d1 = pop_d ();
    double d2 = pop_d ();
    push_d (d2 * d1);
}

static void
div_d   (void)
{
    double d1 = pop_d ();
    double d2 = pop_d ();
    push_d (d2 / d1);
}

static void
l_not  (void)
{
    push_i (!pop_i ());
}

static void
less_i  (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 < i1);
}

static void
leq_i   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 <= i1);
}

static void
more_i  (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 > i1);
}

static void
meq_i   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 >= i1);
}

static void
eq_i    (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 == i1);
}

static void
neq_i   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 != i1);
}

static void
less_d  (void)
{
    double d1 = pop_d ();
    double d2 = pop_d ();
    push_i (d2 < d1);
}

static void
leq_d   (void)
{
    double d1 = pop_d ();
    double d2 = pop_d ();
    push_i (d2 <= d1);
}

static void
more_d  (void)
{
    double d1 = pop_d ();
    double d2 = pop_d ();
    push_i (d2 > d1);
}

static void
meq_d   (void)
{
    double d1 = pop_d ();
    double d2 = pop_d ();
    push_i (d2 >= d1);
}

static void
eq_d    (void)
{
    double d1 = pop_d ();
    double d2 = pop_d ();
    push_i (d2 == d1);
}

static void
neq_d   (void)
{
    double d1 = pop_d ();
    double d2 = pop_d ();
    push_i (d2 != d1);
}

static void
l_sh    (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 << i1);
}

static void
r_sh    (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 >> i1);
}

static void
r_ush   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i ((unsigned) i2 >> i1);
}

static void
b_and   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 & i1);
}

static void
b_or    (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 | i1);
}

static void
b_xor   (void)
{
    int i1 = pop_i ();
    int i2 = pop_i ();
    push_i (i2 ^ i1);
}

static void
b_comp  (void)
{
    int i1 = pop_i ();
    push_i (~i1);
}

#if 0
static void
jump    (void)
{
    struct PC dest;
    dest.nth = pop_i ();
    dest.ast = pop_p ();
}

static void
jump_if (void)
{
    if (pop_i ()) 
	jump ();
}

static void
call    (void)
{
    /* not yet implemented */
}

static void
ret     (void)
{
    /* not yet implemented */
}
/* koreha test */
#endif
