 /*
    XCI: eXperimental C Interpreter for ANSI C
    -- Copyright (c) 2000-2001, gondow@jaist.ac.jp
    $Id: xci.l,v 1.3 2002/07/08 05:40:45 gondow Exp $
 */ 

%{
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include "xci.h"
#include "typedef.h"
/* xci.tab.h is generated by bison with '-d' option */
#include "xci.tab.h"
int  lineno=1;	          /* the currently processing line number */

static char *string_buf;  /* modified by only add_char_to_string (char c); */
static int  string_size;  /* modified by only add_char_to_string (char c); */

static void add_char_to_string   (char c);
static struct AST *process_int   (char const * const yytext, int const radix);
static struct AST *process_float (char const * const yytext);
static char escape_seq_to_char   (const char * const yytext);
%}

/*
  abbreviation definitions ('*' means 'optional') 
      D : decimal digit,   O : octal digit,      H : hexadecimal digit
      L : letter,          E : signed exponent
      IS: integer suffix*, FS: floating suffix*, WS: white spaces
      WS2: white spaces without newlines
 */

D      [0-9]
O      [0-7]
H      [a-fA-F0-9]
L      [a-zA-Z_]
E      [Ee][+-]?{D}+
IS     ([uU]?[lL]?)|([lL][uU])
FS     [fFlL]?
WS     [ \t\f\v\r\n]           
WS2    [ \t\f\v]           

%x comment string

%%
    /* Note: comments are required to be indented in the rule section */

    /* white spaces */
{WS}	    	    	{ if (yytext[0] == '\n') lineno++; } /* skipping */

    /* keywords */
auto	    	    	{ return AUTO;     }
break	     	        { return BREAK;    }
case	    	    	{ return CASE;     }
char	    	    	{ return CHAR;     }
const	     	        { return CONST;    }
continue    	    	{ return CONTINUE; }
default	       	        { return DEFAULT;  }
do    	      	      	{ return DO;       }
double	      	      	{ return DOUBLE;   }
else	    	    	{ return ELSE;     }
enum	    	    	{ return ENUM;     }
extern	      	      	{ return EXTERN;   }
float	     	     	{ return FLOAT;    }
for    	       	        { return FOR;      }
goto	    	    	{ return GOTO;     }
if    	      	      	{ return IF;       }
int    	       	        { return INT;      }
long	    	    	{ return LONG;     }
register    	    	{ return REGISTER; }
return	      	      	{ return RETURN;   }
short	     	     	{ return SHORT;    }
signed	      	      	{ return SIGNED;   }
sizeof	      	        { return SIZEOF;   }
static	      	      	{ return STATIC;   }
struct	      	        { return STRUCT;   }
switch	      	        { return SWITCH;   }
typedef	       	        { return TYPEDEF;  }
union	     	     	{ return UNION;    }
unsigned    	    	{ return UNSIGNED; }
void	    	    	{ return VOID;     }
volatile    	    	{ return VOLATILE; }
while	     	     	{ return WHILE;    }

    /* two or more character operators and separaters
    (in xci.y, these are referred to as literal string tokens ) */
"->"	    	    	{ return ARROW;    }
"=="	    	    	{ return EQ;       }
"!="	    	    	{ return NEQ;      }
"<="	    	    	{ return LESSEQ;   }
">="                    { return MOREEQ;   }
"++"	    	        { return INC;      }
"--"	    	    	{ return DEC;      }
"&&"	    	    	{ return AND;      }
"||"	    	    	{ return OR;       }
">>"	     	        { return RSHIFT;   }
"<<"	     	     	{ return LSHIFT;   }
"+="	    	    	{ return PLUSEQ;   }
"-="	    	    	{ return MINUSEQ;  }
"*="	    	    	{ return MULTEQ;   }
"/="	    	    	{ return DIVEQ;    }
"%="	    	    	{ return MODEQ;    }
"&="	    	    	{ return BANDEQ;   }
"^="	    	    	{ return BEOREQ;   }
"|="	    	    	{ return BIOREQ;   }
">>="	     	        { return RSHIFTEQ; }
"<<="	     	     	{ return LSHIFTEQ; }
"..."	     	     	{ return ELLIPSIS; }

    /* one-character operators and separaters */
";"                     { return ';';      }
"{"                  	{ return '{';      }
"}"                     { return '}';      }
","                     { return ',';      }
":"                     { return ':';      }
"="                     { return '=';      }
"("                     { return '(';      }
")"                     { return ')';      }
"["                     { return '[';      }
"]"                     { return ']';      }
"."                     { return '.';      }
"&"                     { return '&';      }
"!"                     { return '!';      }
"~"                     { return '~';      }
"-"                     { return '-';      }
"+"                     { return '+';      }
"*"                     { return '*';      }
"/"                     { return '/';      }
"%"                     { return '%';      }
"<"                     { return '<';      }
">"                     { return '>';      }
"^"                     { return '^';      }
"|"                     { return '|';      }
"?"                     { return '?';      }

    /* identifiers & literals (constants) */

    /* identifiers, typedef_names, enum constants */
{L}({L}|{D})*           { 
    	    	    	  if (typedef_flag == 0) {
			      yylval = create_AST2 (AST_IDENTIFIER, yytext);
			      return IDENTIFIER;
    	    	    	  }
    	    	    	  switch (typedef_kind (yytext)) {
			  case NK_UNKNOWN: 
			  case NK_IDENTIFIER: 
			      yylval = create_AST2 (AST_IDENTIFIER, yytext);
			      return IDENTIFIER;
			  case NK_TYPEDEF_NAME: 
			      yylval = create_AST2 (AST_TYPEDEF_NAME, yytext);
			      return TYPEDEF_NAME;
			  case NK_ENUMERATION_CONSTANT: 
			      yylval = create_AST2 (AST_ENUMERATION_CONSTANT,
						    yytext);
			      return ENUMERATION_CONSTANT;
			  default:
			      assert (0);
			      break;
    	    	    	  }
    	    	    	}

    /* integer literals: decimal, octal and hexadecimal*/
0{O}+{IS}               { yylval = process_int (yytext,
						INTEGER_CONSTANT_OCTAL);
    	    	    	  return INTEGER_CONSTANT; }
{D}+{IS}                { yylval = process_int (yytext,
						INTEGER_CONSTANT_DECIMAL);
    	    	    	  return INTEGER_CONSTANT; }
0[xX]{H}+{IS}        	{ yylval = process_int (yytext,
						INTEGER_CONSTANT_HEXADECIMAL);
    	    	    	  return INTEGER_CONSTANT; }

    /* floating literals */
{D}+{E}{FS}             |
{D}*"."{D}+{E}?{FS}     |
{D}+"."{D}*{E}?{FS}     { yylval = process_float (yytext);
    	    	    	  return FLOATING_CONSTANT; }

    /* character literals */
L?'\\[ntvbrfa\\\?'"]'   |    /* single escape sequence */
L?'\\{O}{1,3}'          |    /* octal escape sequence */
L?'\\(x|X){H}{1,2}'     |    /* hexadecimal escape sequence */
L?'[^\\']'           	{    /* usual character */
    	    	    	    int c = yytext[1];
    	    	    	    if (c == '\\') c = escape_seq_to_char (yytext+1);
    	    	    	    yylval = create_AST2 (AST_CHARACTER_CONSTANT, c);
    	    	    	    return CHARACTER_CONSTANT;
    	    	    	}

    /* string literals: concatenation is also done here */

    /* start of string literal */
L?\"    	        { BEGIN(string); string_buf = NULL; string_size = 0; }
    /* skpiping ", white-spaces and " in adjacent string literals */
<string>\"{WS}*L?\"     ;
    /* handling escape sequences */
<string>\\[ntvbrfa\\\?'\"]    |
<string>\\{O}{1,3}      |
<string>\\(x|X){H}{1,2} { add_char_to_string (escape_seq_to_char (yytext)); }
    /* end of string literal */
<string>\"    	        {
                            BEGIN(INITIAL);
                            add_char_to_string ('\0');
    	    	    	    yylval = create_AST2 (AST_STRING, string_buf);
                            return STRING;
                        }
    /* error handling */
<string>\n    	        yyerror ("newline in string literal");
<string>\\.    	        yyerror ("unknown escape sequence in string literal");
<string><<EOF>>	        yyerror ("EOF in string literal");
    /* the other characters */
<string>.    	        add_char_to_string (yytext[0]);

    /* comments (this way is easy to read, but not efficient) */
"/*"	    	    	BEGIN(comment);
<comment>"*/" 	    	BEGIN(INITIAL);
<comment><<EOF>>    	yyerror ("EOF in comment");
<comment>\n	   	lineno++; /* skipping any comments */
<comment>.    	      	;         /* skipping any comments */

    /* processing line number directives like the following,
	# 15
	# 15 "foo.h"
	# line 15 "foo.h"
       where we assume there is no double-quote except ones
       surrounding filename.
    */
^{WS2}*#{WS2}*(line)?{WS2}*{D}+{WS2}*(\".*\")?.*\n {
    	    	    	    char *s1, *s2;
			    s1 = strstr (yytext, "line");
    	    	    	    s2 = strchr (yytext, '\"'); /* head of filename */
    	    	    	    if (s1 != NULL && (s2 == NULL || s1 < s2)) {
				s1 += strlen ("line");   /* skip "# line" */
			    } else {
				s1 = strchr (yytext, '#');
    	    	    	    	assert (s1 != NULL);
    	    	    	    	s1++;  /* skip "#" */
			    }
			    lineno = strtol (s1, 0, 0);  /* get line number */
			    /* get filename if exists */
			    if (s2 != NULL) {
				char *s3 = strrchr (yytext, '\"');
    	    	    	    	int  len;
    	    	    	    	if (s3 == s2) {
    	    	    	    	    yyerror ("unbalanced double-quote"
    	    	    	    	    	     "in #line directive");
    	    	    	    	}
    	    	    	    	assert ((s3 != NULL) && (s3 > s2));
    	    	    	    	*s3 = '\0';
    	    	    	    	len = s3 - s2;
    	    	    	    	inc_filename = emalloc (len);
    	    	    	    	strncpy (inc_filename, s2+1, len);
    	    	    	    }
    	    	    	}
^{WS2}*#{WS2}*pragma.*\n        /* skip all pramgas */

    /* the other characters */
.    	     	        yyerror ("unknown character"); /* bad characters */

%%

static void
add_char_to_string (char c)
{
    string_buf = realloc (string_buf, ++string_size);
    string_buf [string_size-1] = c;
}

static struct AST *
process_int (char const * const yytext, int const radix)
{ 
    unsigned long int_ul;
    unsigned int status = 0;

    int is_unsigned = strrchr (yytext, 'U') || strrchr (yytext, 'u');
    int is_long     = strrchr (yytext, 'L') || strrchr (yytext, 'l');

    errno = 0;
    int_ul = strtoul (yytext, 0, 0);
    if (errno != 0) {
	perror ("strtoul:");
    	yyerror ("integer constant out of range");
    }

    assert ((radix == INTEGER_CONSTANT_OCTAL)
	    || (radix == INTEGER_CONSTANT_DECIMAL)
	    || (radix == INTEGER_CONSTANT_HEXADECIMAL));

    status |= radix;
    if (is_unsigned) status |= INTEGER_CONSTANT_UNSIGNED;
    if (is_long)     status |= INTEGER_CONSTANT_LONG;

    return create_AST3 (AST_INTEGER_CONSTANT, int_ul, status);
}

static struct AST *
process_float (char const * const yytext)
{
    long double  float_ld;
    unsigned int status = 0;

    int      result;
    int      is_float = strrchr (yytext, 'F') || strrchr (yytext, 'f');
    int      is_long  = strrchr (yytext, 'L') || strrchr (yytext, 'l');

    result = sscanf (yytext, "%Lf", &float_ld);
    if (result != 1) yyerror ("invalid floating format");

    if (is_float) status |= FLOATING_CONSTANT_FLOAT;
    if (is_long)  status |= FLOATING_CONSTANT_LONG;

    return create_AST3 (AST_FLOATING_CONSTANT, float_ld, status);
}

static char
escape_seq_to_char (const char * const yytext)
{
    int    num;

    assert (yytext[0] == '\\');

    switch (yytext[1]) {
    case 'n':  return '\n'; /* newline */
    case 't':  return '\t'; /* horizontal tab */
    case 'v':  return '\v'; /* vertical tab */
    case 'b':  return '\b'; /* backspace */
    case 'r':  return '\r'; /* carriage return */
    case 'f':  return '\f'; /* formfeed */
    case 'a':  return '\a'; /* audible alert */
    case '\\': return '\\'; /* backslash */
    case '?':  return '\?'; /* question mark */
    case '\'': return '\''; /* single quote */
    case '\"': return '\"'; /* double quote */
    case '0': case '1': case '2': case '3': /* octal number */
    case '4': case '5': case '6': case '7':
	sscanf (yytext+1, "%o", &num);
	if (num > 0xff) yyerror ("illegal octal number");
    	return num;
    case 'x': case 'X':     /* hex number */
	sscanf (yytext+2, "%x", &num);
	if (num > 0xff) yyerror ("illegal hex number");
    	return num;
    default:
    	assert (0); 
    	return yytext [1];
    }
}
