/*
    XCI: eXperimental C Interpreter for ANSI C
    -- Copyright (c) 2000-2001, gondow@jaist.ac.jp
    $Id: xml.c,v 1.2 2002/06/06 14:47:33 gondow Exp $
*/ 

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#include "xml.h"
#include "xci.h"
#include "type.h"
#include "symbol.h"
#include "attr.h"

#define XML_SUFFIX       ".xml"
#define COMMON_FMT       " id=\"X%x\" filename=\"%s\" lineno=\"%d\" "
#define COMMON_ARGS(ast)   ast, ast->inc_filename, ast->lineno
#define TYPE_FMT         " type_ref=\"X%x\" "
#define TYPE_ARGS(ast)   register_type ((ast)->type)
#define bool(x)          ((x) != 0 ? "true" : "false")

/* for type hashing */
#define HASH_INDEX(p) (((int)(p) / sizeof (struct Type *)) & (HASH_SIZE - 1))
/* HASH_SIZE * sizeof (struct TypeList) should be equal to the page size */
enum {HASH_SIZE = 2048};
struct TypeList {
    struct Type      *type;
    int              is_outputted;
    struct TypeList  *next;
};
struct TypeList *type_hash [HASH_SIZE]; /* hash table for all types */
static FILE *xml_fp;	/* output file; e.g., "foo.xml" */
/* ---------------------------------------------------------------------- */
static char *get_xml_filename (char *filename);
static void output_tag (int depth, char *tag, ...);
static void output_xml_ast (int depth, struct AST *ast);
static void output_xml_type (int depth, struct Type *type,
			     int is_const, int is_volatile);
static struct TypeList *find_type (struct Type *type);
static struct Type *register_type (struct Type *type);
static void output_xml_all_type (int depth);
static void output_xml_sym (int depth, struct Symbol *sym);
static void output_xml_all_file_symbol (int depth);
static void output_xml_all_block_symbol (int depth);

static char *namelevel2str (enum NameLevel namelevel);
static char *namespace2str (enum NameSpace namespace);

static struct AST *adjust_ast (struct AST *ast);
static char *case_default_fmt (struct AST *ast_switch);
static char *break_fmt        (struct AST *ast_break);
static char *continue_fmt     (struct AST *ast_continue);
static char *identifier_fmt   (struct AST *ast_identifier);
static char *make_string_escaped (char *str);
/* ---------------------------------------------------------------------- */
void
output_xml (char *filename, char *dtd_filename, struct AST *ast_root)
{
    char *xml_filename;

    xml_filename = get_xml_filename (filename);
    xml_fp = fopen (xml_filename, "w");
    if (xml_fp == NULL) {
	perror ("fopen");
	exit (1);
    }

    output_tag (0, "<?xml version=\"1.0\"?>\n"
		"<!DOCTYPE translation_unit SYSTEM \"%s\">\n"
		"<translation_unit" COMMON_FMT ">",
		dtd_filename, COMMON_ARGS (ast_root));
    output_xml_all_file_symbol (1);
    output_xml_ast (1, ast_root);
    output_xml_all_type (1);
    output_tag (0, "</translation_unit>");
    fclose (xml_fp);
}
/* ---------------------------------------------------------------------- */
/* return the xml filename to output.
   e.g. "hello.xml" for a given filename "../sample/hello.c" */
static char *
get_xml_filename (char *filename)
{
    int size;
    char *xml_filename, *p;

    size = strlen (filename) + strlen (XML_SUFFIX) + 1;
    xml_filename = emalloc (size);
    strcpy (xml_filename, filename);

    p = strrchr (xml_filename, '/');
    if (p != NULL) {
	xml_filename = p+1;
    }

    p = strrchr (xml_filename, '.');
    if (p == NULL) {
	p = strrchr (xml_filename, '\0');
	assert (p != NULL);
    }
    strcpy (p, XML_SUFFIX);
    return xml_filename;
}

/* output xml-tag with 'depth' indent level to xml-fp */
static void
output_tag (int depth, char *tag, ...)
{
    va_list  pvar;
    va_start (pvar, tag);
    fprintf (xml_fp, "%*s", depth*2, "");
    vfprintf (xml_fp, tag, pvar);
    fprintf (xml_fp, "\n");
    va_end (pvar);
}

/* very, very long definition. boring... */
static void
output_xml_ast (int depth, struct AST *ast)
{
    char *fmt;

    switch (ast->ast_type) {
    default:
	assert (0); break;

    /* zero child */
    case AST_declaration_list_opt_null:
    case AST_init_declarator_list_opt_null:
    case AST_struct_or_union_struct:
    case AST_struct_or_union_union:
    case AST_pointer_opt_null:
    case AST_constant_expression_opt_null:
    case AST_identifier_list_opt_null:
    case AST_type_qualifier_list_opt_nil:
    case AST_abstract_declarator_opt_null:
    case AST_parameter_type_list_opt_null:
    case AST_statement_list_opt_null:
    case AST_assignment_operator_eq:
    case AST_assignment_operator_mult:
    case AST_assignment_operator_div:
    case AST_assignment_operator_mod:
    case AST_assignment_operator_add:
    case AST_assignment_operator_sub:
    case AST_assignment_operator_lshift:
    case AST_assignment_operator_rshift:
    case AST_assignment_operator_band:
    case AST_assignment_operator_beor:
    case AST_assignment_operator_bior:
    case AST_unary_operator_addr:
    case AST_unary_operator_deref:
    case AST_unary_operator_add:
    case AST_unary_operator_sub:
    case AST_unary_operator_comp:
    case AST_unary_operator_neg:
	assert (ast->num_child == 0);
	break;

    /* one child */
    case AST_translation_unit_single:
    case AST_external_declaration_func:
    case AST_external_declaration_dec:
    case AST_declaration_specifers_storage:
    case AST_declaration_specifers_spec:
    case AST_declaration_specifers_qual:
    case AST_declaration_list_opt_direct:
    case AST_init_declarator_list_opt_direct:
    case AST_declaration_list_single:
    case AST_type_specifier_struct_or_union:
    case AST_type_specifier_enum:
    case AST_type_specifier_TYPEDEF_NAME:
    case AST_struct_declaration_list_single:
    case AST_struct_declarator_direct:
    case AST_struct_declarator_bitfield_noname:
    case AST_init_declarator_single:
    case AST_init_declarator_dec:
    case AST_specifier_qualifier_list_spec:
    case AST_specifier_qualifier_list_qual:
    case AST_struct_declarator_list_single:
    case AST_enumerator_list_single:
    case AST_enumerator_id:
    case AST_pointer_opt_direct:
    case AST_constant_expression_opt_direct:
    case AST_identifier_list_opt_direct:
    case AST_type_qualifier_list_opt_direct:
    case AST_type_qualifier_list_single:
    case AST_parameter_type_list_direct:
    case AST_parameter_list_single:
    case AST_abstract_declarator_opt_direct:
    case AST_identifier_list_single:
    case AST_initializer_list_single:
    case AST_parameter_type_list_opt_direct:
    case AST_statement_labeled:
    case AST_statement_expression:
    case AST_statement_compound:
    case AST_statement_selection:
    case AST_statement_iteration:
    case AST_statement_jump:
    case AST_expression_opt_direct:
    case AST_statement_list_opt_direct:
    case AST_statement_list_single:
    case AST_expression_direct:
    case AST_assignment_expression_direct:
    case AST_conditional_expression_direct:
    case AST_constant_expression_direct:
    case AST_logical_or_expression_direct:
    case AST_logical_and_expression_direct:
    case AST_inclusive_or_expression_direct:
    case AST_exclusive_or_expression_direct:
    case AST_and_expression_direct:
    case AST_equality_expression_direct:
    case AST_relational_expression_direct:
    case AST_shift_expression_direct:
    case AST_additive_expression_direct:
    case AST_multiplicative_expression_direct:
    case AST_cast_expression_direct:
    case AST_unary_expression_direct:
    case AST_postfix_expression_direct:
    case AST_argument_expression_list_single:
    case AST_constant_INT:
    case AST_constant_CHAR:
    case AST_constant_FLOAT:
    case AST_constant_ENUM:
	assert (ast->num_child == 1);
	output_xml_ast (depth, ast->u.child [0]);
	break;

    /* two child */
    case AST_translation_unit_pair:
    case AST_declaration_specifers_storage_pair:
    case AST_declaration_specifers_spec_pair:
    case AST_declaration_specifers_qual_pair:
    case AST_declaration_list_pair:
    case AST_struct_declaration_list_pair:
    case AST_struct_declarator_bitfield:
    case AST_init_declarator_pair:
    case AST_init_declarator_dec_init:
    case AST_specifier_qualifier_list_spec_pair:
    case AST_specifier_qualifier_list_qual_pair:
    case AST_struct_declarator_list_pair:
    case AST_enumerator_list_pair:
    case AST_enumerator_id_init:
    case AST_type_qualifier_list_pair:
    case AST_parameter_list_pair:
    case AST_identifier_list_pair:
    case AST_initializer_list_pair:
    case AST_statement_list_pair:
    case AST_argument_expression_list_pair:
	assert (ast->num_child == 2);
	output_xml_ast (depth, ast->u.child [0]);
	output_xml_ast (depth, ast->u.child [1]);
	break;

    /* xml tagging */
    case AST_declaration_direct:
	output_tag (depth, "<declaration" COMMON_FMT ">",
		    COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</declaration>");
	break;

    case AST_function_definition_nospec:
	output_tag (depth, "<function_definition" COMMON_FMT ">",
		    COMMON_ARGS (ast));
	sym_begin_function ();
	sym_entry_all_labels (ast->u.child [2]);
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_xml_ast (depth + 1, ast->u.child [2]);
	sym_end_function ();
	output_tag (depth, "</function_definition>");
	break;

    case AST_function_definition_spec:
	output_tag (depth, "<function_definition" COMMON_FMT ">",
		    COMMON_ARGS (ast));
	sym_begin_function ();
	sym_entry_all_labels (ast->u.child [3]);
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_xml_ast (depth + 1, ast->u.child [2]);
	output_xml_ast (depth + 1, ast->u.child [3]);
	sym_end_function ();
	output_tag (depth, "</function_definition>");
	break;

    case AST_storage_class_specifier_auto:
	output_tag (depth, "<auto/>");
	break;
    case AST_storage_class_specifier_register:
	output_tag (depth, "<register/>");
	break;
    case AST_storage_class_specifier_static:
	output_tag (depth, "<static/>");
	break;
    case AST_storage_class_specifier_extern:
	output_tag (depth, "<extern/>");
	break;
    case AST_storage_class_specifier_typedef:
	output_tag (depth, "<typedef/>");
	break;
    case AST_type_specifier_void:
	output_tag (depth, "<void/>");
	break;
    case AST_type_specifier_char:
	output_tag (depth, "<char/>");
	break;
    case AST_type_specifier_short:
	output_tag (depth, "<short/>");
	break;
    case AST_type_specifier_int:
	output_tag (depth, "<int/>");
	break;
    case AST_type_specifier_long:
	output_tag (depth, "<long/>");
	break;
    case AST_type_specifier_float:
	output_tag (depth, "<float/>");
	break;
    case AST_type_specifier_double:
	output_tag (depth, "<double/>");
	break;
    case AST_type_specifier_signed:
	output_tag (depth, "<signed/>");
	break;
    case AST_type_specifier_unsigned:
	output_tag (depth, "<unsigned/>");
	break;
    case AST_type_qualifier_const:
	output_tag (depth, "<const/>");
	break;
    case AST_type_qualifier_volatile:
	output_tag (depth, "<volatile/>");
	break;

    case AST_struct_or_union_specifier_notag:
    case AST_struct_or_union_specifier_tag:
    case AST_struct_or_union_specifier_incomplete:
	assert (ast->u.child [0] != NULL);
	switch (ast->u.child [0]->ast_type) {
	case AST_struct_or_union_struct:
	    output_tag (depth, "<struct_or_union_specifier rhs=\"struct\""
			COMMON_FMT ">", COMMON_ARGS (ast));
	    break;
	case AST_struct_or_union_union:
	    output_tag (depth, "<struct_or_union_specifier rhs=\"union\""
			COMMON_FMT ">", COMMON_ARGS (ast));
	    break;
	default:
	    assert (0);
	    break;
	}
	switch (ast->ast_type) {
	case AST_struct_or_union_specifier_notag:
	case AST_struct_or_union_specifier_incomplete:
	    output_xml_ast (depth + 1, ast->u.child [1]);
	    break;
	case AST_struct_or_union_specifier_tag:
	    output_xml_ast (depth + 1, ast->u.child [1]);
	    output_xml_ast (depth + 1, ast->u.child [2]);
	    break;
	default:
	    assert (0);
	    break;
    	}
	output_tag (depth, "</struct_or_union_specifier>");
	break;

    case AST_struct_declaration_direct:
	output_tag (depth, "<struct_declaration" COMMON_FMT ">",
		    COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</struct_declaration>");
	break;

    case AST_enum_specifier_untag:
	output_tag (depth, "<enum_specifier rhs=\"notag\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</enum_specifier>");
	break;
    case AST_enum_specifier_tag:
	output_tag (depth, "<enum_specifier rhs=\"normal\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</enum_specifier>");
	break;
    case AST_enum_specifier_dec:
	output_tag (depth, "<enum_specifier rhs=\"nobody\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</enum_specifier>");
	break;

    case AST_declarator_direct:
    case AST_function_declarator_direct:
	assert (ast->u.child [0] != NULL);
	switch (ast->u.child [0]->ast_type) {
	case AST_pointer_opt_null:
	    output_tag (depth, "<declarator rhs=\"pointer_null\""
			COMMON_FMT ">", COMMON_ARGS (ast));
	    output_xml_ast (depth + 1, ast->u.child [1]);
	    output_tag (depth, "</declarator>");
	    break;
	case AST_pointer_opt_direct:
	    output_tag (depth, "<declarator rhs=\"pointer\""
			COMMON_FMT ">", COMMON_ARGS (ast));
	    output_xml_ast (depth + 1, ast->u.child [0]);
	    output_xml_ast (depth + 1, ast->u.child [1]);
	    output_tag (depth, "</declarator>");
	    break;
	default:
	    assert (0);
	    break;
	}
	break;

    case AST_direct_declarator_id:
	output_tag (depth, "<declarator rhs=\"id\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</declarator>");
	break;

    case AST_direct_declarator_paren:
	output_tag (depth, "<declarator rhs=\"paren\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</declarator>");
	break;

    case AST_direct_declarator_array:
	output_tag (depth, "<declarator rhs=\"array\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</declarator>");
	break;

    case AST_direct_declarator_func_new:
    case AST_direct_function_declarator_func_new:
	output_tag (depth, "<declarator rhs=\"func_new\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</declarator>");
	break;

    case AST_direct_declarator_func_old:
    case AST_direct_function_declarator_func_old:
	output_tag (depth, "<declarator rhs=\"func_old\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</declarator>");
	break;

    case AST_pointer_single:
	output_tag (depth, "<pointer rhs=\"single\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</pointer>");
	break;

    case AST_pointer_pair:
	output_tag (depth, "<pointer rhs=\"pair\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</pointer>");
	break;

    case AST_parameter_type_list_ellipsis:
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "<ellipsis/>");
	break;

    case AST_parameter_declaration_dec:
	output_tag (depth, "<parameter_declaration rhs=\"dec\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</parameter_declaration>");
	break;

    case AST_parameter_declaration_abs:
	output_tag (depth, "<parameter_declaration rhs=\"abs\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</parameter_declaration>");
	break;

    case AST_initializer_assign:
	output_tag (depth, "<initializer rhs=\"exp\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</initializer>");
	break;

    case AST_initializer_list:
    case AST_initializer_list_comma:
	output_tag (depth, "<initializer rhs=\"list\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</initializer>");
	break;

    case AST_type_name_direct:
	output_tag (depth, "<type_name" COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</type_name>");
	break;

    case AST_abstract_declarator_pointer:
	output_tag (depth, "<abstract_declarator rhs=\"pointer_nil\""
    	    	    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</abstract_declarator>");
	break;

    case AST_abstract_declarator_abs:
	assert (ast->u.child [0] != NULL);
	switch (ast->u.child [0]->ast_type) {
	case AST_pointer_opt_null:
	    output_xml_ast (depth + 1, ast->u.child [1]);
	    break;
	case AST_pointer_opt_direct:
	    output_tag (depth, "<abstract_declarator rhs=\"pointer_pair\""
    	    	    	COMMON_FMT ">", COMMON_ARGS (ast));
	    output_xml_ast (depth + 1, ast->u.child [0]);
	    output_xml_ast (depth + 1, ast->u.child [1]);
	    output_tag (depth, "</abstract_declarator>");
	    break;
	default:
	    assert (0);
	    break;
	}
	break;

    case AST_direct_abstract_declarator_paren:
	output_tag (depth, "<abstract_declarator rhs=\"paren\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</abstract_declarator>");
	break;

    case AST_direct_abstract_declarator_array_nil:
	output_tag (depth, "<abstract_declarator rhs=\"array_nil\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</abstract_declarator>");
	break;

    case AST_direct_abstract_declarator_array_pair:
	output_tag (depth, "<abstract_declarator rhs=\"array_pair\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</abstract_declarator>");
	break;

    case AST_direct_abstract_declarator_func_nil:
	output_tag (depth, "<abstract_declarator rhs=\"func_nil\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</abstract_declarator>");
	break;

    case AST_direct_abstract_declarator_func_pair:
	output_tag (depth, "<abstract_declarator rhs=\"func_pair\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</abstract_declarator>");
	break;

    /* statement */
    case AST_labeled_statement_label:
	output_tag (depth, "<statement rhs=\"label\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</statement>");
	break;

    case AST_labeled_statement_case:
	output_tag (depth, "<statement rhs=\"case\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</statement>");
	break;

    case AST_labeled_statement_default:
	output_tag (depth, "<statement rhs=\"default\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</statement>");
	break;

    case AST_expression_statement_direct:
	output_tag (depth, "<statement rhs=\"expression_opt\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</statement>");
	break;

    case AST_expression_opt_null:
	output_tag (depth, "<empty_expression/>");
	break;

    case AST_compound_statement_decn:
	output_tag (depth, "<statement rhs=\"compound\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	sym_begin_block ();
	sym_copy_static_block (ast);
	output_xml_all_block_symbol (depth + 1);
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	sym_end_block ();
	output_tag (depth, "</statement>");
	break;

    case AST_compound_statement_nodecn:
	output_tag (depth, "<statement rhs=\"compound\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	sym_begin_block ();
	sym_copy_static_block (ast);
	output_xml_all_block_symbol (depth + 1);
	output_xml_ast (depth + 1, ast->u.child [0]);
	sym_end_block ();
	output_tag (depth, "</statement>");
	break;

    case AST_selection_statement_if:
	output_tag (depth, "<statement rhs=\"if\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</statement>");
	break;

    case AST_selection_statement_if_else:
	output_tag (depth, "<statement rhs=\"if-else\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_xml_ast (depth + 1, ast->u.child [2]);
	output_tag (depth, "</statement>");
	break;

    case AST_selection_statement_switch:
	fmt = case_default_fmt (ast);
	output_tag (depth, "<statement rhs=\"switch\" %s"
		    COMMON_FMT ">", fmt, COMMON_ARGS (ast));
	efree (fmt);
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</statement>");
	break;

    case AST_iteration_statement_while:
	output_tag (depth, "<statement rhs=\"while\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</statement>");
	break;

    case AST_iteration_statement_do_while:
	output_tag (depth, "<statement rhs=\"do-while\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</statement>");
	break;

    case AST_iteration_statement_for:
	output_tag (depth, "<statement rhs=\"for\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_xml_ast (depth + 1, ast->u.child [2]);
	output_xml_ast (depth + 1, ast->u.child [3]);
	output_tag (depth, "</statement>");
	break;

    case AST_jump_statement_goto:
	{
	    struct Attribute *attr;
	    attr = get_goto_dest (ast);
	    assert (attr->kind == ATTR_GOTO_DEST);
	    output_tag (depth, "<statement rhs=\"goto\" goto_ref=\"X%x\""
			COMMON_FMT ">",
			attr->u.a_goto.ast_dest,
			COMMON_ARGS (ast));
	    output_xml_ast (depth + 1, ast->u.child [0]);
	    output_tag (depth, "</statement>");
	}
	break;

    case AST_jump_statement_continue:
	fmt = continue_fmt (ast);
	output_tag (depth, "<statement rhs=\"continue\" %s"
		    COMMON_FMT ">", fmt, COMMON_ARGS (ast));
	efree (fmt);
	output_tag (depth + 1, "<continue/>");
	output_tag (depth, "</statement>");
	break;

    case AST_jump_statement_break:
	fmt = break_fmt (ast);
	output_tag (depth, "<statement rhs=\"break\" %s"
		    COMMON_FMT ">", fmt, COMMON_ARGS (ast));
	efree (fmt);
	output_tag (depth + 1, "<break/>");
	output_tag (depth, "</statement>");
	break;

    case AST_jump_statement_return:
	output_tag (depth, "<statement rhs=\"return\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</statement>");
	break;

    /* expression */
    case AST_expression_comma:
	output_tag (depth, "<expression rhs=\"comma\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_assignment_expression_assign:
	assert (ast->u.child [1] != NULL);
	switch (ast->u.child [1]->ast_type) {
	default:
	    assert (0); break;
	case AST_assignment_operator_eq:
	    output_tag (depth, "<expression rhs=\"assign_eq\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_assignment_operator_mult:
	    output_tag (depth, "<expression rhs=\"assign_mult\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_assignment_operator_div:
	    output_tag (depth, "<expression rhs=\"assign_div\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_assignment_operator_mod:
	    output_tag (depth, "<expression rhs=\"assign_mod\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_assignment_operator_add:
	    output_tag (depth, "<expression rhs=\"assign_add\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_assignment_operator_sub:
	    output_tag (depth, "<expression rhs=\"assign_sub\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_assignment_operator_lshift:
	    output_tag (depth, "<expression rhs=\"assign_lshift\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_assignment_operator_rshift:
	    output_tag (depth, "<expression rhs=\"assign_rshift\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_assignment_operator_band:
	    output_tag (depth, "<expression rhs=\"assign_band\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_assignment_operator_beor:
	    output_tag (depth, "<expression rhs=\"assign_beor\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_assignment_operator_bior:
	    output_tag (depth, "<expression rhs=\"assign_bior\""
		    	TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	}
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [2]);
	output_tag (depth, "</expression>");
	break;

    case AST_conditional_expression_cond:
	output_tag (depth, "<expression rhs=\"conditional\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_xml_ast (depth + 1, ast->u.child [2]);
	output_tag (depth, "</expression>");
	break;

    case AST_logical_or_expression_lor:
	output_tag (depth, "<expression rhs=\"logical-or\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_logical_and_expression_land:
	output_tag (depth, "<expression rhs=\"logical-and\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_inclusive_or_expression_bior:
	output_tag (depth, "<expression rhs=\"bitwise-inclusive-or\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_exclusive_or_expression_beor:
	output_tag (depth, "<expression rhs=\"bitwise-exclusive-or\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_and_expression_band:
	output_tag (depth, "<expression rhs=\"bitwise-and\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_equality_expression_eq:
	output_tag (depth, "<expression rhs=\"equal\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_equality_expression_neq:
	output_tag (depth, "<expression rhs=\"not-equal\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_relational_expression_less:
	output_tag (depth, "<expression rhs=\"less\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_relational_expression_more:
	output_tag (depth, "<expression rhs=\"more\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_relational_expression_lesseq:
	output_tag (depth, "<expression rhs=\"less-equal\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_relational_expression_moreeq:
	output_tag (depth, "<expression rhs=\"more-equal\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_shift_expression_lshift:
	output_tag (depth, "<expression rhs=\"left-shift\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_shift_expression_rshift:
	output_tag (depth, "<expression rhs=\"right-shift\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_additive_expression_add:
	output_tag (depth, "<expression rhs=\"add\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_additive_expression_sub:
	output_tag (depth, "<expression rhs=\"sub\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_multiplicative_expression_mult:
	output_tag (depth, "<expression rhs=\"mult\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_multiplicative_expression_div:
	output_tag (depth, "<expression rhs=\"div\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_multiplicative_expression_mod:
	output_tag (depth, "<expression rhs=\"mod\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_cast_expression_cast:
	output_tag (depth, "<expression rhs=\"cast\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_unary_expression_inc:
	output_tag (depth, "<expression rhs=\"pre-increment\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</expression>");
	break;

    case AST_unary_expression_dec:
	output_tag (depth, "<expression rhs=\"pre-decrement\""
	    	    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</expression>");
	break;

    case AST_unary_expression_unary_op:
	assert (ast->u.child [0] != NULL);
	switch (ast->u.child [0]->ast_type) {
	default:
	    assert (0);
	    break;
	case AST_unary_operator_addr:
	    output_tag (depth, "<expression rhs=\"unary-address\""
			TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_unary_operator_deref:
	    output_tag (depth, "<expression rhs=\"unary-deref\""
			TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_unary_operator_add:
	    output_tag (depth, "<expression rhs=\"unary-add\""
			TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_unary_operator_sub:
	    output_tag (depth, "<expression rhs=\"unary-sub\""
			TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_unary_operator_comp:
	    output_tag (depth, "<expression rhs=\"unary-comp\""
			TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	case AST_unary_operator_neg:
	    output_tag (depth, "<expression rhs=\"unary-neg\""
			TYPE_FMT COMMON_FMT ">",
			TYPE_ARGS (ast), COMMON_ARGS (ast));
	    break;
	}
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_unary_expression_sizeof_exp:
	output_tag (depth, "<expression rhs=\"sizeof-exp\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</expression>");
	break;

    case AST_unary_expression_sizeof_type:
	output_tag (depth, "<expression rhs=\"sizeof-type\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</expression>");
	break;

    case AST_postfix_expression_array:
	output_tag (depth, "<expression rhs=\"array\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_postfix_expression_funcall:
	output_tag (depth, "<expression rhs=\"funcall\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_postfix_expression_funcall_noarg:
	output_tag (depth, "<expression rhs=\"funcall\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</expression>");
	break;

    case AST_postfix_expression_dot:
	output_tag (depth, "<expression rhs=\"dot\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_postfix_expression_arrow:
	output_tag (depth, "<expression rhs=\"arrow\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_xml_ast (depth + 1, ast->u.child [1]);
	output_tag (depth, "</expression>");
	break;

    case AST_postfix_expression_inc:
	output_tag (depth, "<expression rhs=\"post-increment\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</expression>");
	break;

    case AST_postfix_expression_dec:
	output_tag (depth, "<expression rhs=\"post-decrement\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</expression>");
	break;

    case AST_primary_expression_paren:
	output_tag (depth, "<expression rhs=\"paren\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</expression>");
	break;

    case AST_primary_expression_ID:
	output_tag (depth, "<expression rhs=\"identifier\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</expression>");
	break;

    case AST_primary_expression_constant:
    case AST_primary_expression_STRING:
	output_tag (depth, "<expression rhs=\"constant\""
		    TYPE_FMT COMMON_FMT ">",
		    TYPE_ARGS (ast), COMMON_ARGS (ast));
	output_xml_ast (depth + 1, ast->u.child [0]);
	output_tag (depth, "</expression>");
	break;


    /* identifier, literal (constant) */
    case AST_IDENTIFIER:
	fmt = identifier_fmt (ast);
	output_tag (depth, "<identifier rhs=\"identifier\" %s"
		    COMMON_FMT ">", fmt, COMMON_ARGS (ast));
	output_tag (depth + 1, "%s", ast->u.string);
	output_tag (depth, "</identifier>");
	break;

    case AST_ENUMERATION_CONSTANT:
	fmt = identifier_fmt (ast);
	output_tag (depth, "<identifier rhs=\"enum_constant\" %s"
		    COMMON_FMT ">", fmt, COMMON_ARGS (ast));
	output_tag (depth + 1, "%s", ast->u.string);
	output_tag (depth, "</identifier>");
	break;

    case AST_TYPEDEF_NAME:
	fmt = identifier_fmt (ast);
	output_tag (depth, "<typedef_name %s"
		    COMMON_FMT ">", fmt, COMMON_ARGS (ast));
	output_tag (depth + 1, "%s", ast->u.string);
	output_tag (depth, "</typedef_name>");
	break;

    case AST_STRING:
	output_tag (depth, "<constant rhs=\"string\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	{
	    char *buf = make_string_escaped (ast->u.string);
	    output_tag (depth + 1, "<![CDATA[%s]]>", buf);
	    efree (buf);
	}
	output_tag (depth, "</constant>");
	break;

    case AST_CHARACTER_CONSTANT:
	output_tag (depth, "<constant rhs=\"char\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_tag (depth + 1, "%d", ast->u.char_i);
	output_tag (depth, "</constant>");
	break;

    case AST_INTEGER_CONSTANT:
	output_tag (depth, "<constant rhs=\"int\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_tag (depth + 1, "%d", ast->u.int_ul);
	output_tag (depth, "</constant>");
	break;

    case AST_FLOATING_CONSTANT:
	output_tag (depth, "<constant rhs=\"float\""
		    COMMON_FMT ">", COMMON_ARGS (ast));
	output_tag (depth + 1, "%Lf", *ast->u.float_ld);
	output_tag (depth, "</constant>");
	break;
    }
}

static void
output_xml_type (int depth, struct Type *type, int is_const, int is_volatile)
{
    int  i;
    struct TypeList *tl;
	
    assert (type != NULL);

    if (type->kind == TYPE_KIND_QUAL) {
	output_xml_type (depth, type->u.t_qual.type,
			 (type->u.t_qual.qualifier & TYPE_QUAL_CONST),
			 (type->u.t_qual.qualifier & TYPE_QUAL_VOLATILE));
	return;
    }

    tl = find_type (type);
    assert (tl != NULL);

    if (tl->is_outputted) {
	output_tag (depth, "<type size=\"%d\" type_ref=\"X%x\""
		    " is_const=\"%s\" is_volatile=\"%s\">",
		    type->size, type, bool (is_const), bool (is_volatile));
	output_tag (depth + 1, "<t_empty/>");
	output_tag (depth, "</type>");
	return;
    }

    output_tag (depth, "<type size=\"%d\" id=\"X%x\""
		" is_const=\"%s\" is_volatile=\"%s\">",
		type->size, type, bool (is_const), bool (is_volatile));

    tl->is_outputted = 1;
    switch (type->kind) {
    case TYPE_KIND_PRIM:
	output_tag (depth + 1, "<t_prim>");
	switch (type->u.t_prim.ptype) {
	default:
	    assert (0);
	    break;
	case PRIM_TYPE_VOID:
	    output_tag (depth + 2, "<t_void/>");
	    break;
	case PRIM_TYPE_CHAR:
	    output_tag (depth + 2,
			"<t_char is_signed=\"%s\" is_unsigned=\"%s\"/>",
			bool (type->u.t_prim.modifier & TYPE_MOD_SIGNED),
			bool (type->u.t_prim.modifier & TYPE_MOD_UNSIGNED));
	    break;
	case PRIM_TYPE_INT:
	    output_tag (depth + 2,
#ifdef XCI_LONGLONG
			"<t_int is_longlong=\"%s\" is_long=\"%s\" is_short=\"%s\""
			" is_signed=\"%s\" is_unsigned=\"%s\"/>",
			bool (type->u.t_prim.modifier & TYPE_MOD_LONGLONG),
#else
			"<t_int is_long=\"%s\" is_short=\"%s\""
			" is_signed=\"%s\" is_unsigned=\"%s\"/>",
#endif /* XCI_LONGLONG */
			bool (type->u.t_prim.modifier & TYPE_MOD_LONG),
			bool (type->u.t_prim.modifier & TYPE_MOD_SHORT),
			bool (type->u.t_prim.modifier & TYPE_MOD_SIGNED),
			bool (type->u.t_prim.modifier & TYPE_MOD_UNSIGNED));
	    break;
	case PRIM_TYPE_FLOAT:
	    output_tag (depth + 2, "<t_float/>");
	    break;
	case PRIM_TYPE_DOUBLE:
	    output_tag (depth + 2, "<t_double is_long=\"%s\"/>",
			bool (type->u.t_prim.modifier & TYPE_MOD_LONG));
	    break;
	}
	output_tag (depth + 1, "</t_prim>");
	break;

    case TYPE_KIND_ENUM:
	if (type->u.t_enum.tag == NULL) {
	    output_tag (depth + 1, "<t_enum num=\"%d\">",
			type->u.t_enum.num);
	} else {
	    output_tag (depth + 1, "<t_enum tag=\"%s\" num=\"%d\">",
			type->u.t_enum.tag, type->u.t_enum.num);
	}
	for (i = 0; i < type->u.t_enum.num; i++) {
	    char *elem = type->u.t_enum.elem [i];
	    char value = type->u.t_enum.value [i];

	    output_tag (depth + 2, "<t_identifier>");
	    output_tag (depth + 3, elem);
	    output_tag (depth + 2, "</t_identifier>");
	    output_tag (depth + 2, "<t_constant>");
	    output_tag (depth + 3, "%d", value);
	    output_tag (depth + 2, "</t_constant>");
	}
	output_tag (depth + 1, "</t_enum>");
	break;

    case TYPE_KIND_STRUCT:
    case TYPE_KIND_UNION:
	if (type->u.t_struct_union.tag == NULL) { 
			/* no tag */
	    output_tag (depth + 1, "<t_%s num=\"%d\">",
			(type->kind == TYPE_KIND_STRUCT ? "struct" : "union"),
			type->u.t_struct_union.num);
	} else {
	    output_tag (depth + 1, "<t_%s tag=\"%s\" num=\"%d\">",
			(type->kind == TYPE_KIND_STRUCT ? "struct" : "union"),
			type->u.t_struct_union.tag,
			type->u.t_struct_union.num);
	}

	for (i = 0; i < type->u.t_struct_union.num; i++) {
	    struct StructUnionField *field;
	    field = type->u.t_struct_union.field [i];
	    if (field->is_bit_field) {
		output_tag (depth + 2,
			    "<t_field name=\"%s\" offset=\"%d\""
			    " bit_offset=\"%d\" bit_size=\"%d\">",
			    field->name, field->offset,
			    field->bit_offset, field->bit_size);
	    } else {
		output_tag (depth + 2,
			    "<t_field name=\"%s\" offset=\"%d\">",
			    field->name, field->offset);
	    }
	    output_xml_type (depth + 3, field->type, 0, 0);
	    output_tag (depth + 2, "</t_field>");
	}
	output_tag (depth + 1, "</t_%s>",
		    (type->kind == TYPE_KIND_STRUCT ? "struct" : "union"));
	break;

    case TYPE_KIND_POINTER:
	output_tag (depth + 1, "<t_pointer>");
	output_xml_type (depth + 2, type->u.t_pointer.type, 0, 0);
	output_tag (depth + 1, "</t_pointer>");
	break;

    case TYPE_KIND_ARRAY:
	output_tag (depth + 1, "<t_array num=\"%d\">", type->u.t_array.num);
	output_xml_type (depth + 2, type->u.t_array.elem_type, 0, 0);
	output_tag (depth + 1, "</t_array>");
	break;

    case TYPE_KIND_FUNCTION:
	output_tag (depth + 1, "<t_function arg_num=\"%d\""
		    " is_ellipsis=\"%s\" is_old_type=\"%s\">",
		    type->u.t_function.arg_num,
		    bool (type->u.t_function.is_ellipsis),
		    bool (type->u.t_function.is_old_type));
	output_xml_type (depth + 2, type->u.t_function.ret_type, 0, 0);
	for (i = 0; i < type->u.t_function.arg_num; i++) {
	    char *arg_name = type->u.t_function.arg_name [i];

	    output_xml_type (depth + 2, type->u.t_function.arg_type [i], 0, 0);
	    if (arg_name != NULL) {
		output_tag (depth + 2, "<t_identifier>");
		output_tag (depth + 3, arg_name);
		output_tag (depth + 2, "</t_identifier>");
    	    }
	}

	output_tag (depth + 1, "</t_function>");
	break;

    case TYPE_KIND_QUAL:
	/* type-qualifiers should be handled before this switch */
	/* falling through */
    default:
	assert (0);
	break;
    }

    output_tag (depth, "</type>");
}

static struct TypeList *
find_type (struct Type *type)
{
    struct TypeList *tl;
    for (tl = type_hash [HASH_INDEX (type)]; tl != NULL; tl = tl->next) {
	if (tl->type == type)
	    return tl;
    }
    return NULL;
}

static struct Type *
register_type (struct Type *type)
{
    struct TypeList *new;
    int    i;

    if (type->kind == TYPE_KIND_QUAL) {
	return register_type (type->u.t_qual.type);
    }

    /* first check if this 'type' is already registered or not */
    if (find_type (type) != NULL)
	return type;

    /* ok, we need to register this type in the hash */
    new = emalloc (sizeof (struct TypeList));
    new->type = type;
    new->is_outputted = 0;
    new->next = type_hash [HASH_INDEX (type)];
    type_hash [HASH_INDEX (type)] = new;

    /* register inner types */
    switch (type->kind) {
    case TYPE_KIND_STRUCT:
    case TYPE_KIND_UNION:
	for (i = 0; i < type->u.t_struct_union.num; i++) {
	    (void) register_type (type->u.t_struct_union.field [i]->type);
	}
	break;

    case TYPE_KIND_FUNCTION:
	(void) register_type (type->u.t_function.ret_type);
	for (i = 0; i < type->u.t_function.arg_num; i++) {
	    (void) register_type (type->u.t_function.arg_type [i]);
	}
	break;

    case TYPE_KIND_POINTER:
	(void) register_type (type->u.t_pointer.type);
	break;

    case TYPE_KIND_ARRAY:
	(void) register_type  (type->u.t_array.elem_type);
	break;

    case TYPE_KIND_PRIM:
    case TYPE_KIND_ENUM:
	break;

    case TYPE_KIND_QUAL:
	/* falling through */
    default:
	assert (0);
	break;
    }
    return type;
}

static void
output_xml_all_type (int depth)
{
    int i;
    struct TypeList *tl;

    /* empty type entry for non-existing types, whose id equals NULL */
    output_tag (depth, "<type size=\"-1\" id=\"X0\""
		" is_const=\"false\" is_volatile=\"false\">");
    output_tag (depth + 1, "<t_empty/>");
    output_tag (depth, "</type>");

    /* output all existing types */
    for (i = 0; i < HASH_SIZE; i++) {
	for (tl = type_hash [i]; tl != NULL; tl = tl->next) {
	    if (!tl->is_outputted)
		output_xml_type (depth, tl->type, 0, 0);
	}
    }
}

static void
output_xml_sym (int depth, struct Symbol *sym)
{
    if (sym->type != NULL) {
	(void) register_type (sym->type);
    }

    output_tag (depth, "<symbol "
		"name=\"%s\" "
		"type_ref=\"X%x\" "
		"ast_ref=\"X%x\" "
		"namespace=\"%s\" "
		"namelevel=\"%s\" "
		"is_tentative=\"%s\" "
		"is_incomplete=\"%s\" "
		"is_declaration=\"%s\" "
		"is_definition=\"%s\" "
		"is_global=\"%s\" "
		"is_enum_const=\"%s\" "
		"is_typedef=\"%s\" "
		"is_static=\"%s\" "
		"is_argument=\"%s\" "
		"is_builtin=\"%s\" "
		"/>",
		sym->name, sym->type,
		adjust_ast (sym->ast),
		namespace2str (sym->namespace),
		namelevel2str (sym->namelevel),
		bool (sym->status & SYM_STAT_TENTATIVE),
		bool (sym->status & SYM_STAT_INCOMPLETE),
		bool (sym->status & SYM_STAT_DECLARATION),
		bool (sym->status & SYM_STAT_DEFINITION),
		bool (sym->status & SYM_STAT_GLOBAL),
		bool (sym->status & SYM_STAT_ENUM_CONST),
		bool (sym->status & SYM_STAT_TYPEDEF),
		bool (sym->status & SYM_STAT_STATIC),
		bool (sym->status & SYM_STAT_ARGUMENT),
		bool (sym->status & SYM_STAT_BUILTIN)
	);
}

static char *
namelevel2str (enum NameLevel namelevel)
{
    switch (namelevel) {
    case NL_FILE:        return "file";
    case NL_BLOCK:       return "block";
    default:             assert (0); return NULL;
    }
}

static char *
namespace2str (enum NameSpace namespace)
{
    switch (namespace) {
    case NS_NORMAL:      return "normal";
    case NS_LABEL:       return "label";
    case NS_TAG:         return "tag";
    case NS_MEMBER:      return "member";
    default:             assert (0); return NULL;
    }
}

static void
output_xml_all_file_symbol (int depth)
{
    struct SymList     *symlist;
    struct Symbol      *sym;

    symlist = sym_table.curr_file;
    for (sym = symlist->symbol; sym != NULL; sym = sym->next) {
	output_xml_sym (depth, sym);
    }
}

static void
output_xml_all_block_symbol (int depth)
{
    struct SymStack *symstack;
    struct Symbol   *sym;
    int             i;

    /* output variables, functions, tags, enum constants, typedef_names */
    symstack = sym_table.func_stack [sym_table.func_sp];
    i = symstack->block_sp;
    for (sym = symstack->block_stack [i]; sym != NULL; sym = sym->next) {
	output_xml_sym (depth, sym);
    }

    /* output labels */
    for (sym = symstack->label; sym != NULL; sym = sym->next) {
	output_xml_sym (depth, sym);
    }
}

static struct AST *
adjust_ast (struct AST *ast)
{
    switch (ast->ast_type) {
    /* for declarations */
    case AST_init_declarator_dec:
    case AST_init_declarator_dec_init:
	return ast->u.child [0]; /* AST_declarator_direct */

    /* for enumerator-constants */
    case AST_enumerator_id:
    case AST_enumerator_id_init:
	return ast->u.child [0]; /* AST_IDENTIFIER */

    /* for new-type parameters */
    case AST_parameter_declaration_dec:
	return ast->u.child [1]; /* AST_declarator_direct */

    /* for old-type parameters */
    case AST_declaration_list_single:
    case AST_declaration_list_pair:
	return NULL;             /* implicit parameter declarations */
    case AST_declarator_direct:
	return ast;              /* AST_declarator_direct */

    case AST_IDENTIFIER:
	return NULL;             /* implicit function declarations */

    default:
	return ast;
    }
}

static char *
case_default_fmt (struct AST *ast_switch)
{
    int    i, case_num;
    struct Attribute *attr;
    char   *buf, *p, ref [20];

    attr = get_block_sw_table (ast_switch);
    assert (attr != NULL);
    assert (attr->kind == ATTR_SWITCH_TABLE);
    case_num = attr->u.a_sw_table.case_num;

    /* 
       here is an over estimation:
	   the length of ID like X12345678 < 20.
	   the length of "case_refs=\"\" switch_refs=\"\"" < 50.
     */
    buf = emalloc ((case_num + 1) * 20 + 50); /* 1 for default ID */

    p = buf;
#if defined (CASE_REFS) || defined (DEFAULT_REF)
    assert (0);
#endif
#define CASE_REFS    "case_refs=\""
#define DEFAULT_REF  "\" default_ref=\"X%x\""
    sprintf (p, CASE_REFS);
    p += strlen (CASE_REFS);

    for (i = 0; i < attr->u.a_sw_table.case_num; i++) {
	sprintf (ref, "X%x ", (unsigned) attr->u.a_sw_table.ast_case [i]);
	sprintf (p, ref);
	p += strlen (ref);
    }
    sprintf (p, DEFAULT_REF, (unsigned) attr->u.a_sw_table.ast_default);
    return buf;
#undef CASE_REFS
#undef DEFAULT_REF
}

static char *
break_fmt (struct AST *ast_break)
{
    char *buf;
    assert (ast_break->ast_type == AST_jump_statement_break);
    buf = emalloc (40);
    sprintf (buf, "break_ref=\"X%x\"",
	     (unsigned) get_break_exit (ast_break));
    return buf;
}

static char *
continue_fmt (struct AST *ast_continue)
{
    char *buf;
    buf = emalloc (40);
    assert (ast_continue->ast_type == AST_jump_statement_continue);
    sprintf (buf, "continue_ref=\"X%x\"",
	     (unsigned) get_continue_exit (ast_continue));
    return buf;
}

static char *
identifier_fmt (struct AST *ast_identifier)
{
    struct Symbol *sym;
    char *buf;

    assert ((ast_identifier->ast_type == AST_IDENTIFIER)
	    || (ast_identifier->ast_type == AST_ENUMERATION_CONSTANT)
	    || (ast_identifier->ast_type == AST_TYPEDEF_NAME));
    sym = sym_lookup (ast_identifier->u.string, NS_NORMAL);
    buf = emalloc (40);
    if (sym == NULL) {
	sprintf (buf, "ref=\"X%x\"", (unsigned) NULL);
    } else {
	sprintf (buf, "ref=\"X%x\"", (unsigned) adjust_ast (sym->ast));
    }
    return buf;
}

static char *
make_string_escaped (char *str)
{
    int  i, j;
    int  len;
    char *ret;

    /* first, count the length including backslashes */
    len = 1; /* for the last NUL character '\0' */
    for (i = 0; str [i] != '\0'; i++) {
	switch (str [i]) {
	case '\n': case '\t': case '\v': case '\b':
	case '\r': case '\f': case '\a': case '\\':
	case '\?': case '\'': case '\"':
	    len += 2;     /* '\n' */
	    break;
	default:
	    if (' ' <= str [i] && str [i] <= '~') {
		len++;    /* 'X' */
	    } else {
		len += 4; /* '\ooo' */
	    }
	    break;
	}
    }
    /* then, store the converted string into 'ret' */
    ret = malloc (len);
    for (i = 0, j = 0; str [i] != '\0'; i++) {
	switch (str [i]) {
	case '\n': case '\t': case '\v': case '\b':
	case '\r': case '\f': case '\a': case '\\':
	case '\?': case '\'': case '\"':
	    ret [j] = '\\';
	    switch (str [i]) {
	    case '\n': ret [j + 1] = 'n';  break;
	    case '\t': ret [j + 1] = 't';  break;
	    case '\v': ret [j + 1] = 'v';  break;
	    case '\b': ret [j + 1] = 'b';  break;
	    case '\r': ret [j + 1] = 'r';  break;
	    case '\f': ret [j + 1] = 'f';  break;
	    case '\a': ret [j + 1] = 'a';  break;
	    case '\\': ret [j + 1] = '\\'; break;
	    case '\?': ret [j + 1] = '\?'; break;
	    case '\'': ret [j + 1] = '\''; break;
	    case '\"': ret [j + 1] = '\"'; break;
	    default:   assert (0);         break;
	    }
	    j += 2;
	    break;
	default:
	    if (' ' <= str [i] && str [i] <= '~') {
		ret [j] = str [i];
		j++;
	    } else {
		ret [j] = '\\';
		sprintf (&ret [j], "\\%03o", str [i]);
		j += 4;
	    }
	    break;
	}
    }
    ret [j] = '\0';
    j++;
    assert (j == len);
    return ret;
}
/* ---------------------------------------------------------------------- */
